/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_string.h"
#endif
/* End of auto-generated includes */

#include "LfapAR.h"
#include "LfapMsg.h"


/******************************************************************************/
/*                                                                            */
/*  Function: resetLfapAR                                                     */
/*                                                                            */
/*  Resets the internal IEs of the LFAP_AR message structure to 0.            */
/*                                                                            */
/******************************************************************************/
void resetLfapAR (LFAP_AR* ar, lfapui8_t resetFlags)
{
    if (resetFlags == LFAP_RESET_FLAGS) ar->encode_flag = LFAP_NOT_ENCODED;

    lfap_memset(&(ar->command), 0, LFAP_IE_SIZE);
    lfap_memset(&(ar->flow_id), 0, LFAP_IE_SIZE);
    lfap_memset(&(ar->fas_ip_addr), 0, LFAP_IE_SIZE);
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapAR                                                      */
/*                                                                            */
/*  Initializes the internals of the LFAP_AR message structure.               */
/*                                                                            */
/******************************************************************************/
void initLfapAR (LFAP_AR* ar)
{
    resetLfapAR(ar, LFAP_RESET_FLAGS);
    lfap_buffer_init(&(ar->lfap_msg), 0, 0);

    ar->encode_flag = LFAP_NOT_ENCODED;

    ar->hdr.opcode = LFAP_AR_OPCODE;
    ar->hdr.status = LFAP_MSG_SUCCESS;
    ar->hdr.msgid = getNextLfapMsgId();
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapARcommand                                               */
/*                                                                            */
/*  Initializes the internals of the LFAP_AR message structure, along with    */
/*  setting the command IE (the mandatory IE for an LFAP_AR message).         */
/*                                                                            */
/******************************************************************************/
void initLfapARcommand (LFAP_AR* ar, lfapui8_t command)
{
    initLfapAR(ar);
    setCommandLfapAR(ar, command);
}

/******************************************************************************/
/*                                                                            */
/*  Function: initRcvdLfapAR                                                  */
/*                                                                            */
/*  Initializes the internals of the LFAP_AR message structure, from a        */
/*  received message.  LFAP_SUCCESS is returned upon success, LFAP_ERROR is   */
/*  returned if there is an error. lfap_buffer_get_error(&ar->lfap_msg) can   */
/* be used if there is an error.                                              */
/*                                                                            */
/******************************************************************************/
int initRcvdLfapAR (LFAP_AR* ar, void* buf, int buf_len)
{
    LFAPIE ie;
    int get_header_val = LFAP_SUCCESS;

    resetLfapAR(ar, LFAP_RESET_FLAGS);
    
    lfap_buffer_init(&ar->lfap_msg, buf, buf_len);
    lfap_buffer_insert(&ar->lfap_msg, buf_len);
    ar->encode_flag = LFAP_ENCODE_COMPLETE;

    get_header_val = lfap_get_header(&ar->lfap_msg, &ar->hdr);
    if (get_header_val == LFAP_SUCCESS) {
        lfap_memset(&ie, 0, LFAP_IE_SIZE);
        while (lfap_get_IE(&ar->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
          setComponentLfapAR(ar, &ie);
    }

    return get_header_val;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setComponentLfapAR                                              */
/*                                                                            */
/*  Sets one of the internal IEs of the LFAP_AR message structure.            */
/*                                                                            */
/******************************************************************************/
int setComponentLfapAR (LFAP_AR* ar, LFAPIE* ie)
{
    int rc = LFAP_SUCCESS;
    switch (ie->type) {
        case LFAP_COMMAND_CODE:
            ar->command = *ie;
            break;
        case LFAP_FLOW_ID:
            ar->flow_id = *ie;
            break;
        case LFAP_FAS_IP_ADDR:
            ar->fas_ip_addr = *ie;
            break;
        default:
            rc = LFAP_ERROR;
            break;
    }
    if (rc == LFAP_SUCCESS)
        ar->encode_flag |= LFAP_ENCODE_IE;

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setCommandLfapAR                                                */
/*                                                                            */
/*  Sets the command IE of the LFAP_AR message structure.                     */
/*                                                                            */
/******************************************************************************/
void setCommandLfapAR (LFAP_AR* ar, lfapui8_t command)
{
    ar->command.type = LFAP_COMMAND_CODE;
    ar->command.rsvd_ui8.reserved = 0;
    ar->command.rsvd_ui8.val = command;

    ar->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFlowIdLfapAR                                                 */
/*                                                                            */
/*  Sets the flow_id IE of the LFAP_AR message structure.                     */
/*                                                                            */
/******************************************************************************/
void setFlowIdLfapAR (LFAP_AR* ar, lfapui32_t fas_id_hi, lfapui32_t fas_id_lo,
                      lfapui32_t cce_id)
{
  ar->flow_id.fid_ie.type = LFAP_FLOW_ID;
  ar->flow_id.fid_ie.fas_id[0] = fas_id_hi;
  ar->flow_id.fid_ie.fas_id[1] = fas_id_lo;
  ar->flow_id.fid_ie.cce_id = cce_id;
  
  ar->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFASIpAddrLfapAR                                              */
/*                                                                            */
/*  Sets the fas_ip_addr IE of the LFAP_AR message structure.                 */
/*                                                                            */
/******************************************************************************/
void setFASIpAddrLfapAR (LFAP_AR* ar, char* addr, lfapui16_t family,
                         lfapui16_t len)
{
  if ((family == LFAP_IP_VERSION_4) || (family == LFAP_IP_VERSION_6)) {
    ar->fas_ip_addr.addr_ie.type = LFAP_FAS_IP_ADDR;
    ar->fas_ip_addr.addr_ie.val = addr;
    ar->fas_ip_addr.addr_ie.family = family;
    ar->fas_ip_addr.addr_ie.len = len;
    
    ar->encode_flag |= LFAP_ENCODE_IE;
  }
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeIEsLfapAR                                                 */
/*                                                                            */
/*  Encodes AR IE's to the fun_buf.                                           */
/*                                                                            */
/******************************************************************************/
static void encodeIEsLfapAR (LFAP_AR* ar)
{
  LFAPBuffer* ar_buf = &(ar->lfap_msg);

    if (ar->command.type)
        lfap_add_IE(ar_buf, &ar->command);
    if (ar->flow_id.type)
        lfap_add_IE(ar_buf, &ar->flow_id);
    if (ar->fas_ip_addr.type)
        lfap_add_IE(ar_buf, &ar->fas_ip_addr);
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeFixedLfapAR                                               */
/*                                                                            */
/*  Encodes the fixed portion of the LFAP_AR message by placing the IEs into  */
/*  the fixed portion of the LFAPBuffer.  If fixed information is desired to  */
/*  be encoded, this must be called before any other encode routine is        */
/*  called.  The IEs of the LFAP_AR message are reset after the encoding is   */
/*  complete.                                                                 */
/*                                                                            */
/******************************************************************************/
int encodeFixedLfapAR (LFAP_AR* ar)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* ar_buf = &(ar->lfap_msg);

  if ((!((ar->encode_flag & LFAP_ENCODE_COMPLETE) ||
         (ar->encode_flag & LFAP_ENCODE_RECORD))) &&
      (ar->encode_flag & LFAP_ENCODE_IE)) {
    lfap_begin_message(ar_buf, &ar->hdr);
    lfap_begin_multiple(ar_buf);
    
    lfap_begin_fixed(ar_buf);
    encodeIEsLfapAR(ar);
    rc = lfap_end_fixed(ar_buf);
    
    ar->encode_flag &= ~LFAP_ENCODE_IE;
    ar->encode_flag |= LFAP_ENCODE_FIXED;
    resetLfapAR(ar, LFAP_NO_RESET_FLAGS);
  }
  
  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeRecordLfapAR                                              */
/*                                                                            */
/*  Encodes a single record of the LFAP_AR message by placing the IEs into    */
/*  the LFAPBuffer.  The IEs of the LFAP_AR message are reset after the       */
/*  encoding is complete.                                                     */
/*                                                                            */
/******************************************************************************/
int encodeRecordLfapAR (LFAP_AR* ar)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* ar_buf = &(ar->lfap_msg);
  
  if ((!(ar->encode_flag & LFAP_ENCODE_COMPLETE)) &&
      (ar->encode_flag & LFAP_ENCODE_IE)) {
    if (!(ar->encode_flag & LFAP_ENCODE_RECORD) &&
        !(ar->encode_flag & LFAP_ENCODE_FIXED)) {
      lfap_begin_message(ar_buf, &ar->hdr);
      lfap_begin_multiple(ar_buf);
    }
    lfap_begin_record(ar_buf);
    encodeIEsLfapAR(ar);
    rc = lfap_end_record(ar_buf);
    
    ar ->encode_flag &= ~LFAP_ENCODE_IE;
    ar->encode_flag |= LFAP_ENCODE_RECORD;
    resetLfapAR(ar, LFAP_NO_RESET_FLAGS);
  }
  
  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeCompletedLfapAR                                           */
/*                                                                            */
/*  Finishes the encoding process of the LFAP_AR message into the LFAPBuffer. */
/*  The LFAP_AR message (in the form of the LFAPBuffer) is ready to be sent.  */
/*                                                                            */
/******************************************************************************/
int encodeCompletedLfapAR (LFAP_AR* ar)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* ar_buf = &(ar->lfap_msg);
  
  if (!(ar->encode_flag & LFAP_ENCODE_COMPLETE)) {
    if ((ar->encode_flag & LFAP_ENCODE_RECORD) ||
        (ar->encode_flag & LFAP_ENCODE_FIXED)) {
      if (ar->encode_flag & LFAP_ENCODE_IE) {
        lfap_begin_record(ar_buf);
        encodeIEsLfapAR(ar);
        lfap_end_record(ar_buf);
      }
      lfap_end_multiple(ar_buf);
    }
    else {
      lfap_begin_message(ar_buf, &ar->hdr);
      encodeIEsLfapAR(ar);
    }
    rc = lfap_end_message(ar_buf);
    ar->encode_flag |= LFAP_ENCODE_COMPLETE;
  }
  
  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: decodeNextRecordLfapAR                                          */
/*                                                                            */
/*  Decodes the next record in the multiple record LFAP_AR message.  The IEs  */
/*  of the record are placed into the IEs of the LFAP_AR message.  If there   */
/*  are no more records, the LFAP_AR's IEs are reset and LFAP_EMPTY_BUFFER    */
/*  is returned, else LFAP_SUCCESS is returned.                               */
/*                                                                            */
/******************************************************************************/
int decodeNextRecordLfapAR (LFAP_AR* ar)
{
  LFAPIE ie;
  int msg_decoded = lfap_next_record(&ar->lfap_msg);
  
  resetLfapAR(ar, LFAP_NO_RESET_FLAGS);
  
  if (msg_decoded != LFAP_EMPTY_BUFFER) {
    lfap_memset(&ie, 0, LFAP_IE_SIZE);
    while (lfap_get_IE(&ar->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
      setComponentLfapAR(ar, &ie);
  }
  
  return msg_decoded;
}

/******************************************************************************/
/*                                                                            */
/*  Function: verifyLfapAR                                                    */
/*                                                                            */
/*  Verifies which IEs of the received LFAP_AR message are encoded.           */
/*  LFAP_AR_INVALID is returned if the mandatory IEs are not encoded,         */
/*  otherwise the returned value indicates which IEs are encoded in the       */
/*  message.                                                                  */
/*                                                                            */
/******************************************************************************/
lfapui8_t verifyLfapAR (LFAP_AR* ar)
{
    lfapui8_t verify = LFAP_AR_INVALID;

    if (ar->command.type)
    {
        verify = LFAP_AR_COMMAND;
        if (ar->flow_id.type)
            verify += LFAP_AR_FLOW_ID;
        if (ar->fas_ip_addr.type)
            verify += LFAP_AR_FAS_IP_ADDR;
    }

    return verify;
}
