/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_string.h"
#endif
/* End of auto-generated includes */

#include "LfapARA.h"
#include "LfapMsg.h"


/******************************************************************************/
/*                                                                            */
/*  Function: resetLfapARA                                                    */
/*                                                                            */
/*  Resets the internal IEs of the LFAP_ARA message structure to 0.           */
/*                                                                            */
/******************************************************************************/
void resetLfapARA (LFAP_ARA* ara, lfapui8_t resetFlags)
{
    if ( resetFlags == LFAP_RESET_FLAGS) ara->encode_flag = LFAP_NOT_ENCODED;

    lfap_memset(&(ara->flow_id), 0, LFAP_IE_SIZE);
    lfap_memset(&(ara->fail_code), 0, LFAP_IE_SIZE);
    lfap_memset(&(ara->flow_prefix), 0, LFAP_IE_SIZE);
    lfap_memset(&(ara->time), 0, LFAP_IE_SIZE);
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapARA                                                     */
/*                                                                            */
/*  Initializes the internals of the LFAP_ARA message structure.              */
/*                                                                            */
/******************************************************************************/
void initLfapARA (LFAP_ARA* ara)
{
    resetLfapARA(ara, LFAP_RESET_FLAGS);
    lfap_buffer_init(&(ara->lfap_msg), 0, 0);

    ara->encode_flag = LFAP_NOT_ENCODED;

    ara->hdr.opcode = LFAP_ARA_OPCODE;
    ara->hdr.status = LFAP_MSG_SUCCESS;
    ara->hdr.msgid = getNextLfapMsgId();
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapARAfromAR                                               */
/*                                                                            */
/*  Initializes the internals of the LFAP_ARA message structure, from a       */
/*  received LFAP_AR message.                                                 */
/*                                                                            */
/******************************************************************************/
void initLfapARAfromAR (LFAP_ARA* ara, LFAP_AR* ar)
{
    resetLfapARA(ara, LFAP_RESET_FLAGS);
    lfap_buffer_init(&(ara->lfap_msg), 0, 0);

    ara->encode_flag = LFAP_NOT_ENCODED;

    ara->hdr.opcode = LFAP_ARA_OPCODE;
    ara->hdr.status = ar->hdr.status;
    ara->hdr.msgid = ar->hdr.msgid;
}

/******************************************************************************/
/*                                                                            */
/*  Function: initRcvdLfapARA                                                 */
/*                                                                            */
/*  Initializes the internals of the LFAP_ARA message structure, from a       */
/*  received message.  LFAP_SUCCESS is returned upon success, LFAP_ERROR is   */
/*  returned if there is an error. lfap_buffer_get_error(&ara->lfap_msg) can  */
/*  be used if there is an error.                                             */
/*                                                                            */
/******************************************************************************/
int initRcvdLfapARA (LFAP_ARA* ara, void* buf, int buf_len)
{
  LFAPIE ie;
  int get_header_val = LFAP_SUCCESS;
  
  resetLfapARA(ara, LFAP_RESET_FLAGS);
  
  lfap_buffer_init(&ara->lfap_msg, buf, buf_len);
  lfap_buffer_insert(&ara->lfap_msg, buf_len);
  
  get_header_val = lfap_get_header(&ara->lfap_msg, &ara->hdr);
  if (get_header_val == LFAP_SUCCESS) {
    lfap_memset(&ie, 0, LFAP_IE_SIZE);
    while (lfap_get_IE(&ara->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
      setComponentLfapARA(ara, &ie);
  }
  
  return get_header_val;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setComponentLfapARA                                             */
/*                                                                            */
/*  Sets one of the internal IEs of the LFAP_ARA message structure.           */
/*                                                                            */
/******************************************************************************/
int setComponentLfapARA (LFAP_ARA* ara, LFAPIE* ie)
{
    int rc = LFAP_SUCCESS;
    switch (ie->type) {
        case LFAP_FAIL_CODE:
            ara->fail_code = *ie;
            break;
        case LFAP_FLOW_ID_PREFIX:
            ara->flow_prefix = *ie;
            break;
        case LFAP_FLOW_ID:
            ara->flow_id = *ie;
            break;
        case LFAP_TIME:
            ara->time = *ie;
            break;
        default:
            rc = LFAP_ERROR;
            break;
    }
    if (rc == LFAP_SUCCESS)
        ara->encode_flag |= LFAP_ENCODE_IE;

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFailCodeLfapARA                                              */
/*                                                                            */
/*  Sets the fail_code IE of the LFAP_ARA message structure.                  */
/*                                                                            */
/******************************************************************************/
void setFailCodeLfapARA (LFAP_ARA* ara, lfapui32_t code)
{
    ara->fail_code.rsvd_ui8.type = LFAP_FAIL_CODE;
    ara->fail_code.rsvd_ui8.reserved = 0;
    ara->fail_code.rsvd_ui8.val = code;

    ara->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFlowIdLfapARA                                                */
/*                                                                            */
/*  Sets the flow_id IE of the LFAP_ARA message structure.                    */
/*                                                                            */
/******************************************************************************/
void setFlowIdLfapARA (LFAP_ARA* ara, lfapui32_t fas_id_hi,
                       lfapui32_t fas_id_lo, lfapui32_t cce_id)
{
  ara->flow_id.fid_ie.type = LFAP_FLOW_ID;
  ara->flow_id.fid_ie.fas_id[0] = fas_id_hi;
  ara->flow_id.fid_ie.fas_id[1] = fas_id_lo;
  ara->flow_id.fid_ie.cce_id = cce_id;

  ara->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFlowPrefixLfapARA                                            */
/*                                                                            */
/*  Sets the flow_prefix IE of the LFAP_ARA message structure.                */
/*                                                                            */
/******************************************************************************/
void setFlowPrefixLfapARA (LFAP_ARA* ara, lfapui32_t fas_id_hi,
                           lfapui32_t fas_id_lo)
{
    ara->flow_prefix.pfx.type = LFAP_FLOW_ID_PREFIX;
    ara->flow_prefix.pfx.fas_id[0] = fas_id_hi;
    ara->flow_prefix.pfx.fas_id[1] = fas_id_lo;

    ara->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setTimeLfapARA                                                  */
/*                                                                            */
/*  Sets the time IE of the LFAP_ARA message structure.                       */
/*                                                                            */
/******************************************************************************/
void setTimeLfapARA (LFAP_ARA* ara, lfapui32_t time)
{
    ara->time.i32_ie.type = LFAP_TIME;
    ara->time.i32_ie.val = time;

    ara->encode_flag |= LFAP_ENCODE_IE;

}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeIEsLfapARA                                                */
/*                                                                            */
/*  Encodes ARA IE's to the ara_buf.                                          */
/*                                                                            */
/******************************************************************************/
static void encodeIEsLfapARA (LFAP_ARA* ara)
{
  LFAPBuffer* ara_buf = &(ara->lfap_msg);


  if (ara->fail_code.type)
    lfap_add_IE(ara_buf, &ara->fail_code);
  if (ara->flow_id.type)
    lfap_add_IE(ara_buf, &ara->flow_id);
  if (ara->flow_prefix.type)
    lfap_add_IE(ara_buf, &ara->flow_prefix);
  if (ara->time.type)
    lfap_add_IE(ara_buf, &ara->time);
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeFixedLfapARA                                              */
/*                                                                            */
/*  Encodes the fixed portion of the LFAP_ARA message by placing the IEs into */
/*  the  fixed portion of the LFAPBuffer.  If fixed information is desired to */
/*  be encoded, this must be called before any other encode routine is        */
/*  called.  The IEs of the LFAP_ARA message are reset after the encoding is  */
/*  complete.                                                                 */
/*                                                                            */
/******************************************************************************/
int encodeFixedLfapARA (LFAP_ARA* ara)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* ara_buf = &(ara->lfap_msg);
  
  if ((!((ara->encode_flag & LFAP_ENCODE_COMPLETE) ||
         (ara->encode_flag & LFAP_ENCODE_RECORD))) &&
      (ara->encode_flag & LFAP_ENCODE_IE)) {
    lfap_begin_message(ara_buf, &ara->hdr);
    lfap_begin_multiple(ara_buf);
    
    lfap_begin_fixed(ara_buf);
    encodeIEsLfapARA(ara);
    rc = lfap_end_fixed(ara_buf);
    
    ara->encode_flag &= ~LFAP_ENCODE_IE;
    ara->encode_flag |= LFAP_ENCODE_FIXED;
    resetLfapARA(ara, LFAP_NO_RESET_FLAGS);
  }

  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeRecordLfapARA                                             */
/*                                                                            */
/*  Encodes a single record of the LFAP_ARA message by placing the IEs into   */
/*  the LFAPBuffer.  The IEs of the LFAP_ARA message are reset after the      */
/*  encoding is complete.                                                     */
/*                                                                            */
/******************************************************************************/
int encodeRecordLfapARA (LFAP_ARA* ara)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* ara_buf = &(ara->lfap_msg);
  
  if ((!(ara->encode_flag & LFAP_ENCODE_COMPLETE)) &&
      (ara->encode_flag & LFAP_ENCODE_IE)) {
    if (!(ara->encode_flag & LFAP_ENCODE_RECORD) &&
        !(ara->encode_flag & LFAP_ENCODE_FIXED)) {
      lfap_begin_message(ara_buf, &ara->hdr);
      lfap_begin_multiple(ara_buf);
    }
    lfap_begin_record(ara_buf);
    encodeIEsLfapARA(ara);
    rc = lfap_end_record(ara_buf);
    
    ara->encode_flag &= ~LFAP_ENCODE_IE;
    ara->encode_flag |= LFAP_ENCODE_RECORD;
    resetLfapARA(ara, LFAP_NO_RESET_FLAGS);
  }
  
  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeCompletedLfapARA                                          */
/*                                                                            */
/*  Finishes the encoding process of the LFAP_ARA message into the            */
/*  LFAPBuffer.  The LFAP_ARA message (in the form of the LFAPBuffer) is      */
/*  ready to be sent.                                                         */
/*                                                                            */
/******************************************************************************/
int encodeCompletedLfapARA (LFAP_ARA* ara)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* ara_buf = &(ara->lfap_msg);
  
  if (!(ara->encode_flag & LFAP_ENCODE_COMPLETE)) {
    if ((ara->encode_flag & LFAP_ENCODE_RECORD) ||
        (ara->encode_flag & LFAP_ENCODE_FIXED)) {
      if (ara->encode_flag & LFAP_ENCODE_IE) {
        lfap_begin_record(ara_buf);
        encodeIEsLfapARA(ara);
        lfap_end_record(ara_buf);
      }
      lfap_end_multiple(ara_buf);
    }
    else {
      lfap_begin_message(ara_buf, &ara->hdr);
      encodeIEsLfapARA(ara);
    }
    rc = lfap_end_message(ara_buf);
    ara->encode_flag |= LFAP_ENCODE_COMPLETE;
  }
  
  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: decodeNextRecordLfapARA                                         */
/*                                                                            */
/*  Decodes the next record in the multiple record LFAP_ARA message.  The IEs */
/*  of the record are placed into the IEs of the LFAP_ARA message.  If there  */
/*  are no more records, the LFAP_ARA's IEs are reset and LFAP_EMPTY_BUFFER   */
/*  is returned, else LFAP_SUCCESS is returned.                               */
/*                                                                            */
/******************************************************************************/
int decodeNextRecordLfapARA (LFAP_ARA* ara)
{
  LFAPIE ie;
  int msg_decoded = lfap_next_record(&ara->lfap_msg);
  
  resetLfapARA(ara, LFAP_NO_RESET_FLAGS);
  
  if (msg_decoded != LFAP_EMPTY_BUFFER) {
    lfap_memset(&ie, 0, LFAP_IE_SIZE);
    while (lfap_get_IE(&ara->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
      setComponentLfapARA(ara, &ie);
  }
  
  return msg_decoded;
}

/******************************************************************************/
/*                                                                            */
/*  Function: verifyLfapARA                                                   */
/*                                                                            */
/*  Verifies which IEs of the received LFAP_ARA message are encoded.          */
/*  LFAP_ARA_VALID is always returned, as an LFAP_ARA message does not have   */
/*  any mandatory IEs.  The returned value indicates which IEs are encoded in */
/*  the message.                                                              */
/*                                                                            */
/******************************************************************************/
lfapui8_t verifyLfapARA (LFAP_ARA* ara)
{
    lfapui8_t verify = LFAP_ARA_VALID;

    if (ara->flow_id.type)
        verify += LFAP_ARA_FLOW_ID;
    if (ara->fail_code.type)
        verify += LFAP_ARA_FAIL_CODE;
    if (ara->flow_prefix.type)
        verify += LFAP_ARA_FLOW_PREFIX;
    if (ara->time.type)
        verify += LFAP_ARA_TIME;

    return verify;
}
