/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/*
 * routines for manipulating LFAPBuffer's
 */

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_stdlib.h"
#include "mls_string.h"
#endif
/* End of auto-generated includes */

#include "LfapPrivate.h"


/*
 * Initialize an LFAPBuffer.
 */
int lfap_buffer_init(LFAPBuffer *lbuf, void *buf, int buflen)
{
  lfap_memset(lbuf, 0, SIZEOF_LFAP_BUFFER);

  if (buflen > 0) {
	if (buf != NULL) {
		lbuf->buf = (lfapui8_t*) buf;
		lbuf->buflen = buflen;
	}
	else {
		return lfap_buffer_alloc(lbuf,buflen);
	}
  }

  return LFAP_SUCCESS;
}


/*
 * Prepare an LFAPBuffer for reuse.
 */
void lfap_buffer_reset(LFAPBuffer *lbuf)
{
  lfap_buffer_init(lbuf, lbuf->buf, lbuf->buflen);
}


/*
 * Attempt to extract "nbyte" bytes from an LFAPBuffer.
 */
lfapui8_t* lfap_buffer_extract(LFAPBuffer *lbuf, int nbyte)
{
  lfapui8_t *cbuf;

  if (lbuf->msglen - lbuf->off < nbyte) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return NULL;
  }

  cbuf = lbuf->buf + lbuf->off;
  lbuf->off += nbyte;

  return cbuf;
}


/*
 * Attempt to insert "nbyte" bytes into an LFAPBuffer.  The buffer will
 * be resized if necessary.
 */
lfapui8_t* lfap_buffer_insert(LFAPBuffer *lbuf, int nbyte)
{
  lfapui8_t *cbuf;

  if (lbuf->buflen - lbuf->msglen < nbyte && lfap_buffer_alloc(lbuf, nbyte))
	return NULL;

  cbuf = lbuf->buf + lbuf->msglen;
  lbuf->msglen += nbyte;

  return cbuf;
}


/*
 * Default buffer allocation routine.  The buffer will be resized to hold
 * at least "nbyte" additional bytes.
 */
static int default_buffer_alloc(LFAPBuffer *lbuf, int nbyte)
{
  int pow2, need;

  need = nbyte + lbuf->msglen;

  if (need > LFAP_MAX_MSGSIZE) {
	lbuf->err = LFAP_MSGSIZE;
	return LFAP_ERROR;
  }

  /*
   *  To prevent excessive resizing:  (1) don't allocate buffers of less than
   *  256 bytes and (2) round the buffer size up to a power of 2.
   */
  for (pow2 = 256; pow2 < need; pow2 <<= 1) {}

  lbuf->buf = lbuf->buf ? (lfapui8_t*)lfap_realloc(lbuf->buf, pow2)
                        : (lfapui8_t*)lfap_malloc(pow2);

  if (lbuf->buf == NULL) {
	lbuf->err = LFAP_NOMEM;
	return LFAP_ERROR;
  }

  lbuf->buflen = pow2;

  return LFAP_SUCCESS;
}

/*
 *  Default buffer deallocation routine.
 */
static void default_buffer_free(LFAPBuffer *lbuf)
{
  lfap_free(lbuf->buf);
}


/*
 *  Global function pointers so that the default buffer allocation and
 *  deallocation may be overridden by user-defined routines.
 */

lfap_buffer_alloc_t	lfap_buffer_alloc	= default_buffer_alloc;
lfap_buffer_free_t	lfap_buffer_free	= default_buffer_free;
