/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 2001
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Paul Calato
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_string.h"
#endif
/* End of auto-generated includes */

#include "LfapCR.h"
#include "LfapMsg.h"


/******************************************************************************/
/*                                                                            */
/*  Function: resetLfapCR                                                     */
/*                                                                            */
/*  Resets the internal IEs of the LFAP_CR message structure to 0.            */
/*                                                                            */
/******************************************************************************/
void resetLfapCR (LFAP_CR* cr, lfapui8_t resetFlags)
{
    if (resetFlags == LFAP_RESET_FLAGS) cr->encode_flag = LFAP_NOT_ENCODED;
    lfap_memset(&(cr->fas_ip_addr), 0, LFAP_IE_SIZE);
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapCR                                                      */
/*                                                                            */
/*  Initializes the internals of the LFAP_CR message structure.               */
/*                                                                            */
/******************************************************************************/
void initLfapCR (LFAP_CR* cr)
{
    resetLfapCR(cr, LFAP_RESET_FLAGS);
    lfap_buffer_init(&(cr->lfap_msg), 0, 0);

    cr->encode_flag = LFAP_NOT_ENCODED;

    cr->hdr.opcode = LFAP_CR_OPCODE;
    cr->hdr.status = LFAP_MSG_SUCCESS;
    cr->hdr.msgid = getNextLfapMsgId();
}

/******************************************************************************/
/*                                                                            */
/*  Function: initRcvdLfapCR                                                  */
/*                                                                            */
/*  Initializes the internals of the LFAP_CR message structure, from a        */
/*  received message.  LFAP_SUCCESS is returned upon success, LFAP_ERROR is   */
/*  returned if there is an error.  lfap_buffer_get_error(&cr->lfap_msg) can be    */
/*  used if there is an error.                                                */
/*                                                                            */
/******************************************************************************/
int initRcvdLfapCR (LFAP_CR* cr, void* buf, int buf_len)
{
    LFAPIE ie;
    int get_header_val = LFAP_SUCCESS;

    resetLfapCR(cr, LFAP_RESET_FLAGS);

    lfap_buffer_init(&cr->lfap_msg, buf, buf_len);
    lfap_buffer_insert(&cr->lfap_msg, buf_len);
    cr->encode_flag = LFAP_ENCODE_COMPLETE;

    get_header_val = lfap_get_header(&cr->lfap_msg, &cr->hdr);
    if (get_header_val == LFAP_SUCCESS) {
        lfap_memset(&ie, 0, LFAP_IE_SIZE);
        while (lfap_get_IE(&cr->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
        setComponentLfapCR(cr, &ie);
    }

    return get_header_val;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setComponentLfapCR                                              */
/*                                                                            */
/*  Sets one of the internal IEs of the LFAP_CR message structure.            */
/*                                                                            */
/******************************************************************************/
int setComponentLfapCR (LFAP_CR* cr, LFAPIE* ie)
{
    int rc = LFAP_SUCCESS;
    switch (ie->type) {
        case LFAP_TRANSPORT_FAS_IP:
            cr->fas_ip_addr = *ie;
            break;
        default:
            rc = LFAP_ERROR;
            break;
    }

    if (rc == LFAP_SUCCESS)
        cr->encode_flag |= LFAP_ENCODE_IE;

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFASIpAddrLfapCR                                              */
/*                                                                            */
/*  Sets the fas_ip_addr IE of the LFAP_CR message structure. The mandatory   */
/*  IE for the message.                                                       */
/*                                                                            */
/******************************************************************************/
void setFASIpAddrLfapCR (LFAP_CR* cr, char* addr, 
                         lfapui16_t family,
                         lfapui16_t len)
{
    if ((family == LFAP_IP_VERSION_4) || (family == LFAP_IP_VERSION_6)) {
        cr->fas_ip_addr.addr_ie.type = LFAP_TRANSPORT_FAS_IP;
        cr->fas_ip_addr.addr_ie.val = addr;
        cr->fas_ip_addr.addr_ie.family = family;
        cr->fas_ip_addr.addr_ie.len = len;

        cr->encode_flag |= LFAP_ENCODE_IE;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeIEsLfapCR                                                 */
/*                                                                            */
/*  Encodes CR IE's to the fun_buf.                                           */
/*                                                                            */
/******************************************************************************/
static void encodeIEsLfapCR (LFAP_CR* cr)
{
    LFAPBuffer* cr_buf = &(cr->lfap_msg);

    if (cr->fas_ip_addr.type)
        lfap_add_IE(cr_buf, &cr->fas_ip_addr);
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeFixedLfapCR                                               */
/*                                                                            */
/*  Encodes the fixed portion of the LFAP_CR message by placing the IEs into  */
/*  the fixed portion of the LFAPBuffer.  If fixed information is desired to  */
/*  be encoded, this must be called before any other encode routine is        */
/*  called.  The IEs of the LFAP_CR message are reset after the encoding is   */
/*  complete.                                                                 */
/*                                                                            */
/******************************************************************************/
int encodeFixedLfapCR (LFAP_CR* cr)
{
    int rc = LFAP_ERROR;
    LFAPBuffer* cr_buf = &(cr->lfap_msg);

    if ((!((cr->encode_flag & LFAP_ENCODE_COMPLETE) ||
           (cr->encode_flag & LFAP_ENCODE_RECORD))) &&
           (cr->encode_flag & LFAP_ENCODE_IE)) {
        lfap_begin_message(cr_buf, &cr->hdr);
        lfap_begin_multiple(cr_buf);

        lfap_begin_fixed(cr_buf);

        encodeIEsLfapCR(cr);

        rc = lfap_end_fixed(cr_buf);

        cr->encode_flag &= ~LFAP_ENCODE_IE;
        cr->encode_flag |= LFAP_ENCODE_FIXED;
        resetLfapCR(cr, LFAP_NO_RESET_FLAGS);
    }

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeRecordLfapCR                                              */
/*                                                                            */
/*  Encodes a single record of the LFAP_CR message by placing the IEs into    */
/*  the LFAPBuffer.  The IEs of the LFAP_CR message are reset after the       */
/*  encoding is complete.                                                     */
/*                                                                            */
/******************************************************************************/
int encodeRecordLfapCR (LFAP_CR* cr)
{
    int rc = LFAP_ERROR;
    LFAPBuffer* cr_buf = &(cr->lfap_msg);

    if ((!(cr->encode_flag & LFAP_ENCODE_COMPLETE)) &&
          (cr->encode_flag & LFAP_ENCODE_IE)) {
        if (!(cr->encode_flag & LFAP_ENCODE_RECORD) &&
            !(cr->encode_flag & LFAP_ENCODE_FIXED)) {
          lfap_begin_message(cr_buf, &cr->hdr);
          lfap_begin_multiple(cr_buf);
        }
        lfap_begin_record(cr_buf);

        encodeIEsLfapCR(cr);

        rc = lfap_end_record(cr_buf);

        cr ->encode_flag &= ~LFAP_ENCODE_IE;
        cr->encode_flag |= LFAP_ENCODE_RECORD;
        resetLfapCR(cr, LFAP_NO_RESET_FLAGS);
    }

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeCompletedLfapCR                                           */
/*                                                                            */
/*  Finishes the encoding process of the LFAP_CR message into the LFAPBuffer. */
/*  The LFAP_CR message (in the form of the LFAPBuffer) is ready to be sent.  */
/*                                                                            */
/******************************************************************************/
int encodeCompletedLfapCR (LFAP_CR* cr)
{
    int rc = LFAP_ERROR;
    LFAPBuffer* cr_buf = &(cr->lfap_msg);

    if (!(cr->encode_flag & LFAP_ENCODE_COMPLETE)) {
        if ((cr->encode_flag & LFAP_ENCODE_RECORD) ||
            (cr->encode_flag & LFAP_ENCODE_FIXED)) {
            if (cr->encode_flag & LFAP_ENCODE_IE) {
                lfap_begin_record(cr_buf);
                encodeIEsLfapCR(cr);
                lfap_end_record(cr_buf);
            }
            lfap_end_multiple(cr_buf);
        }
        else {
            lfap_begin_message(cr_buf, &cr->hdr);
            encodeIEsLfapCR(cr);
        }
        rc = lfap_end_message(cr_buf);
        cr->encode_flag |= LFAP_ENCODE_COMPLETE;
    }

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: decodeNextRecordLfapCR                                          */
/*                                                                            */
/*  Decodes the next record in the multiple record LFAP_CR message.  The IEs  */
/*  of the record cre placed into the IEs of the LFAP_CR message.  If there   */
/*  are no more records, the LFAP_CR's IEs cre reset and LFAP_EMPTY_BUFFER    */
/*  is returned, else LFAP_SUCCESS is returned.                               */
/*                                                                            */
/******************************************************************************/
int decodeNextRecordLfapCR (LFAP_CR* cr)
{
    LFAPIE ie;
    int msg_decoded = lfap_next_record(&cr->lfap_msg);

    resetLfapCR(cr, LFAP_NO_RESET_FLAGS);

    if (msg_decoded != LFAP_EMPTY_BUFFER) {
        lfap_memset(&ie, 0, LFAP_IE_SIZE);
        while (lfap_get_IE(&cr->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
            setComponentLfapCR(cr, &ie);
    }

    return msg_decoded;
}

/******************************************************************************/
/*                                                                            */
/*  Function: verifyLfapCR                                                    */
/*                                                                            */
/*  Verifies which IEs of the received LFAP_CR message are encoded.           */
/*  LFAP_CR_INVALID is returned if the mandatory IEs are not encoded,         */
/*  otherwise the returned value indicates which IEs are encoded in the       */
/*  message.                                                                  */
/*                                                                            */
/******************************************************************************/
lfapui8_t verifyLfapCR (LFAP_CR* cr)
{
    lfapui8_t verify = LFAP_CR_INVALID;

    if (cr->fas_ip_addr.type)
      verify = LFAP_CR_FAS_IP_ADDR;

    return verify;
}
