/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 2001
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Paul Calato
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_string.h"
#endif
/* End of auto-generated includes */

#include "LfapDR.h"
#include "LfapMsg.h"


/******************************************************************************/
/*                                                                            */
/*  Function: resetLfapDR                                                     */
/*                                                                            */
/*  Resets the internal IEs of the LFAP_DR message structure to 0.            */
/*                                                                            */
/******************************************************************************/
void resetLfapDR (LFAP_DR* dr, lfapui8_t resetFlags)
{
    if (resetFlags == LFAP_RESET_FLAGS) dr->encode_flag = LFAP_NOT_ENCODED;
    lfap_memset(&(dr->fas_ip_addr), 0, LFAP_IE_SIZE);
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapDR                                                      */
/*                                                                            */
/*  Initializes the internals of the LFAP_DR message structure.               */
/*                                                                            */
/******************************************************************************/
void initLfapDR (LFAP_DR* dr)
{
    resetLfapDR(dr, LFAP_RESET_FLAGS);
    lfap_buffer_init(&(dr->lfap_msg), 0, 0);

    dr->encode_flag = LFAP_NOT_ENCODED;

    dr->hdr.opcode = LFAP_DR_OPCODE;
    dr->hdr.status = LFAP_MSG_SUCCESS;
    dr->hdr.msgid = getNextLfapMsgId();
}

/******************************************************************************/
/*                                                                            */
/*  Function: initRcvdLfapDR                                                  */
/*                                                                            */
/*  Initializes the internals of the LFAP_DR message structure, from a        */
/*  received message.  LFAP_SUCCESS is returned upon success, LFAP_ERROR is   */
/*  returned if there is an error.  lfap_buffer_get_error(&dr->lfap_msg) can be    */
/*  used if there is an error.                                                */
/*                                                                            */
/******************************************************************************/
int initRcvdLfapDR (LFAP_DR* dr, void* buf, int buf_len)
{
    LFAPIE ie;
    int get_header_val = LFAP_SUCCESS;

    resetLfapDR(dr, LFAP_RESET_FLAGS);

    lfap_buffer_init(&dr->lfap_msg, buf, buf_len);
    lfap_buffer_insert(&dr->lfap_msg, buf_len);
    dr->encode_flag = LFAP_ENCODE_COMPLETE;

    get_header_val = lfap_get_header(&dr->lfap_msg, &dr->hdr);
    if (get_header_val == LFAP_SUCCESS) {
        lfap_memset(&ie, 0, LFAP_IE_SIZE);
        while (lfap_get_IE(&dr->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
        setComponentLfapDR(dr, &ie);
    }

    return get_header_val;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setComponentLfapDR                                              */
/*                                                                            */
/*  Sets one of the internal IEs of the LFAP_DR message structure.            */
/*                                                                            */
/******************************************************************************/
int setComponentLfapDR (LFAP_DR* dr, LFAPIE* ie)
{
    int rc = LFAP_SUCCESS;
    switch (ie->type) {
        case LFAP_FAS_IP_ADDR:
            dr->fas_ip_addr = *ie;
            break;
        default:
            rc = LFAP_ERROR;
            break;
    }

    if (rc == LFAP_SUCCESS)
        dr->encode_flag |= LFAP_ENCODE_IE;

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFASIpAddrLfapDR                                              */
/*                                                                            */
/*  Sets the fas_ip_addr IE of the LFAP_DR message structure. The mandatory   */
/*                                                                            */
/******************************************************************************/
void setFASIpAddrLfapDR (LFAP_DR* dr, char* addr, 
                         lfapui16_t family,
                         lfapui16_t len)
{
    if ((family == LFAP_IP_VERSION_4) || (family == LFAP_IP_VERSION_6)) {
        dr->fas_ip_addr.addr_ie.type = LFAP_FAS_IP_ADDR;
        dr->fas_ip_addr.addr_ie.val = addr;
        dr->fas_ip_addr.addr_ie.family = family;
        dr->fas_ip_addr.addr_ie.len = len;

        dr->encode_flag |= LFAP_ENCODE_IE;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeIEsLfapDR                                                 */
/*                                                                            */
/*  Encodes DR IE's to the fun_buf.                                           */
/*                                                                            */
/******************************************************************************/
static void encodeIEsLfapDR (LFAP_DR* dr)
{
    LFAPBuffer* dr_buf = &(dr->lfap_msg);

    if (dr->fas_ip_addr.type)
        lfap_add_IE(dr_buf, &dr->fas_ip_addr);
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeFixedLfapDR                                               */
/*                                                                            */
/*  Encodes the fixed portion of the LFAP_DR message by placing the IEs into  */
/*  the fixed portion of the LFAPBuffer.  If fixed information is desired to  */
/*  be encoded, this must be called before any other encode routine is        */
/*  called.  The IEs of the LFAP_DR message are reset after the encoding is   */
/*  complete.                                                                 */
/*                                                                            */
/******************************************************************************/
int encodeFixedLfapDR (LFAP_DR* dr)
{
    int rc = LFAP_ERROR;
    LFAPBuffer* dr_buf = &(dr->lfap_msg);

    if ((!((dr->encode_flag & LFAP_ENCODE_COMPLETE) ||
           (dr->encode_flag & LFAP_ENCODE_RECORD))) &&
           (dr->encode_flag & LFAP_ENCODE_IE)) {
        lfap_begin_message(dr_buf, &dr->hdr);
        lfap_begin_multiple(dr_buf);

        lfap_begin_fixed(dr_buf);
        encodeIEsLfapDR(dr);
        rc = lfap_end_fixed(dr_buf);

        dr->encode_flag &= ~LFAP_ENCODE_IE;
        dr->encode_flag |= LFAP_ENCODE_FIXED;
        resetLfapDR(dr, LFAP_NO_RESET_FLAGS);
    }

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeRecordLfapDR                                              */
/*                                                                            */
/*  Encodes a single record of the LFAP_DR message by placing the IEs into    */
/*  the LFAPBuffer.  The IEs of the LFAP_DR message are reset after the       */
/*  encoding is complete.                                                     */
/*                                                                            */
/******************************************************************************/
int encodeRecordLfapDR (LFAP_DR* dr)
{
    int rc = LFAP_ERROR;
    LFAPBuffer* dr_buf = &(dr->lfap_msg);

    if ((!(dr->encode_flag & LFAP_ENCODE_COMPLETE)) &&
          (dr->encode_flag & LFAP_ENCODE_IE)) {
        if (!(dr->encode_flag & LFAP_ENCODE_RECORD) &&
            !(dr->encode_flag & LFAP_ENCODE_FIXED)) {
            lfap_begin_message(dr_buf, &dr->hdr);
            lfap_begin_multiple(dr_buf);
        }
        lfap_begin_record(dr_buf);
        encodeIEsLfapDR(dr);
        rc = lfap_end_record(dr_buf);

        dr ->encode_flag &= ~LFAP_ENCODE_IE;
        dr->encode_flag |= LFAP_ENCODE_RECORD;
        resetLfapDR(dr, LFAP_NO_RESET_FLAGS);
    }

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeCompletedLfapDR                                           */
/*                                                                            */
/*  Finishes the encoding process of the LFAP_DR message into the LFAPBuffer. */
/*  The LFAP_DR message (in the form of the LFAPBuffer) is ready to be sent.  */
/*                                                                            */
/******************************************************************************/
int encodeCompletedLfapDR (LFAP_DR* dr)
{
    int rc = LFAP_ERROR;
    LFAPBuffer* dr_buf = &(dr->lfap_msg);

    if (!(dr->encode_flag & LFAP_ENCODE_COMPLETE)) {
        if ((dr->encode_flag & LFAP_ENCODE_RECORD) ||
            (dr->encode_flag & LFAP_ENCODE_FIXED)) {
            if (dr->encode_flag & LFAP_ENCODE_IE) {
                lfap_begin_record(dr_buf);
                encodeIEsLfapDR(dr);
                lfap_end_record(dr_buf);
            }
            lfap_end_multiple(dr_buf);
        }
        else {
            lfap_begin_message(dr_buf, &dr->hdr);
            encodeIEsLfapDR(dr);
        }
        rc = lfap_end_message(dr_buf);
        dr->encode_flag |= LFAP_ENCODE_COMPLETE;
    }

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: decodeNextRecordLfapDR                                          */
/*                                                                            */
/*  Decodes the next record in the multiple record LFAP_DR message.  The IEs  */
/*  of the record dre placed into the IEs of the LFAP_DR message.  If there   */
/*  are no more records, the LFAP_DR's IEs dre reset and LFAP_EMPTY_BUFFER    */
/*  is returned, else LFAP_SUCCESS is returned.                               */
/*                                                                            */
/******************************************************************************/
int decodeNextRecordLfapDR (LFAP_DR* dr)
{
    LFAPIE ie;
    int msg_decoded = lfap_next_record(&dr->lfap_msg);

    resetLfapDR(dr, LFAP_NO_RESET_FLAGS);

    if (msg_decoded != LFAP_EMPTY_BUFFER) {
        lfap_memset(&ie, 0, LFAP_IE_SIZE);
        while (lfap_get_IE(&dr->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
            setComponentLfapDR(dr, &ie);
    }

    return msg_decoded;
}

/******************************************************************************/
/*                                                                            */
/*  Function: verifyLfapDR                                                    */
/*                                                                            */
/*  Verifies which IEs of the received LFAP_DR message are encoded.           */
/*  LFAP_DR_INVALID is returned if the mandatory IEs are not encoded,         */
/*  otherwise the returned value indicates which IEs are encoded in the       */
/*  message.                                                                  */
/*                                                                            */
/******************************************************************************/
lfapui8_t verifyLfapDR (LFAP_DR* dr)
{
    lfapui8_t verify = LFAP_DR_INVALID;

    if (dr->fas_ip_addr.type)
      verify = LFAP_DR_FAS_IP_ADDR;

    return verify;
}
