/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#include "mls_syscalls.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "cdefs.h"
#include "nucleus.h"
#include "ni_api.h" 
#include "mls_in.h"
#include "mls_inet.h"
#include "mls_resolv.h"
#include "mls_netdb.h"
#include "mls_reent.h"
#include "mls_stdio.h"
#include "mls_stdlib.h"
#include "mls_select.h"
#include "byteorder.h"
#include "mls_syscalls.h"
#endif


/* End of auto-generated includes */

#include "LfapDebug.h"
#include "LfapMsg.h"

/* CCE rcvd messages / FAS send messages */
#include "LfapAR.h"
#include "LfapARA.h"
#include "LfapVRA.h"

/* CCE send messages / FAS rcvd messages */
#include "LfapFAR.h"
#include "LfapFUN.h"
#include "LfapVR.h"
#include "LfapCR.h"
#include "LfapKA.h"
#include "LfapDR.h"
#include "LfapCAN.h"
#include "LfapCRN.h"
#include "LfapFER.h"

static lfapui16_t previous_msg_id = 0;
static lfapui8_t previous_msg_opcode = 0;
static int current_msg_send_or_recv = LFAP_SEND;
static int previous_msg_send_or_recv = LFAP_SEND;

#ifdef _WIN32

static time_t LFAP_CURRENT_TIME;
static void printLfapCurrentTime_ (void)
{
    time(&LFAP_CURRENT_TIME);
    fprintf(stderr, ctime((const long*) &LFAP_CURRENT_TIME));
}

#else

static struct timeval LFAP_CURRENT_TIME;
static void printLfapCurrentTime_ (void)
{
    gettimeofday(&LFAP_CURRENT_TIME, (struct timezone*) 0);
    fprintf(stderr, ctime((const long*) &LFAP_CURRENT_TIME.tv_sec));
}

#endif  /* _WIN32 */

static void printLfapHeader_ (LFAPHeader* hdr, LFAPBuffer* buf)
{
    lfapui8_t* ptr;
    lfapui8_t version = 0;

    if ((previous_msg_id == hdr->msgid) &&
        (previous_msg_opcode == hdr->opcode) &&
        (previous_msg_send_or_recv == current_msg_send_or_recv)) {
        fprintf(stderr, " (multi-record)");
    }
    else {
        previous_msg_id = hdr->msgid;
        previous_msg_opcode = hdr->opcode;
        previous_msg_send_or_recv = current_msg_send_or_recv;
    }

    fprintf(stderr, "\n\tid:\t\t%d\n\tstatus:\t\t", hdr->msgid);
    if (hdr->status == LFAP_MSG_SUCCESS)
        fprintf(stderr, "success\n");
    else if (hdr->status == LFAP_MSG_VERSION)
        fprintf(stderr, "version error\n");
    else
        fprintf(stderr, "unknown\n");
    fprintf(stderr, "\tversion:\t");
    if ((ptr = buf->buf)) version = (lfapui8_t) *(ptr);
    if (version)
        fprintf(stderr, "%d\n\n", version);         /* receiving */
    else
        fprintf(stderr, "%d\n\n", LFAP_VERSION);    /* sending   */
}

static void printLfapFlowId_ (lfapui16_t type, lfapui32_t cce_id,
                              lfapui32_t fas_id0, lfapui32_t fas_id1)
{
    fprintf(stderr, "flow id:");
    if (type) {
        fprintf(stderr, "\n\t  cce id -\t%lu\n", cce_id);
        fprintf(stderr, "\t  fas pre(hi) -\t%lu\n", fas_id0);
        fprintf(stderr, "\t  fas pre(lo) -\t%lu\n", fas_id1);
    }
    else
        fprintf(stderr, "\tNone\n");
}

static void printLfapFailCode_ (lfapui16_t type, lfapui32_t val)
{
    fprintf(stderr, "\tfail code:\t");
    if (type) {
        switch (val) {
            case LFAP_NO_SUCH_FLOW:
                fprintf(stderr, "no such flow\n");
                break;
            default:
                fprintf(stderr, "NOT VALID\n");
                break;
        }
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapQualifier_ (lfapui16_t type, LFAPQualifierList* list)
{
    lfapui16_t i;
    lfapui32_t lo_val = 0;
    lfapui32_t hi_val = 0;
    LFAPQualifier* qualifier;
    lfapui16_t ul_size = LFAP_UINT32_SIZE;

    fprintf(stderr, "\tqualifiers:\t");
    if (type) {
        fprintf(stderr, "\n");
        for (i = 0; i < list->items; i++) {
      qualifier = &list->q_list[i];
      if (qualifier->len <= ul_size) {
	lfap_memcpy(&lo_val, qualifier->val, ul_size);
            }
            else {
	lfap_memcpy(&hi_val, qualifier->val, ul_size);
	lfap_memcpy(&lo_val, qualifier->val + ul_size, ul_size);
            }
            hi_val = ntohl(hi_val);
            lo_val = ntohl(lo_val);

      switch (qualifier->id) {
                case LFAP_CXN_PRIORITY_ID:
                    fprintf(stderr, "\t  priority -\t");
                    switch (lo_val) {
                        case LFAP_CXN_PRIORITY_PREMIUM:
                            fprintf(stderr, "premium\n");
                            break;
                        case LFAP_CXN_PRIORITY_HIGH:
                            fprintf(stderr, "high\n");
                            break;
                        case LFAP_CXN_PRIORITY_MEDIUM:
                            fprintf(stderr, "medium\n");
                            break;
                        case LFAP_CXN_PRIORITY_LOW:
                            fprintf(stderr, "low\n");
                            break;
                        default:
                            fprintf(stderr, "unknown - %lu\n", lo_val);
                            break;
                    }
                    break;
                case LFAP_CHECKPOINT_ID:
                    fprintf(stderr, "\t  checkpoint -\t");
                    switch (lo_val) {
                        case LFAP_CHECKPOINT_HOURLY:
                            fprintf(stderr, "hourly\n");
                            break;
                        case LFAP_CHECKPOINT_DAILY:
                            fprintf(stderr, "daily\n");
                            break;
                        case LFAP_CHECKPOINT_WEEKLY:
                            fprintf(stderr, "weekly\n");
                            break;
                        case LFAP_CHECKPOINT_5_MINUTE:
                            fprintf(stderr, "5 minutes\n");
                            break;
                        case LFAP_CHECKPOINT_15_MINUTE:
                            fprintf(stderr,"15 minutes\n");
                            break;
                        case LFAP_CHECKPOINT_USER_1:
                            fprintf(stderr,"user 1\n");
                            break;
                        case LFAP_CHECKPOINT_USER_2:
                            fprintf(stderr,"user 2\n");
                            break;
                        default:
                            fprintf(stderr, "unknown - %lu\n", lo_val);
                            break;
                    }
                    break;
                default:
                    fprintf(stderr, "\t  unknown\n");
                    break;
            }
        }
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapAddr_ (lfapui16_t type, LFAPAddress* addr)
{
    lfapui16_t i;

    fprintf(stderr, "addr:\t");
    if (type) {
        fprintf(stderr, "\n\t  family -\t%d\n", addr->family);
    fprintf(stderr, "\t  addrlen -\t%d\n", addr->len);
        fprintf(stderr, "\t  addr -\t");
    for (i = 0; i < addr->len; i++) {
      fprintf(stderr, "%d ", (lfapui8_t) addr->val[i]);
        }
        fprintf(stderr, "\n");
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapSrcPort_ (lfapui16_t type, lfapui32_t port)
{
    fprintf(stderr, "\tsrc port:\t");
    switch (type) {
        case LFAP_SRC_PORT_UDP:
            fprintf(stderr, "udp, %lu\n", port);
            break;
        case LFAP_SRC_PORT_TCP:
            fprintf(stderr, "tcp, %lu\n", port);
            break;
        case LFAP_SRC_SOCKET_IPX:
            fprintf(stderr, "ipx, %lu\n", port);
            break;
        default:
            fprintf(stderr, "None\n");
            break;
    }
}

static void printLfapFlowState_ (lfapui16_t type, lfapui32_t val)
{
    fprintf(stderr, "\tflow state:\t");
    if (type) {
        switch (val) {
            case LFAP_INACTIVE:
                fprintf(stderr, "Inactive\n");
                break;
            case LFAP_ACTIVE:
                fprintf(stderr, "Active\n");
                break;
            default:
                fprintf(stderr, "NOT VALID\n");
                break;
        }
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapCommand_ (lfapui16_t type, lfapui32_t val)
{
    fprintf(stderr, "\tcommand:\t");
    if (type) {
        switch (val) {
            case LFAP_RETURN_INDICATED_FLOWS:
                fprintf(stderr, "indicated flows\n");
                break;
            case LFAP_RETURN_FLOW_PREFIX:
                fprintf(stderr, "flow prefix\n");
                break;
            case LFAP_CONNECTION_ACCEPTED:
                fprintf(stderr, "connection accepted\n");
                break;
            case LFAP_CONNECTION_REJECTED:
                fprintf(stderr, "connection rejected\n");
                break;
            case LFAP_LIST_OF_FASS:
                fprintf(stderr, "list of fas's\n");
                break;
            case LFAP_DISCONNECT:
                fprintf(stderr, "disconnect\n");
                break;
            case LFAP_KEEPALIVE:
                fprintf(stderr, "keepalive\n");
                break;
            case LFAP_RETURN_TIME:
                fprintf(stderr, "return time\n");
                break;
            case LFAP_REQUESTED_IEs:
                fprintf(stderr, "requested IEs\n");
                break;
            default:
                fprintf(stderr, "NOT VALID\n");
                break;
        }
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapFlowPrefix_ (lfapui16_t type, lfapui32_t fid0,
                                  lfapui32_t fid1)
{
    fprintf(stderr, "\tflow prefix:\t");
    if (type)
        fprintf(stderr, "%lu %lu\n", fid0, fid1);
    else
        fprintf(stderr, "None\n");
}

static void printLfapTime_ (lfapui16_t type, lfapui32_t val)
{
    fprintf(stderr, "\ttime:\t\t");
    if (type) {
        char *tmp = ctime((const long*)&val);
        fprintf(stderr, "(%lu)  ", val);
        fprintf(stderr, " %s", tmp ? tmp : "Time is before 1970\n");
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapCount_ (lfapui16_t type, LFAPDataCount* count)
{
    fprintf(stderr, "count:\t");
    if (type) {
        if (type == LFAP_PACKETS_TOTAL || type == LFAP_BYTES_TOTAL)
            fprintf(stderr, "total\n");
        else
            fprintf(stderr, "delta\n");
        fprintf(stderr, "\t\t\trcvd: %lu %lu  sent: %lu %lu\n",
	    count->dc_rcvd.hi_val, count->dc_rcvd.lo_val,
	    count->dc_sent.hi_val, count->dc_sent.lo_val);
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapProtocol_ (lfapui16_t type, lfapui8_t* data, lfapui16_t len)
{
    lfapui16_t i;
    lfapui16_t p_len;
    lfapui32_t port = 0;

    fprintf(stderr, "\tprotocol:\t");
    if (type) {
        fprintf(stderr, "\n\t  data -\t");
        p_len = data[0];
        fprintf(stderr, "%d", p_len);
        if (p_len == 16) /* 16 is the length for IP traffic, for now */ {
            for (i = 1; i < 13; i++)
                fprintf(stderr, ".%d", (unsigned int) data[i]);
            memcpy(&port, &data[13], 4);
            port = ntohl(port);
            if (port < 16777216) fprintf(stderr, ".0");
            if (port < 65535) fprintf(stderr, ".0");
            if (port < 256) fprintf(stderr, ".0");
            fprintf(stderr, ".%lu", port);
        }
        else {
            for (i = 1; i <= p_len; i++)
                fprintf(stderr, ".%d", (unsigned int) data[i]);
        }

        if (p_len == 16) {
            fprintf(stderr, "  (");
            /* add in more later */
            switch (data[7]) {
                case 8:
                    fprintf(stderr, "IP");
                    break;
                default:
                    break;
            }
            fprintf(stderr, ", ");
            switch (data[12]) {
                case 0:
                    fprintf(stderr, "None");
                    break;
                case 1:
                    fprintf(stderr, "ICMP");
                    break;
                case 2:
                    fprintf(stderr, "IGMP");
                    break;
                case 6:
                    fprintf(stderr, "TCP");
                    break;
                case 17:
                    fprintf(stderr, "UDP");
                    break;
                default:
                    fprintf(stderr, "Other");
                    break;
            }
            fprintf(stderr, ", ");
            switch (port) {
                case 0:
                    fprintf(stderr, "None)");
                    break;
                case 7:
                    fprintf(stderr, "echo)");
                    break;
                case 20:
                    fprintf(stderr, "ftp-data)");
                    break;
                case 21:
                    fprintf(stderr, "ftp-control)");
                    break;
                case 23:
                    fprintf(stderr, "telnet)");
                    break;
                case 25:
                    fprintf(stderr, "SMTP)");
                    break;
                case 53:
                    fprintf(stderr, "DNS)");
                    break;
                case 80:
                    fprintf(stderr, "http)");
                    break;
                case 111:
                    fprintf(stderr, "Sun RPC)");
                    break;
                case 137:
                    fprintf(stderr, "NETBIOS Name)");
                    break;
                case 138:
                    fprintf(stderr, "NETBIOS Datagram)");
                    break;
                case 139:
                    fprintf(stderr, "NETBIOS Session)");
                    break;
                case 161:
                    fprintf(stderr, "SNMP)");
                    break;
                case LFAP_PORT:
                    fprintf(stderr, "LFAP)");
                    break;
                case 99999:
                    fprintf(stderr, "other)");
                    break;
                default:
                    fprintf(stderr, "???)");
                    break;
            }
        }
        else if (data[7] == 0x88 && data[8] == 0x47 ){
          fprintf(stderr, " (MPLS)");
        }
        fprintf(stderr, "\n\t  length -\t%d\n", len);
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapClientData_ (lfapui16_t type, char* data, lfapui16_t len)
{
    fprintf(stderr, "\tclient data:\t");
    if (type) {
        fprintf(stderr, "\n\t  data -\t%s\n", data);
        fprintf(stderr, "\t  length -\t%d\n", len);
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapSrcAS_ (lfapui16_t type, lfapui32_t src_as)
{
    fprintf(stderr, "\tSrc AS\t\t");
    if (type) {
        fprintf(stderr, "%lu  ssr obtains value later\n", src_as); 
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapDestAS_ (lfapui16_t type, lfapui32_t dest_as)
{
    fprintf(stderr, "\tDest AS:\t\t");
    if (type) {
        fprintf(stderr, "%lu\n", dest_as);
    }
    else
        fprintf(stderr, "None\n");
}

static void printLfapPort_ (lfapui16_t type, lfapui32_t port)
{
    fprintf(stderr, "port:\t");
    if (type)
        fprintf(stderr, "%lu\n", port);
    else
        fprintf(stderr, "None\n");
}

static void printLfapMplsLabel_ (LFAPIE labels[])
{
  int i;
    fprintf(stderr, "MPLS Label:\t{");

    for(i = 0; i < LFAP_MAX_LABELS; i++) {
      if (labels[i].type)
        fprintf(stderr, "%u", labels[i].mplsLabel.val);
      else
        fprintf(stderr, "None");

      if ((i+1) < LFAP_MAX_LABELS)
        fprintf(stderr, ", ");
        
    }
    fprintf(stderr, "}\n");
}

static void printLfapFECs_ (LFAPIE fecs[], int send_or_recv)
{
  int i;
  
  struct in_addr IPaddr;

  if (!fecs[0].type) {
    fprintf(stderr, "\tFECs:\t\tNone\n");
    return;
  }

  for(i = 0; i < LFAP_MAX_FECs; i++) {
    if (fecs[i].type) {

      lfapui32_t tmp_addr;
      lfapui8_t fectype;

      if (send_or_recv == LFAP_SEND)
        fectype = fecs[i].octetStr_ie.val[0];
      else
        fectype = fecs[i].ldp.fec.fecType;

      switch (fectype) {
        
      case LFAP_LDP_PREFIX_FEC:
        {
          struct lfapLdpFECPrefix_s *fec;

          if (send_or_recv == LFAP_SEND)
            fec = (struct lfapLdpFECPrefix_s *)fecs[i].octetStr_ie.val;
          else
            fec = &fecs[i].ldp.fec.prefix;

          fprintf(stderr, "\tPrefix FEC:\t");
          fprintf(stderr, "\n\t  family -\t%d\n", fec->family);
          fprintf(stderr, "\t  prefix len -\t%u\n", fec->len);
          fprintf(stderr, "\t  addr -\t");
          tmp_addr = htonl(fec->addr);
          IPaddr = *(struct in_addr*)&tmp_addr;
          fprintf(stderr, "%s ", inet_ntoa(IPaddr));
        }
        break;
        
      case LFAP_LDP_HOST_FEC:
        {
          struct lfapLdpFECHost_s *fec;

          if (send_or_recv == LFAP_SEND)
            fec = (struct lfapLdpFECHost_s *)fecs[i].octetStr_ie.val;
          else
            fec = &fecs[i].ldp.fec.host;

          fprintf(stderr, "\tHost FEC:\t");
          fprintf(stderr, "\n\t  family -\t%d\n", fec->family);
          fprintf(stderr, "\t  addr len -\t%u\n", fec->len);
          fprintf(stderr, "\t  addr -\t");
          tmp_addr = htonl(fec->addr[0]);
          IPaddr = *(struct in_addr*)&tmp_addr;
          fprintf(stderr, "%s ", inet_ntoa(IPaddr));

          if ( fec->len >= 8 ) {
            IPaddr = *(struct in_addr*)&fec->addr[1];
            fprintf(stderr, " - %s ", inet_ntoa(IPaddr));
          }

          if ( fec->len >= 12 ) {
            IPaddr = *(struct in_addr*)&fec->addr[2];
            fprintf(stderr, " - %s ", inet_ntoa(IPaddr));
          }
        }
          break;
        
      case LFAP_LDP_VC_FEC:
        {
          struct lfapLdpFECVC_s *fec;
          char *vctypes[] = { "Unknown", "FRAME_RELAY", "ATM_AAL5", 
                              "ATM_TRANSPARENT", "ETHERNET_VLAN", 
                              "ETHERNET", "HDLC", "PPP", "CEM", "ATM_VCC", "ATM_VPC" };

          if (send_or_recv == LFAP_SEND)
            fec = (struct lfapLdpFECVC_s *)fecs[i].octetStr_ie.val;
          else
            fec = &fecs[i].ldp.fec.vc;

          fprintf(stderr, "\tVC FEC:");
          fprintf(stderr, "\n\t  VC type:  -\t%s\n", 
                  vctypes[fec->type < LFAP_MAX_LDP_FEC_TYPE ? 
                          fec->type : 0 ]);
          fprintf(stderr, "\t  Group ID -\t%lu\n", fec->groupID);
          if ( fec->len >= 4 ) {
            fprintf(stderr, "\t  VC ID -\t%lu\n", fec->VCID);
          }
          break;
        }
      default:
          fprintf(stderr, "\tUnknown FEC:\t%d", fectype);

      } /* switch */
    } /* if */
  } /* for */

  fprintf(stderr, "\n");
}

static void printLfapVlanID_ (lfapui16_t type, lfapui16_t s_id, lfapui16_t d_id)
{
    fprintf(stderr, "\tVLAN ID:\t");
    if (type)
        fprintf(stderr, "Source: %u\tDest: %u\n", s_id, d_id);
    else
        fprintf(stderr, "None\n");
}

static void printLfapCoS_ (LFAPIE CoS_Ar[][LFAP_COS_RT+1])
{
  lfapui8_t found = 0;
  lfapui8_t t, rt;
  lfapui8_t  ct;
  
  char *cos_types[] = {"Error", "IPv4", "IPv6", "MPLS", "VLAN" };
  
  for( t = 1; t <= LFAP_COS_MAX_TYPES; t++) {
    for( rt = 1; rt <= LFAP_COS_RT; rt++) {
      if ( CoS_Ar[t][rt].type ) {
        found = 1;
        fprintf(stderr, "\tCoS:\t\t");
        ct = CoS_Ar[t][rt].CoS.CoS_type;
        fprintf(stderr, "R = %u  T = %u  Type = %s   Value = %u\n",
                CoS_Ar[t][rt].CoS.RT & LFAP_COS_RECEIVED    ? 1 : 0,
                CoS_Ar[t][rt].CoS.RT & LFAP_COS_TRANSMITTED ? 1 : 0,
                cos_types[ ct >= 1 && ct <= LFAP_COS_MAX_TYPES ? ct : 0],
                CoS_Ar[t][rt].CoS.val);
      }
    }
  }
  
  if ( !found ) {
    fprintf(stderr, "\tCoS:\t");
    fprintf(stderr, "\tNone\n");
  }
}

static void printLfapUpTime_ (lfapui32_t time)
{
	/* assuming time is in seconds */

    lfapui32_t sec = 0;
    lfapui32_t min = 0;
    lfapui32_t hour = 0;
    lfapui32_t days = 0;
    lfapui32_t t = time;

    if (t >= 86400) {
        days = t / 86400;
        t = t%86400;
    }
        if (t >= 3600) {
        hour = t / 3600;
        t = t%3600;
    }
    if (t >= 60) {
        min = t / 60;
        t = t%60;
    }
    if (t >= 1) {
        sec = t;
    }

    printf("%lu %.2lu:%.2lu:%.2lu\n", days, hour, min, sec);
}


/******************************************************************************/
/*                                                                            */
/*  Function: printLFAPStats                                                  */
/*                                                                            */
/*  Prints out the all of the statistics the LFAP API.                        */
/*                                                                            */
/******************************************************************************/
void printLFAPStats (const LFAPApiStats* stats)
{
    lfapui32_t server;
    char buffer[256];
    lfapui16_t stat = 0;
    const LFAPApiStats* s;

    fprintf(stderr, "\n\n*******************\nLFAP API Statistics\n");
    fprintf(stderr, "*******************\n\n    Time:\t\t");
    printLfapCurrentTime_();
    fprintf(stderr, "\n");

    if (!stats) {
        fprintf(stderr, "    None Available\n\n");
        return;
    }
    else if (!stats->num_servers) {
        fprintf(stderr, "    NO Servers in Server List\n\n");
        return;
    }

    while (stat < stats->num_servers)
    {
        s = &stats[stat];

	server = s->this_server;
        sprintf(buffer, "%d.%d.%d.%d",
                (int) (server >> 24) & 0xff,
                (int) (server >> 16) & 0xff,
                (int) (server >>  8) & 0xff,
                (int) (server >>  0) & 0xff);
        fprintf(stderr, "    FAS:\t\t%s\n\n", buffer);

        fprintf(stderr, "\tconn_time:\t\t");
        fprintf(stderr, ctime((const long*)&s->conn_time));
        fprintf(stderr, "\tup_time:\t\t");
        printLfapUpTime_(s->up_time);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tnum_servers:\t\t%u\n", s->num_servers);
        fprintf(stderr, "\tthis_server:\t\t%lu\n", server);
        fprintf(stderr, "\tthis_server_in_list:\t%u\n",s->this_server_in_list);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tretry_conn_success:\t%lu\n", s->retry_conn_success);
        fprintf(stderr, "\tretry_conn_failure:\t%lu\n", s->retry_conn_failure);
        fprintf(stderr, "\n");
        fprintf(stderr, "\ttotal_bytes_sent:\t%lu\n", s->total_bytes_sent);
        fprintf(stderr, "\ttotal_bytes_rcvd:\t%lu\n", s->total_bytes_rcvd);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tmsgs_sent_success:\t%lu\n", s->msgs_sent_success);
        fprintf(stderr, "\tmsgs_sent_failure:\t%lu\n", s->msgs_sent_failure);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tmsgs_rcvd_corrupt:\t%lu\n", s->msgs_rcvd_corrupt);
        fprintf(stderr, "\tmsgs_rcvd_error:\t%lu\n", s->msgs_rcvd_error);
        fprintf(stderr, "\tmsgs_rcvd_success:\t%lu\n", s->msgs_rcvd_success);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tmsgs_send_buf:\t\t%lu\n", s->msgs_send_buf);
        fprintf(stderr, "\tmsgs_rcvd_buf:\t\t%lu\n", s->msgs_rcvd_buf);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tmsg_vr_sent:\t\t%lu\n", s->msg_vr_sent);
        fprintf(stderr, "\tmsg_vra_rcvd:\t\t%lu\n", s->msg_vra_rcvd);
        fprintf(stderr, "\tmsg_cr_sent:\t\t%lu\n", s->msg_cr_sent);
        fprintf(stderr, "\tmsg_can_rcvd:\t\t%lu\n", s->msg_can_rcvd);
        fprintf(stderr, "\tmsg_crn_rcvd:\t\t%lu\n", s->msg_crn_rcvd);
        fprintf(stderr, "\tmsg_dr_rcvd:\t\t%lu\n", s->msg_dr_rcvd);
        fprintf(stderr, "\tmsg_fer_rcvd:\t\t%lu\n", s->msg_fer_rcvd);
        fprintf(stderr, "\tmsg_far_sent:\t\t%lu\n", s->msg_far_sent);
        fprintf(stderr, "\tmsg_fun_sent:\t\t%lu\n", s->msg_fun_sent);
        fprintf(stderr, "\tmsg_ar_sent:\t\t%lu\n", s->msg_ar_sent);
        fprintf(stderr, "\tmsg_ar_rcvd:\t\t%lu\n", s->msg_ar_rcvd);
        fprintf(stderr, "\tmsg_ara_sent:\t\t%lu\n", s->msg_ara_sent);
        fprintf(stderr, "\tmsg_ara_rcvd:\t\t%lu\n", s->msg_ara_rcvd);
        fprintf(stderr, "\tmsg_ka_rcvd:\t\t%lu\n", s->msg_ka_rcvd);
        fprintf(stderr, "\tmsg_unk_rcvd:\t\t%lu\n", s->msg_unk_rcvd);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tdropped_send_msgs:\t%lu\n", s->dropped_send_msgs);
        fprintf(stderr, "\tdropped_send_msgs_conn:\t");
        fprintf(stderr, "%lu\n", s->dropped_send_msgs_conn);
        fprintf(stderr, "\tdropped_vr_msgs:\t%lu\n", s->dropped_vr_msgs);
        fprintf(stderr, "\tdropped_ar_msgs:\t%lu\n", s->dropped_ar_msgs);
        fprintf(stderr, "\tdropped_ara_msgs:\t%lu\n", s->dropped_ara_msgs);
        fprintf(stderr, "\tdropped_far_msgs:\t%lu\n", s->dropped_far_msgs);
        fprintf(stderr, "\tdropped_fun_i_msgs:\t%lu\n", s->dropped_fun_i_msgs);
        fprintf(stderr, "\tdropped_fun_msgs:\t%lu\n", s->dropped_fun_msgs);
        fprintf(stderr, "\tdropped_rcvd_msgs:\t%lu\n", s->dropped_rcvd_msgs);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tlost_bytes_sent:\t");
        fprintf(stderr, "%llu\n", s->lost_bytes_sent);
        fprintf(stderr, "\tlost_packets_sent:\t");
        fprintf(stderr, "%llu\n", s->lost_packets_sent);
        fprintf(stderr, "\tlost_bytes_rcvd:\t");
        fprintf(stderr, "%llu\n", s->lost_bytes_rcvd);
        fprintf(stderr, "\tlost_packets_rcvd:\t");
        fprintf(stderr, "%llu\n", s->lost_packets_rcvd);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tflow_setups:\t\t%lu\n", s->flow_setups);
        fprintf(stderr, "\tflow_teardowns:\t\t%lu\n", s->flow_teardowns);
        fprintf(stderr, "\tactive_flows:\t\t%lu\n", s->active_flows);
        fprintf(stderr, "\n");
        fprintf(stderr, "\tpeak_msgs_send_buf:\t%lu\n",s->peak_msgs_send_buf);
        fprintf(stderr, "\tpeak_active_flows:\t%lu\n", s->peak_active_flows);
        fprintf(stderr, "\n\n");
        stat++;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: printLfapList                                                   */
/*                                                                            */
/*  Prints out the members of the LFAP list.                                  */
/*                                                                            */
/******************************************************************************/
void printLfapList (LFAPList* list)
{
    void* member;
    lfapui16_t entries;
    LFAPListIterator iter;

    entries = lfapListEntries(list);
    switch (entries) {
        case 0:
            fprintf(stderr, "   The List is Empty\n");
            break;
        case 1:
            fprintf(stderr, "   The List has 1 entry\n");
            break;
        default:
            fprintf(stderr,"   The List has %d entries\n", entries);
            break;
    }
	
    if (entries) {
        fprintf(stderr, "    The List Members:\n");
        initLfapIterList(&iter, list);
        while ((member = incLfapIter(&iter)))
            fprintf(stderr, "      %lu\n", (lfapui32_t) member);
    }

    fprintf(stderr, "\n");
}

/******************************************************************************/
/*                                                                            */
/*  Function: PRINT_LFAP_MESSAGE                                              */
/*                                                                            */
/*  Prints out the message structure (header and all internal IEs).  If addr  */
/*  is not 0, the address is also printed out.  send_or_recv is either        */
/*  LFAP_SEND, or LFAP_RECV, indicating if this message is being sent or was  */
/*  received.                                                                 */
/*                                                                            */
/******************************************************************************/
void PRINT_LFAP_MESSAGE (void* msg, const char* addr, int send_or_recv)
{
    LFAP_AR* ar;
    LFAP_CR* cr;
    LFAP_KA* ka;
    LFAP_VR* vr;
    LFAP_DR* dr;
    LFAP_FER* fer;
    LFAP_ARA* ara;
    LFAP_CAN* can;
    LFAP_CRN* crn;
    LFAP_VRA* vra;
    LFAP_FUN* fun;
    LFAPHeader* hdr;
    LFAPBuffer* buf;
    LFAPMsg* lfap_msg;
    lfapui8_t type = 0;
    LFAP_FAR* lfap_far;
    lfapui16_t valid_msg = 1;

    lfap_msg = (LFAPMsg*) msg;
    buf = &lfap_msg->buf;
    hdr = &lfap_msg->hdr;
    type = getLfapMsgOpcode(lfap_msg);

    current_msg_send_or_recv = send_or_recv;
    if (send_or_recv == LFAP_SEND) fprintf(stderr, "**** SENDING ");
    else if (send_or_recv == LFAP_RECV) fprintf(stderr, "!!!! RECEIVED ");
    else fprintf(stderr, "???? ");
    switch (type) {
        case LFAP_VR_OPCODE:
            fprintf(stderr, "VR"); break;
        case LFAP_VRA_OPCODE:
            fprintf(stderr, "VRA"); break;
        case LFAP_AR_OPCODE:
            fprintf(stderr, "AR"); break;
        case LFAP_CR_OPCODE:
            fprintf(stderr, "CR"); break;
        case LFAP_DR_OPCODE:
            fprintf(stderr, "DR"); break;
        case LFAP_CAN_OPCODE:
            fprintf(stderr, "CAN"); break;
        case LFAP_CRN_OPCODE:
            fprintf(stderr, "CRN"); break;
        case LFAP_FER_OPCODE:
            fprintf(stderr, "FER"); break;
        case LFAP_KA_OPCODE:
            fprintf(stderr, "KA"); break;
        case LFAP_ARA_OPCODE:
            fprintf(stderr, "ARA"); break;
        case LFAP_FAR_OPCODE:
            fprintf(stderr, "FAR"); break;
        case LFAP_FUN_OPCODE:
            fprintf(stderr, "FUN"); break;
        default:
            valid_msg = 0;
            fprintf(stderr, "INVALID MESSAGE"); break;
   }
    if (send_or_recv == LFAP_SEND) {
        if (addr) fprintf(stderr, " to %s", addr);
        fprintf(stderr, " ****\n\n    Time sent:\t\t");
    }
    else if (send_or_recv == LFAP_RECV) {
        if (addr) fprintf(stderr, " from %s", addr);
        fprintf(stderr, " !!!!\n\n    Time received:\t");
    }
    else {
        if (addr) fprintf(stderr, " ???? %s", addr);
        fprintf(stderr, " ????\n\n    Time:\t\t");
    }
    printLfapCurrentTime_();

    if (valid_msg) fprintf(stderr, "\n    Header\n\n");
    switch (type) {
        case LFAP_VR_OPCODE:
            vr = (LFAP_VR*) msg;
            fprintf(stderr, "\ttype:\t\tVR");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            fprintf(stderr, "\tEmpty is valid\n");
            break;
        case LFAP_VRA_OPCODE:
            vra = (LFAP_VRA*) msg;
            fprintf(stderr, "\ttype:\t\tVRA");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            fprintf(stderr, "\tEmpty is valid\n");
            break;
        case LFAP_CAN_OPCODE:
            can = (LFAP_CAN*) msg;
            fprintf(stderr, "\ttype:\t\tCAN");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            fprintf(stderr, "\tEmpty is valid\n");
            break;
        case LFAP_CRN_OPCODE:
            crn = (LFAP_CRN*) msg;
            fprintf(stderr, "\ttype:\t\tCRN");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            fprintf(stderr, "\tEmpty is valid\n");
            break;
        case LFAP_FER_OPCODE:
            fer = (LFAP_FER*) msg;
            fprintf(stderr, "\ttype:\t\tFER");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            fprintf(stderr, "\tEmpty is valid\n");
            break;
        case LFAP_AR_OPCODE:
            ar = (LFAP_AR*) msg;
            fprintf(stderr, "\ttype:\t\tAR");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            printLfapCommand_(ar->command.type, ar->command.rsvd_ui8.val);
            fprintf(stderr, "\t");
    printLfapFlowId_(ar->flow_id.fid_ie.type,
		     ar->flow_id.fid_ie.cce_id,
		     ar->flow_id.fid_ie.fas_id[0],
		     ar->flow_id.fid_ie.fas_id[1]);
            fprintf(stderr, "\tfas ");
    printLfapAddr_(ar->fas_ip_addr.type, &ar->fas_ip_addr.addr_ie);
            break;
        case LFAP_ARA_OPCODE:
            ara = (LFAP_ARA*) msg;
            fprintf(stderr, "\ttype:\t\tARA");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n\t");
            printLfapFlowId_(ara->flow_id.fid_ie.type,
		     ara->flow_id.fid_ie.cce_id,
		     ara->flow_id.fid_ie.fas_id[0],
		     ara->flow_id.fid_ie.fas_id[1]);
            printLfapFailCode_(ara->fail_code.type, ara->fail_code.rsvd_ui8.val);
            printLfapFlowPrefix_(ara->flow_prefix.pfx.type,
                                 ara->flow_prefix.pfx.fas_id[0],
                                 ara->flow_prefix.pfx.fas_id[1]);
            printLfapTime_(ara->time.i32_ie.type, ara->time.i32_ie.val);
            break;
        case LFAP_FAR_OPCODE:
            lfap_far = (LFAP_FAR*) msg;
            fprintf(stderr, "\ttype:\t\tFAR");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n\t");
    printLfapFlowId_(lfap_far->flow_id.fid_ie.type,
		     lfap_far->flow_id.fid_ie.cce_id,
		     lfap_far->flow_id.fid_ie.fas_id[0],
		     lfap_far->flow_id.fid_ie.fas_id[1]);
            fprintf(stderr, "\tsrc ");
            printLfapAddr_(lfap_far->src_addr.type, &lfap_far->src_addr.addr_ie);
            fprintf(stderr, "\tdest ");
            printLfapAddr_(lfap_far->dest_addr.type, &lfap_far->dest_addr.addr_ie);
            printLfapSrcPort_(lfap_far->src_port.type,
		      lfap_far->src_port.i32_ie.val);
            printLfapProtocol_(lfap_far->protocol.type,
		       (lfapui8_t*)lfap_far->protocol.octetStr_ie.val,
		       lfap_far->protocol.octetStr_ie.len);
            printLfapClientData_(lfap_far->client_data.type,
			 lfap_far->client_data.octetStr_ie.val,
			 lfap_far->client_data.octetStr_ie.len);
            fprintf(stderr, "\tsrc cce ");
            printLfapAddr_(lfap_far->src_cce_addr.type,
		   &lfap_far->src_cce_addr.addr_ie);
            printLfapQualifier_(lfap_far->qualifier.type,
                               &lfap_far->qualifier.qual_ie);
            printLfapSrcAS_ (lfap_far->src_AS.type, lfap_far->src_AS.i32_ie.val );
            printLfapDestAS_(lfap_far->dest_AS.type, lfap_far->dest_AS.i32_ie.val);
            fprintf(stderr, "\tingress ");
            printLfapPort_(lfap_far->ingress_port.type,
		   lfap_far->ingress_port.i32_ie.val);
            fprintf(stderr, "\tegress ");
            printLfapPort_(lfap_far->egress_port.type,
                           lfap_far->egress_port.i32_ie.val);
            printLfapVlanID_(lfap_far->vlan_id.type,
                             lfap_far->vlan_id.i16_ie.val[0],
                             lfap_far->vlan_id.i16_ie.val[1]  );
            printLfapCoS_(lfap_far->CoS_Arr);

            fprintf(stderr, "\tIn ");
            printLfapMplsLabel_(lfap_far->inMplsLabel);
            fprintf(stderr, "\tOut ");
            printLfapMplsLabel_(lfap_far->outMplsLabel);

            printLfapFECs_(lfap_far->ldpFEC, send_or_recv);

            fprintf(stderr, "\tnext hop ");
            printLfapAddr_(lfap_far->next_hop_addr.type,
                          &lfap_far->next_hop_addr.addr_ie);
            printLfapTime_(lfap_far->time.i32_ie.type, lfap_far->time.i32_ie.val);
            break;
        case LFAP_FUN_OPCODE:
            fun = (LFAP_FUN*) msg;
            fprintf(stderr, "\ttype:\t\tFUN");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n\t");
    printLfapFlowId_(fun->flow_id.fid_ie.type,
		     fun->flow_id.fid_ie.cce_id,
		     fun->flow_id.fid_ie.fas_id[0],
		     fun->flow_id.fid_ie.fas_id[1]);
    printLfapFlowState_(fun->flow_state.type, fun->flow_state.rsvd_ui8.val);
    printLfapTime_(fun->time.i32_ie.type, fun->time.i32_ie.val);
            fprintf(stderr, "\tbyte ");
    printLfapCount_(fun->byte_count.dataCount_ie.type, &fun->byte_count.dataCount_ie);
            fprintf(stderr, "\tpacket ");
    printLfapCount_(fun->packet_count.dataCount_ie.type, &fun->packet_count.dataCount_ie);
            break;
        case LFAP_CR_OPCODE:
            cr = (LFAP_CR*) msg;
            fprintf(stderr, "\ttype:\t\tCR");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            fprintf(stderr, "\tfas ");
            printLfapAddr_(cr->fas_ip_addr.type, &cr->fas_ip_addr.addr_ie);
            break;
        case LFAP_DR_OPCODE:
            dr = (LFAP_DR*) msg;
            fprintf(stderr, "\ttype:\t\tDR");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            fprintf(stderr, "\tfas ");
            printLfapAddr_(dr->fas_ip_addr.type, &dr->fas_ip_addr.addr_ie);
            break;
        case LFAP_KA_OPCODE:
            ka = (LFAP_KA*) msg;
            fprintf(stderr, "\ttype:\t\tKA");
            printLfapHeader_(hdr, buf);
            fprintf(stderr, "    Message body\n\n");
            fprintf(stderr, "\tEmpty is valid\n");
            break;
        default:
            break;
	}

	fprintf(stderr, "\n\n");
}

void print_lfap_msg( char *buff, lfapui16_t bufsize, lfapui32_t ip, int direction)
{
    LFAP_VR  vr;
    LFAP_VRA vra;
    LFAP_CR  cr;
    LFAP_CAN can;
    LFAP_CRN crn;
    LFAP_FER fer;
    LFAP_AR  ar;
    LFAP_ARA ara;
    LFAP_FUN fun;
    LFAP_FAR lfap_far;
    LFAP_KA  ka;
    LFAP_DR  dr;
    char ip_str[16];

    sprintf(ip_str, "%d.%d.%d.%d",
                (int) (ip >> 24) & 0xff,
                (int) (ip >> 16) & 0xff,
                (int) (ip >>  8) & 0xff,
                (int) (ip >>  0) & 0xff);

    switch( lfap_msg_get_opcode(buff) ) {

        case LFAP_VR_OPCODE :
          initRcvdLfapVR( &vr, buff, bufsize );
          PRINT_LFAP_MESSAGE( &vr, ip_str, direction );
          break;
        case LFAP_VRA_OPCODE :
          initRcvdLfapVRA( &vra, buff, bufsize );
          PRINT_LFAP_MESSAGE( &vra, ip_str, direction );
          break;

        case LFAP_CR_OPCODE :
          initRcvdLfapCR( &cr, buff, bufsize );
          do {
            PRINT_LFAP_MESSAGE( &cr, ip_str, direction );
          } while ( decodeNextRecordLfapCR( &cr ) != LFAP_EMPTY_BUFFER );
          break;

        case LFAP_CAN_OPCODE :
          initRcvdLfapCAN( &can, buff, bufsize );
          PRINT_LFAP_MESSAGE( &can, ip_str, direction );
          break;

        case LFAP_CRN_OPCODE :
          initRcvdLfapCRN( &crn, buff, bufsize );
          PRINT_LFAP_MESSAGE( &crn, ip_str, direction );
          break;

        case LFAP_FER_OPCODE :
          initRcvdLfapFER( &fer, buff, bufsize );
          PRINT_LFAP_MESSAGE( &fer, ip_str, direction );
          break;

        case LFAP_FAR_OPCODE :
          initRcvdLfapFAR( &lfap_far, buff, bufsize );
          PRINT_LFAP_MESSAGE( &lfap_far, ip_str, direction );
          break;
        case LFAP_FUN_OPCODE :
          initRcvdLfapFUN( &fun, buff, bufsize );
          do {
            PRINT_LFAP_MESSAGE( &fun, ip_str, direction );
          } while ( decodeNextRecordLfapFUN( &fun ) != LFAP_EMPTY_BUFFER );
          break;
        case LFAP_AR_OPCODE :
          initRcvdLfapAR( &ar, buff, bufsize );
          do {
            PRINT_LFAP_MESSAGE( &ar, ip_str, direction );
          } while ( decodeNextRecordLfapAR( &ar ) != LFAP_EMPTY_BUFFER );
          break;
        case LFAP_ARA_OPCODE :
          initRcvdLfapARA( &ara, buff, bufsize );
          PRINT_LFAP_MESSAGE( &ara, ip_str, direction );
          break;

        case LFAP_KA_OPCODE :
          initRcvdLfapKA( &ka, buff, bufsize );
          PRINT_LFAP_MESSAGE( &ka, ip_str, direction );
          break;

        case LFAP_DR_OPCODE :
          initRcvdLfapDR( &dr, buff, bufsize );
          PRINT_LFAP_MESSAGE( &dr, ip_str, direction );
          break;

        default :
          fprintf(stderr, "Cannot print message, unknown opcode = %d\n", lfap_msg_get_opcode(buff));
    }
}
