/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/*
 * lfap message decoding routines
 */

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_types.h"
#endif
/* End of auto-generated includes */

#include "Lfap.h"
#include "LfapPrivate.h"

/*
 * macros for decoding integers from network to host order
 */
#define DECODE8(src,dst)  { (dst)  = (lfapui8_t) *(src)++; }
#define DECODE16(src,dst) { (dst)  = (lfapui16_t)*(src)++ <<  8; \
                            (dst) |= (lfapui16_t)*(src)++; }
#define DECODE32(src,dst) { (dst) =  (lfapui32_t)*(src)++ << 24; \
                            (dst) |= (lfapui32_t)*(src)++ << 16; \
                            (dst) |= (lfapui32_t)*(src)++ <<  8; \
                            (dst) |= (lfapui32_t)*(src)++; }


/*
 *  _decode_unknown
 *  Handles an unknown IE type encountered during decoding.
 */
static int _decode_unknown(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  /* All we can do is move the pointer past the IE */
  if ((ptr = lfap_buffer_extract(lbuf,len+PADLEN(len))) == NULL)
	return LFAP_ERROR;

  return LFAP_SUCCESS;
}


/*
 *  _decode_count
 *  Decodes the body of an LFAPDataCount IE
 */
static int _decode_count(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  if (len != 16) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,16)) == NULL)
	return LFAP_ERROR;

  DECODE32(ptr,ie->dataCount_ie.dc_rcvd.hi_val);
  DECODE32(ptr,ie->dataCount_ie.dc_rcvd.lo_val);
  DECODE32(ptr,ie->dataCount_ie.dc_sent.hi_val);
  DECODE32(ptr,ie->dataCount_ie.dc_sent.lo_val);

  return LFAP_SUCCESS;
}


/*
 *  _decode_flowid
 *  Decodes the body of an LFAPFlowID IE
 */
static int _decode_flowid(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;
  lfapui16_t fas_len, cce_len;

  fas_len = len >> 8;
  cce_len = len & 0xff;

  if ((fas_len != 0 && fas_len != 8) || (cce_len != 4)) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,fas_len + cce_len)) == NULL)
	return LFAP_ERROR;

  if (fas_len) {
    DECODE32(ptr,ie->fid_ie.fas_id[0]);
    DECODE32(ptr,ie->fid_ie.fas_id[1]);
  }
  else {
    ie->fid_ie.fas_id[0] = ie->fid_ie.fas_id[1] = 0;
  }

  DECODE32(ptr,ie->fid_ie.cce_id);

  return LFAP_SUCCESS;
}


/*
 *  _decode_octstr
 *  Decodes the body of an LFAPOctetStr IE
 */
static int _decode_octstr(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  if ((ptr = lfap_buffer_extract(lbuf,len+PADLEN(len))) == NULL)
	return LFAP_ERROR;

  ie->octetStr_ie.len = len;
  ie->octetStr_ie.val = (char*)ptr;

  return LFAP_SUCCESS;
}


/*
 *  _decode_addr
 *  Decodes the body of an LFAPAddress IE
 */
static int _decode_addr(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  if (len < 4) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,len+PADLEN(len))) == NULL)
	return LFAP_ERROR;

  DECODE16(ptr,ie->addr_ie.family);
  DECODE16(ptr,ie->addr_ie.len);

  if (ie->addr_ie.len != len - 4) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  ie->addr_ie.val = (char *)ptr;

  return LFAP_SUCCESS;
}


/*
 *  _decode_uint32
 *  Decodes the body of an LFAPUInt32 IE
 */
static int _decode_uint32(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  if (len != 4) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,4)) == NULL)
	return LFAP_ERROR;

  DECODE32(ptr,ie->i32_ie.val);

  return LFAP_SUCCESS;
}


/*
 *  _decode_uint16x2
 *  Decodes the body of an LFAPUInt16x2 IE
 */
static int _decode_uint16x2(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  if (len != 4) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,4)) == NULL)
	return LFAP_ERROR;

  DECODE16(ptr,ie->i16_ie.val[0]);
  DECODE16(ptr,ie->i16_ie.val[1]);

  return LFAP_SUCCESS;
}


/*
 *  _decode_rsvd_uint8
 *  decodes the body of an lfaprsvduint8 ie
 */
static int _decode_rsvd_uint8(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;
  lfapui8_t *rptr;

  if (len != 4) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,4)) == NULL)
	return LFAP_ERROR;

  rptr = (lfapui8_t *) &((LFAP_encode_decode_RsvdUint8 *)&ie->rsvd_ui8)->reserved_val;
  DECODE8(ptr,*rptr); 
  rptr++; 
  DECODE8(ptr,*rptr); 
  rptr++;
  DECODE8(ptr,*rptr);
  DECODE8(ptr,ie->rsvd_ui8.val);

  return LFAP_SUCCESS;
}

/*
 *  _decode_rsvd_uint16
 *  decodes the body of an lfaprsvduint16 ie
 */
static int _decode_rsvd_uint16(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  if (len != 4) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,4)) == NULL)
	return LFAP_ERROR;

  DECODE16(ptr,ie->rsvd_ui16.reserved);
  DECODE16(ptr,ie->rsvd_ui16.val);

  return LFAP_SUCCESS;
}

/*
 *  _decode_rsvd_uint24
 *  decodes the body of an lfaprsvduint24 ie
 */
static int _decode_rsvd_uint24(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;
  lfapui8_t *vptr;

  if (len != 4) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,4)) == NULL)
	return LFAP_ERROR;

  vptr = (lfapui8_t *) &((LFAP_encode_decode_RsvdUint24 *)&ie->rsvd_ui24)->reserved_val;
  vptr++; /* skip over the reserved byte */
  DECODE8(ptr,ie->rsvd_ui24.reserved);
  DECODE8(ptr,*vptr);
  vptr++;
  DECODE8(ptr,*vptr);
  vptr++;
  DECODE8(ptr,*vptr);

  return LFAP_SUCCESS;
}

/*
 *  _decode_ldpFEC
 *  Decodes the body of an LFAPLDPFEC IE
 */
static int _decode_ldpfec(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  if (len < 1) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,1)) == NULL)
	return LFAP_ERROR;

  DECODE8(ptr, ie->ldp.fec.fecType);
  switch (ie->ldp.fec.fecType) {

  case LFAP_LDP_PREFIX_FEC:
    if (len != 8 || (ptr = lfap_buffer_extract(lbuf,7)) == NULL)
      return LFAP_ERROR;
    DECODE16(ptr, ie->ldp.fec.prefix.family);
    DECODE8(ptr,  ie->ldp.fec.prefix.len);
    DECODE32(ptr, ie->ldp.fec.prefix.addr);
    break;

  case LFAP_LDP_HOST_FEC:
    if ((ptr = lfap_buffer_extract(lbuf,3)) == NULL)
      return LFAP_ERROR;
    DECODE16(ptr, ie->ldp.fec.host.family);
    DECODE8(ptr, ie->ldp.fec.host.len);
    if ( (ptr = lfap_buffer_extract(lbuf,ie->ldp.fec.host.len)) == NULL)
      return LFAP_ERROR;
 
    if ((ie->ldp.fec.host.len + 4) != len)
      return LFAP_ERROR;

    if ( ie->ldp.fec.host.len == 4 ){
      DECODE32(ptr, ie->ldp.fec.host.addr[0]);
    }
    else if (ie->ldp.fec.host.len == 8) {
      DECODE32(ptr, ie->ldp.fec.host.addr[0]);
      DECODE32(ptr, ie->ldp.fec.host.addr[1]);
    }
    else if (ie->ldp.fec.host.len == 12 ) {
      DECODE32(ptr, ie->ldp.fec.host.addr[0]);
      DECODE32(ptr, ie->ldp.fec.host.addr[1]);
      DECODE32(ptr, ie->ldp.fec.host.addr[2]);
    }
    else
      return LFAP_ERROR;
    break;

  case LFAP_LDP_VC_FEC:
    if ((ptr = lfap_buffer_extract(lbuf,7)) == NULL)
      return LFAP_ERROR;
    DECODE16(ptr, ie->ldp.fec.vc.type);
    DECODE8(ptr, ie->ldp.fec.vc.len);
    DECODE32(ptr, ie->ldp.fec.vc.groupID);

    if ((ptr = lfap_buffer_extract(lbuf,ie->ldp.fec.vc.len+PADLEN(len))) == NULL)
      return LFAP_ERROR;    

    if ( ie->ldp.fec.vc.len >= 4 ) {
      DECODE32(ptr, ie->ldp.fec.vc.VCID);
    }
    if ( (ie->ldp.fec.vc.len + PADLEN(len) + 8) != len )
      return LFAP_ERROR;
    break;

  default:
    /* Unknown FEC types are treated as octet strings          */
    /* subtract the first byte we already retrieved as fecType */
    if ((ptr = lfap_buffer_extract(lbuf,len+PADLEN(len) - 1)) == NULL)
      return LFAP_ERROR;

    ie->ldp.fec.unknown.buf = (char*)--ptr; /* back up to the fectype */
  }

  return LFAP_SUCCESS;
}

/*
 *  _decode_CoS
 *  decodes the body of a Class of Service ie
 */
static int _decode_CoS(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;
  
  if (len != 4) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,4)) == NULL)
	return LFAP_ERROR;

  /* High order byte */
  ie->CoS.RT   = *ptr >> 6;
  ie->CoS.CoS_type = *ptr & 0x3F;

  /* lower 3 bytes which is the value */
  ie->CoS.val =  ((lfapui32_t)*(++ptr) << 16);
  ie->CoS.val |= ((lfapui32_t)*(++ptr) << 8);
  ie->CoS.val |= *(++ptr);

  return LFAP_SUCCESS;
}


/*
 *  _decode_qualifier
 *  Decodes the body of an LFAPQualifierList IE
 */
static int _decode_qualifier(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr, *end;
  LFAPQualifier *qual;

  if ((ptr = lfap_buffer_extract(lbuf,len+PADLEN(len))) == NULL)
	return LFAP_ERROR;

  ie->qual_ie.items = 0;

  end = ptr + len;

  while (ptr < end) {
	if (end - ptr < 4) {
		lbuf->err = LFAP_MSG_CORRUPTED;
		return LFAP_ERROR;
	}

    qual = ie->qual_ie.q_list + ie->qual_ie.items;

    DECODE16(ptr,qual->id);
    DECODE16(ptr,qual->len);

    len = qual->len + PADLEN(qual->len);

	if (end - ptr < len) {
		lbuf->err = LFAP_MSG_CORRUPTED;
		return LFAP_ERROR;
	}

    qual->val = (char *)ptr;
	ptr += len;

    ie->qual_ie.items++;
  }

  return LFAP_SUCCESS;
}


/*
 *  _decode_prefix
 *  Decodes the body of an LFAPFlowIDPrefix IE
 */
static int _decode_prefix(LFAPBuffer *lbuf, LFAPIE *ie, lfapui16_t len)
{
  lfapui8_t *ptr;

  if (len != 8) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  if ((ptr = lfap_buffer_extract(lbuf,8)) == NULL)
	return LFAP_ERROR;

  DECODE32(ptr,ie->pfx.fas_id[0]);
  DECODE32(ptr,ie->pfx.fas_id[1]);

  return LFAP_SUCCESS;
}


/* prototype for the '_decode' functions */
typedef int (*vptr)(LFAPBuffer*, LFAPIE*, lfapui16_t len);

vptr _decode_vtbl[LFAP_MAX_IE_CODE];

void lfap_init_decode() {
  int ie;

  /* Initialize array to all unknown */
  for(ie = 0; ie < LFAP_MAX_IE_CODE; ie++) {
    _decode_vtbl[ie] = _decode_unknown;
  }

  /* Fill in the known IE decode routines */
  _decode_vtbl[LFAP_TRANSPORT_FAS_IP]          = _decode_addr;
  _decode_vtbl[LFAP_TRANSPORT_MULTIPLE_RECORD] = _decode_unknown;
  _decode_vtbl[LFAP_FLOW_ID]                   = _decode_flowid;
  _decode_vtbl[LFAP_SRC_ADDR]                  = _decode_addr;
  _decode_vtbl[LFAP_DST_ADDR]                  = _decode_addr;
  _decode_vtbl[LFAP_FLOW_QUALIFIER]            = _decode_qualifier;
  _decode_vtbl[LFAP_TIME]                      = _decode_uint32;
  _decode_vtbl[LFAP_UTC_TIME]                  = _decode_uint32;
  _decode_vtbl[LFAP_DELTA_TIME]                = _decode_unknown; /* Not yet implemented */
  _decode_vtbl[LFAP_CLASS_OF_SERVICE]          = _decode_CoS;
  _decode_vtbl[LFAP_SRC_CCE_ADDR]              = _decode_addr;
  _decode_vtbl[LFAP_CLIENT_DATA]               = _decode_octstr;
  _decode_vtbl[LFAP_COMMAND_CODE]              = _decode_rsvd_uint8;
  _decode_vtbl[LFAP_IE_LIST]                   = _decode_unknown; /* Not yet implemented */
  _decode_vtbl[LFAP_FAS_IP_ADDR]               = _decode_addr;
  _decode_vtbl[LFAP_FAIL_CODE]                 = _decode_rsvd_uint8;
  _decode_vtbl[LFAP_FLOW_ID_PREFIX]            = _decode_prefix;
  _decode_vtbl[LFAP_FLOW_STATE]                = _decode_rsvd_uint8;
  _decode_vtbl[LFAP_BYTES_TOTAL]               = _decode_count;
  _decode_vtbl[LFAP_BYTES_DELTA]               = _decode_count;
  _decode_vtbl[LFAP_PACKETS_TOTAL]             = _decode_count;
  _decode_vtbl[LFAP_PACKETS_DELTA]             = _decode_count;
  _decode_vtbl[LFAP_PROTOCOL_ID]               = _decode_octstr;
  _decode_vtbl[LFAP_SRC_PORT_UDP]              = _decode_uint32;
  _decode_vtbl[LFAP_SRC_PORT_TCP]              = _decode_uint32;
  _decode_vtbl[LFAP_SRC_SOCKET_IPX]            = _decode_uint32;
  _decode_vtbl[LFAP_SRC_AS]                    = _decode_uint32;
  _decode_vtbl[LFAP_DEST_AS]                   = _decode_uint32;
  _decode_vtbl[LFAP_INGRESS_PORT]              = _decode_uint32;
  _decode_vtbl[LFAP_EGRESS_PORT]               = _decode_uint32;
  _decode_vtbl[LFAP_VLAN_ID]                   = _decode_uint16x2;
  _decode_vtbl[LFAP_MPLS_IN_LABEL]             = _decode_uint32;
  _decode_vtbl[LFAP_MPLS_OUT_LABEL]            = _decode_uint32;
  _decode_vtbl[LFAP_NEXT_HOP_ADDR]             = _decode_addr;
  _decode_vtbl[LFAP_LDP_FEC]                   = _decode_ldpfec;
  /* The rest of the IE's are not yet implemented */
}

/*
 *  _decode_ie_hdr
 *  Decodes the type/length header of an LFAP IE
 */
static int _decode_ie_hdr(LFAPBuffer *lbuf, LFAPIE *ie)
{
  lfapui8_t *ptr;
  lfapui16_t ie_len;

  if (!lbuf->form.off) {

	/*
	 * We're not inside a formatted record, so read the IE header
	 * from the end of the message.
	 */
	if ((ptr = lfap_buffer_extract(lbuf, 4)) == NULL)
		return LFAP_ERROR;
  }
  else {
	/*
	 * This is a formatted record, we read the IE header from the
         * format specifier.
	 */

	if (lbuf->form.off >= lbuf->form.end) {
		lbuf->err = LFAP_MSG_CORRUPTED;
		return LFAP_ERROR;
	}

	ptr = lbuf->buf + lbuf->form.off;

	lbuf->form.off += 4;
  }

  DECODE16(ptr,ie->type);
  DECODE16(ptr,ie_len);

  return (int)ie_len;
}


/*
 *  _decode_mult_hdr
 *  Initializes the cursors necessary for decoding a multiple-record IE
 */
static int _decode_mult_hdr(LFAPBuffer *lbuf, lfapui16_t ie_len)
{
  lfapui8_t *ptr;
  lfapui16_t fixedlen, formlen;

  lbuf->mlt.end = lbuf->off + ie_len;

  if ((ptr = lfap_buffer_extract(lbuf, 4)) == 0)
	return LFAP_ERROR;

  DECODE16(ptr,fixedlen);
  DECODE16(ptr,formlen);

  lbuf->form.beg = lbuf->off + fixedlen;
  lbuf->form.end = lbuf->form.beg + formlen;
  lbuf->form.off = 0;

  lbuf->mlt.beg = lbuf->off;
  lbuf->mlt.off = lbuf->form.end;

  return LFAP_SUCCESS;
}


/*
 *  decode just the message length from an lfap message header
 */
int lfap_decode_msglen(void *buf)
{
  lfapui8_t* ptr;
  lfapui16_t msglen;

  ptr = LFAP_LENGTH_OFFSET + (lfapui8_t*)buf;

  DECODE16(ptr,msglen);

  return msglen;
}


/*
 *  decode the header of an LFAPMessage
 */
int lfap_get_header(LFAPBuffer *lbuf, LFAPHeader *hdr)
{
  lfapui8_t *ptr, version, reserved;
  lfapui16_t msglen;

  if ((ptr = lfap_buffer_extract(lbuf, LFAP_HEADER_SIZE)) == NULL)
	return LFAP_ERROR;

  DECODE8(ptr,version);
  if (version != LFAP_VERSION) {
	lbuf->err = LFAP_MSG_VERSION;
	return LFAP_ERROR;
  }

  DECODE8(ptr,hdr->opcode);

  DECODE8(ptr,reserved);
  if (reserved != 0) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  DECODE8(ptr,hdr->status);
  DECODE16(ptr,hdr->msgid);

  DECODE16(ptr,msglen);
  if (msglen != lbuf->msglen - lbuf->off || (msglen & 0x3)) {
	lbuf->err = LFAP_MSG_CORRUPTED;
	return LFAP_ERROR;
  }

  return LFAP_SUCCESS;
}


/*
 *  Decodes the next available IE into the passed in LFAPIE struct.
 *
 */
int lfap_get_IE(LFAPBuffer *lbuf, LFAPIE *ie)
{
  int first, len;

  if (lbuf->off == lbuf->msglen ||
      lbuf->off == lbuf->mlt.end ||
      (lbuf->form.off && lbuf->form.off >= lbuf->form.end)) {
	lbuf->err = LFAP_NODATA;
	return LFAP_ERROR;
  }

  if (lbuf->off == lbuf->form.beg) {
	lbuf->form.off = lbuf->form.beg;
	lbuf->off = lbuf->mlt.off;
  }

  if ((len = _decode_ie_hdr(lbuf, ie)) < 0)
	return LFAP_ERROR;

  if (ie->type == LFAP_TRANSPORT_MULTIPLE_RECORD) {
	first = (lbuf->off == LFAP_HEADER_SIZE + 4);

	if (_decode_mult_hdr(lbuf, (lfapui16_t)len) == LFAP_SUCCESS) {
		if (first) {
			return lfap_get_IE(lbuf, ie);
		}
		lbuf->err = LFAP_NODATA;
	}
	return LFAP_ERROR;
  }

  return CALL_VIRTUAL(ie,_decode)(lbuf, ie, (lfapui16_t)len);
}


/*
 *  Advances the cursors to the next record in the message (if any).
 */
int lfap_next_record(LFAPBuffer *lbuf)
{
  LFAPIE ie;
  int len;

  if (lbuf->off == lbuf->msglen) {
	lbuf->err = LFAP_NODATA;
	return LFAP_ERROR;
  }

  if (lbuf->off == lbuf->mlt.end) {
	lbuf->form.off = 0;

	if ((len = _decode_ie_hdr(lbuf, &ie)) < 0)
		return LFAP_ERROR;

	if (ie.type != LFAP_TRANSPORT_MULTIPLE_RECORD) {
		lbuf->err = LFAP_MSG_CORRUPTED;
		return LFAP_ERROR;
	}

	if (_decode_mult_hdr(lbuf, (lfapui16_t)len) == LFAP_ERROR)
		return LFAP_ERROR;

	return LFAP_SUCCESS;
  }

  if (lbuf->form.off && lbuf->form.off == lbuf->form.end) {
	lbuf->mlt.off = lbuf->off;
	lbuf->off = lbuf->mlt.beg;
	lbuf->form.off = 0;

	return LFAP_SUCCESS;
  }

  return LFAP_SUCCESS;
}
