/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_string.h"
#endif
/* End of auto-generated includes */

#include "LfapFAR.h"
#include "LfapMsg.h"


/******************************************************************************/
/*                                                                            */
/*  Function: resetLfapFAR                                                    */
/*                                                                            */
/*  Resets the internal IEs of the LFAP_FAR message structure to 0.           */
/*                                                                            */
/******************************************************************************/
void resetLfapFAR (LFAP_FAR* lfap_far)
{

  lfap_memset(lfap_far, 0, sizeof(LFAP_FAR));

  /* PAC_REMOVE     lfap_memset(&(lfap_far->flow_id), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->time), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->src_addr), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->dest_addr), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->src_port), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->protocol), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->tos), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->client_data), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->src_cce_addr), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->qualifier), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->src_dest_as), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->ingress_port), 0, LFAP_IE_SIZE); */
  /* PAC_REMOVE     lfap_memset(&(lfap_far->egress_port), 0, LFAP_IE_SIZE); */
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapFAR                                                     */
/*                                                                            */
/*  Initializes the internals of the LFAP_FAR message structure.              */
/*                                                                            */
/******************************************************************************/
void initLfapFAR (LFAP_FAR* lfap_far)
{
    resetLfapFAR(lfap_far);
    lfap_buffer_init(&(lfap_far->lfap_msg), 0, 0);

    lfap_far->hdr.opcode = LFAP_FAR_OPCODE;
    lfap_far->hdr.status = LFAP_MSG_SUCCESS;
    lfap_far->hdr.msgid = getNextLfapMsgId();
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapFARflowId                                               */
/*                                                                            */
/*  Initializes the internals of the LFAP_FAR message structure, along with   */
/*  setting the flow id IE (a mandatory IE for an LFAP_FAR message).          */
/*                                                                            */
/******************************************************************************/
void initLfapFARflowId (LFAP_FAR* lfap_far, lfapui32_t fas_id_hi,
                        lfapui32_t fas_id_lo, lfapui32_t cce_id)
{
    initLfapFAR(lfap_far);
    setFlowIdLfapFAR(lfap_far, fas_id_hi, fas_id_lo, cce_id);
}

/******************************************************************************/
/*                                                                            */
/*  Function: initRcvdLfapFAR                                                 */
/*                                                                            */
/*  Initializes the internals of the LFAP_FAR message structure, from a       */
/*  received message.  LFAP_SUCCESS is returned upon success, LFAP_ERROR is   */
/*  returned if there is an error.  lfap_buffer_get_error(&far->buf) can be   */
/*  used if there is an error.                                                */
/*                                                                            */
/*  WARNING - caller must ensure buf is valid as long as lfap_far is in       */
/*            use. lfap_far may contain pointers into buf.                    */
/*                                                                            */
/******************************************************************************/
int initRcvdLfapFAR (LFAP_FAR* lfap_far, void* buf, int buf_len)
{
  LFAPIE ie;
  int get_header_val = LFAP_SUCCESS;
  
  resetLfapFAR(lfap_far);
  
  lfap_buffer_init(&lfap_far->lfap_msg, buf, buf_len);
  lfap_buffer_insert(&lfap_far->lfap_msg, buf_len);
  
  get_header_val = lfap_get_header(&lfap_far->lfap_msg, &lfap_far->hdr);
  if (get_header_val == LFAP_SUCCESS) {
    lfap_memset(&ie, 0, LFAP_IE_SIZE);
    while (lfap_get_IE(&lfap_far->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
      setComponentLfapFAR(lfap_far, &ie);
  }
  
  return get_header_val;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setComponentLfapFAR                                             */
/*                                                                            */
/*  Sets one of the internal IEs of the LFAP_FAR message structure.           */
/*                                                                            */
/******************************************************************************/
int setComponentLfapFAR (LFAP_FAR* lfap_far, LFAPIE* ie)
{
    int rc = LFAP_SUCCESS;

    switch (ie->type) {
        case LFAP_FLOW_ID:
            lfap_far->flow_id = *ie;
            break;
        case LFAP_TIME:
            lfap_far->time = *ie;
            break;
        case LFAP_SRC_ADDR:
            lfap_far->src_addr = *ie;
            break;
        case LFAP_DST_ADDR:
            lfap_far->dest_addr = *ie;
            break;
        case LFAP_SRC_PORT_UDP:
        case LFAP_SRC_PORT_TCP:
        case LFAP_SRC_SOCKET_IPX:
            lfap_far->src_port = *ie;
            break;
        case LFAP_PROTOCOL_ID:
            {
              lfapui8_t len;
              lfapui8_t off;
              lfapui8_t i;
              
              /* decode the RMON string */
              off = 1; /* skip the zero byte, that holds the length */
              len = ie->octetStr_ie.val[0];
              for( i = 0; i < LFAP_MAX_PROTOCOL_LIST && off <= len; ++i) {
                memcpy(&lfap_far->protocolList[i], &ie->octetStr_ie.val[off], sizeof(lfapui32_t));
                lfap_far->protocolList[i] = ntohl(lfap_far->protocolList[i]);
                off += sizeof(lfapui32_t);
              }
              lfap_far->protocolListCnt = i+1;
              /* assign the IE */
              lfap_far->protocol = *ie;
            }
            break;
        case LFAP_CLIENT_DATA:
            lfap_far->client_data = *ie;
            break;
        case LFAP_SRC_CCE_ADDR:
            lfap_far->src_cce_addr = *ie;
            break;
        case LFAP_FLOW_QUALIFIER:
            lfap_far->qualifier = *ie;
            break;
        case LFAP_SRC_AS:
            lfap_far->src_AS = *ie;
            break;
        case LFAP_DEST_AS:
            lfap_far->dest_AS = *ie;
            break;
        case LFAP_INGRESS_PORT:
            lfap_far->ingress_port = *ie;
            break;
        case LFAP_EGRESS_PORT:
            lfap_far->egress_port = *ie;
            break;
        case LFAP_VLAN_ID:
            lfap_far->vlan_id = *ie;
            break;
        case LFAP_MPLS_IN_LABEL:
            if ( lfap_far->inLabelCount < LFAP_MAX_LABELS )
              lfap_far->inMplsLabel[lfap_far->inLabelCount++] = *ie;
            /* PAC - TODO should bump some error counter on else */
            break;
        case LFAP_MPLS_OUT_LABEL:
            if ( lfap_far->outLabelCount < LFAP_MAX_LABELS )
              lfap_far->outMplsLabel[lfap_far->outLabelCount++] = *ie;
            /* PAC - TODO should bump some error counter on else */
            break;
        case LFAP_LDP_FEC:
            if ( lfap_far->fecCount < LFAP_MAX_FECs )
              lfap_far->ldpFEC[lfap_far->fecCount++] = *ie;
            /* PAC - TODO should bump some error counter on else */
            break;
        case LFAP_NEXT_HOP_ADDR:
            lfap_far->next_hop_addr = *ie;
            break;
        case LFAP_CLASS_OF_SERVICE:
            if ( ie->CoS.RT & LFAP_COS_RECEIVED ) {
              lfap_far->CoS_Arr[ie->CoS.CoS_type][LFAP_COS_RECEIVED] = *ie;
              lfap_far->CoS_Arr[ie->CoS.CoS_type][LFAP_COS_RECEIVED].CoS.RT = LFAP_COS_RECEIVED;
            }

            if ( ie->CoS.RT & LFAP_COS_TRANSMITTED ) {
              lfap_far->CoS_Arr[ie->CoS.CoS_type][LFAP_COS_TRANSMITTED] = *ie;
              lfap_far->CoS_Arr[ie->CoS.CoS_type][LFAP_COS_TRANSMITTED].CoS.RT = LFAP_COS_TRANSMITTED;
            }
            break;
        default:
            rc = LFAP_ERROR;
            break;
    }

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFlowIdLfapFAR                                                */
/*                                                                            */
/*  Sets the flow_id IE of the LFAP_FAR message structure.                    */
/*                                                                            */
/******************************************************************************/
void setFlowIdLfapFAR (LFAP_FAR* lfap_far, lfapui32_t fas_id_hi,
                       lfapui32_t fas_id_lo, lfapui32_t cce_id)
{
  lfap_far->flow_id.fid_ie.type = LFAP_FLOW_ID;
  lfap_far->flow_id.fid_ie.fas_id[0] = fas_id_hi;
  lfap_far->flow_id.fid_ie.fas_id[1] = fas_id_lo;
  lfap_far->flow_id.fid_ie.cce_id = cce_id;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setTimeLfapFAR                                                  */
/*                                                                            */
/*  Sets the time IE of the LFAP_FAR message structure.                       */
/*                                                                            */
/******************************************************************************/
void setTimeLfapFAR (LFAP_FAR* lfap_far, lfapui32_t time)
{
  lfap_far->time.i32_ie.type = LFAP_TIME;
  lfap_far->time.i32_ie.val = time;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setSrcAddrLfapFAR                                               */
/*                                                                            */
/*  Sets the src_addr IE of the LFAP_FAR message structure.                   */
/*                                                                            */
/******************************************************************************/
void setSrcAddrLfapFAR (LFAP_FAR* lfap_far, char* addr, lfapui16_t family,
                        lfapui16_t len)
{
  lfap_far->src_addr.addr_ie.type = LFAP_SRC_ADDR;
  lfap_far->src_addr.addr_ie.val = addr;
  lfap_far->src_addr.addr_ie.family = family;
  lfap_far->src_addr.addr_ie.len = len;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setDestAddrLfapFAR                                              */
/*                                                                            */
/*  Sets the dest_addr IE of the LFAP_FAR message structure.                  */
/*                                                                            */
/******************************************************************************/
void setDestAddrLfapFAR (LFAP_FAR* lfap_far, char* addr, lfapui16_t family,
                         lfapui16_t len)
{
  lfap_far->dest_addr.addr_ie.type = LFAP_DST_ADDR;
  lfap_far->dest_addr.addr_ie.val = addr;
  lfap_far->dest_addr.addr_ie.family = family;
  lfap_far->dest_addr.addr_ie.len = len;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setSrcPortLfapFAR                                               */
/*                                                                            */
/*  Sets the src_port IE of the LFAP_FAR message structure.                   */
/*                                                                            */
/******************************************************************************/
void setSrcPortLfapFAR (LFAP_FAR* lfap_far, lfapui16_t type, lfapui16_t port)
{
  lfap_far->src_port.i32_ie.type = type;
  lfap_far->src_port.i32_ie.val = port;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setProtocolLfapFAR                                              */
/*                                                                            */
/*  Sets the protocol IE of the LFAP_FAR message structure.                   */
/*  WARNING - protocol buffer must be valid until encodeCompletedLfapFAR      */
/*            is called                                                       */
/*                                                                            */
/******************************************************************************/
void setProtocolLfapFAR (LFAP_FAR* lfap_far, char* protocol, lfapui16_t len)
{
  lfap_far->protocol.octetStr_ie.type = LFAP_PROTOCOL_ID;
  lfap_far->protocol.octetStr_ie.val = protocol;
  lfap_far->protocol.octetStr_ie.len = len;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setClientDataLfapFAR                                            */
/*                                                                            */
/*  Sets the client_data IE of the LFAP_FAR message structure.                */
/*  WARNING - data buffer must be valid until encodeCompletedLfapFAR          */
/*            is called                                                       */
/*                                                                            */
/******************************************************************************/
void setClientDataLfapFAR (LFAP_FAR* lfap_far, char* data, lfapui16_t len)
{
  lfap_far->client_data.octetStr_ie.type = LFAP_CLIENT_DATA;
  lfap_far->client_data.octetStr_ie.val = data;
  lfap_far->client_data.octetStr_ie.len = len;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setSrcCCEAddrLfapFAR                                            */
/*                                                                            */
/*  Sets the source CCE address IE of the LFAP_FAR message structure.         */
/*                                                                            */
/******************************************************************************/
void setSrcCCEAddrLfapFAR (LFAP_FAR* lfap_far, char* addr, lfapui16_t family,
                           lfapui16_t len)
{
    if ((family == LFAP_IP_VERSION_4) || (family == LFAP_IP_VERSION_6)) {
      lfap_far->src_cce_addr.addr_ie.type = LFAP_SRC_CCE_ADDR;
      lfap_far->src_cce_addr.addr_ie.val = addr;
      lfap_far->src_cce_addr.addr_ie.family = family;
      lfap_far->src_cce_addr.addr_ie.len = len;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: setSrcDestASLfapFAR                                             */
/*                                                                            */
/*  Sets the source and destination AS numbers IE of the LFAP_FAR message     */
/*  structure.                                                                */
/*                                                                            */
/******************************************************************************/
void setSrcASLfapFAR (LFAP_FAR* lfap_far, lfapui32_t src_as)
{
    lfap_far->src_AS.i16_ie.type = LFAP_SRC_AS;
    lfap_far->src_AS.i32_ie.val = src_as;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setIngressPortLfapFAR                                           */
/*                                                                            */
/*  Sets the ingress port IE of the LFAP_FAR message structure.               */
/*                                                                            */
/******************************************************************************/
void setIngressPortLfapFAR (LFAP_FAR* lfap_far, lfapui32_t ingress_port)
{
  lfap_far->ingress_port.i32_ie.type = LFAP_INGRESS_PORT;
  lfap_far->ingress_port.i32_ie.val = ingress_port;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setEgressPortLfapFAR                                            */
/*                                                                            */
/*  Sets the egress port IE of the LFAP_FAR message structure.                */
/*                                                                            */
/******************************************************************************/
void setEgressPortLfapFAR (LFAP_FAR* lfap_far, lfapui32_t egress_port)
{
  lfap_far->egress_port.i32_ie.type = LFAP_EGRESS_PORT;
  lfap_far->egress_port.i32_ie.val = egress_port;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setVlanIDLfapFAR                                                */
/*                                                                            */
/*  Sets the VLAN ID IE of the LFAP_FAR message structure.                    */
/*                                                                            */
/******************************************************************************/
void setVlanIDLfapFAR(LFAP_FAR* lfap_far, lfapui16_t s_vlan_id, lfapui16_t d_vlan_id)
{
    lfap_far->vlan_id.i16_ie.type = LFAP_VLAN_ID;
    lfap_far->vlan_id.i16_ie.val[0] = s_vlan_id;
    lfap_far->vlan_id.i16_ie.val[1] = d_vlan_id;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setMplsLabelLfapFAR                                             */
/*                                                                            */
/*  Sets a MPLS Label IE of the LFAP_FAR message structure.                   */
/*                                                                            */
/******************************************************************************/
void setMplsLabelLfapFAR(LFAP_FAR* lfap_far, lfapui8_t in, lfapui32_t label)
{
  if ( in == LFAP_IN_LABEL) {
    if (lfap_far->inLabelCount < LFAP_MAX_LABELS ) {
      lfap_far->inMplsLabel[lfap_far->inLabelCount].mplsLabel.type = LFAP_MPLS_IN_LABEL;
      lfap_far->inMplsLabel[lfap_far->inLabelCount].mplsLabel.reserved = 0;
      lfap_far->inMplsLabel[lfap_far->inLabelCount].mplsLabel.val = label;
      lfap_far->inLabelCount++;
    }
  }
  else {
    if (lfap_far->outLabelCount < LFAP_MAX_LABELS ) {
      lfap_far->outMplsLabel[lfap_far->outLabelCount].mplsLabel.type = LFAP_MPLS_OUT_LABEL;
      lfap_far->outMplsLabel[lfap_far->outLabelCount].mplsLabel.reserved = 0;
      lfap_far->outMplsLabel[lfap_far->outLabelCount].mplsLabel.val = label;
      lfap_far->outLabelCount++;
    }
  }
}

/******************************************************************************/
/*                                                                            */
/*  Function: setLDPFECLfapFAR                                                */
/*                                                                            */
/*  Sets the LDP FEC IE of the LFAP_FAR message structure. ldpFEC must        */
/*  already be in the format specified by RFC 3036 and/or                     */
/*  draft-martini-l2circuit-trans-mpls-07.txt                                 */
/*                                                                            */
/*  WARNING - the fec  is treated as an  octet string. This was done as a     */
/*            performance optimization to save making a copy.                 */
/*                                                                            */
/*  WARNING - ldpFEC buffer passed in  must be valid until                    */
/*            encodeCompletedLfapFAR is called.                               */
/*                                                                            */
/******************************************************************************/
void setLDPFECLfapFAR (LFAP_FAR* lfap_far, lfapui8_t* ldpFEC, lfapui16_t len)
{
  if (lfap_far->fecCount < LFAP_MAX_FECs) {
    lfap_far->ldpFEC[lfap_far->fecCount].octetStr_ie.type = LFAP_LDP_FEC;
    lfap_far->ldpFEC[lfap_far->fecCount].octetStr_ie.val = (char *)ldpFEC;
    lfap_far->ldpFEC[lfap_far->fecCount].octetStr_ie.len = len;
    lfap_far->fecCount++;
  }
}

/******************************************************************************/
/*                                                                            */
/*  Function: setCoSLfapFAR                                                   */
/*                                                                            */
/*  Sets the Class of Service IE of the LFAP_FAR message structure.                    */
/*                                                                            */
/******************************************************************************/
void setCoSLfapFAR(LFAP_FAR* lfap_far, CoS_Types_t type, CoS_Types_t rt, lfapui32_t val)
{
  if ( rt & LFAP_COS_TRANSMITTED ) {
    lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED].type = LFAP_CLASS_OF_SERVICE;
    lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED].CoS.CoS_type = type;
    lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED].CoS.RT = LFAP_COS_TRANSMITTED;
    lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED].CoS.val = val;
  }

  if ( rt & LFAP_COS_RECEIVED ) {
    lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED].type = LFAP_CLASS_OF_SERVICE;
    lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED].CoS.CoS_type = type;
    lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED].CoS.RT = LFAP_COS_RECEIVED;
    lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED].CoS.val = val;
  }


}

/******************************************************************************/
/*                                                                            */
/*  Function: getNextQualifierLfapFAR_                                        */
/*                                                                            */
/*  Gets the next LFAPQualifier from the list of qualifiers.  If the list is  */
/*  full (items == LFAP_MAX_QUALIFIERS) 0 is returned.                        */
/*                                                                            */
/******************************************************************************/
static LFAPQualifier* getNextQualifierLfapFAR_ (LFAP_FAR* lfap_far)
{
  LFAPQualifier* qualifier = 0;
  
  lfap_far->qualifier.type = LFAP_FLOW_QUALIFIER;
  if (lfap_far->qualifier.qual_ie.items < LFAP_MAX_QUALIFIERS)
    qualifier = &lfap_far->qualifier.qual_ie.q_list[lfap_far->qualifier.qual_ie.items++];

  return qualifier;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setCheckpointLfapFAR                                            */
/*                                                                            */
/*  Sets the checkpoint qualifier IE of the LFAP_FAR message structure.       */
/*                                                                            */
/******************************************************************************/
void setCheckpointLfapFAR (LFAP_FAR* lfap_far, lfapui32_t checkpoint)
{
    LFAPQualifier* qualifier;

    if ((qualifier = getNextQualifierLfapFAR_(lfap_far))) {
      checkpoint = htonl(checkpoint);
      lfap_memcpy(lfap_far->checkpoint, &checkpoint, LFAP_UINT32_SIZE);
      qualifier->id = LFAP_CHECKPOINT_ID;
      qualifier->len = LFAP_UINT32_SIZE;
      qualifier->val = (char*) &lfap_far->checkpoint;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: setCxnPriorityLfapFAR                                           */
/*                                                                            */
/*  Sets the connection priority qualifier IE of the LFAP_FAR message         */
/*  structure.                                                                */
/*                                                                            */
/******************************************************************************/
void setCxnPriorityLfapFAR (LFAP_FAR* lfap_far, lfapui32_t cxn_priority)
{
  LFAPQualifier* qualifier;
  
  if ((qualifier = getNextQualifierLfapFAR_(lfap_far))) {
    cxn_priority = htonl(cxn_priority);
    lfap_memcpy(lfap_far->cxn_priority, &cxn_priority, LFAP_UINT32_SIZE);
    qualifier->id = LFAP_CXN_PRIORITY_ID;
    qualifier->len = LFAP_UINT32_SIZE;
    qualifier->val = (char*) &lfap_far->cxn_priority;
  }
}

/******************************************************************************/
/*                                                                            */
/*  Function: setNextHopAddrLfapFAR                                               */
/*                                                                            */
/*  Sets the next_hop_addr IE of the LFAP_FAR message structure.                   */
/*                                                                            */
/******************************************************************************/
void setNextHopAddrLfapFAR (LFAP_FAR* lfap_far, char* addr, lfapui16_t family,
                            lfapui16_t len)
{
    lfap_far->next_hop_addr.addr_ie.type = LFAP_NEXT_HOP_ADDR;
    lfap_far->next_hop_addr.addr_ie.val = addr;
    lfap_far->next_hop_addr.addr_ie.family = family;
    lfap_far->next_hop_addr.addr_ie.len = len;
}

/******************************************************************************/
/*                                                                            */
/*  Function: seDestASLfapFAR                                             */
/*                                                                            */
/*  Sets the destination AS numbers IE of the LFAP_FAR message     */
/*  structure.                                                                */
/*                                                                            */
/******************************************************************************/
void setDestASLfapFAR (LFAP_FAR* lfap_far,  lfapui32_t dest_as)
{
    lfap_far->dest_AS.i32_ie.type = LFAP_DEST_AS;
    lfap_far->dest_AS.i32_ie.val = dest_as;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeCompletedLfapFAR                                          */
/*                                                                            */
/*  Encodes the LFAP_FAR message by placing the IEs into the LFAPBuffer.      */
/*  The LFAP_FAR message (in the form of the LFAPBuffer) is ready to be sent. */
/*                                                                            */
/******************************************************************************/
int encodeCompletedLfapFAR (LFAP_FAR* lfap_far)
{
    int type, i;
    LFAPBuffer* far_buf = (&lfap_far->lfap_msg);

    lfap_begin_message(far_buf, &lfap_far->hdr);

    /*          WARNING WARNING WARNING

       The order is important, in other LFAP api code there is a giant 
       hack where we dig into  the message after it is built to change 
       a value. It depends on the order of the encoding to work and the ies
       encoded before the hack. At some point the hack should get changed 
       to a less error prone method. (see LfapFlowMgr.c) 
    */

    if (lfap_far->flow_id.type)
        lfap_add_IE(far_buf, &lfap_far->flow_id);
    if (lfap_far->time.type)
        lfap_add_IE(far_buf, &lfap_far->time);
    if (lfap_far->src_addr.type)
        lfap_add_IE(far_buf, &lfap_far->src_addr);
    if (lfap_far->dest_addr.type)
        lfap_add_IE(far_buf, &lfap_far->dest_addr);
    if (lfap_far->src_AS.type)
        lfap_add_IE(far_buf, &lfap_far->src_AS); /* HACK is on this field */
    if (lfap_far->src_port.type)
        lfap_add_IE(far_buf, &lfap_far->src_port);
    if (lfap_far->protocol.type)
        lfap_add_IE(far_buf, &lfap_far->protocol);
    if (lfap_far->client_data.type)
        lfap_add_IE(far_buf, &lfap_far->client_data);
    if (lfap_far->src_cce_addr.type)
        lfap_add_IE(far_buf, &lfap_far->src_cce_addr);
    if (lfap_far->qualifier.type)
        lfap_add_IE(far_buf, &lfap_far->qualifier);
    if (lfap_far->ingress_port.type)
        lfap_add_IE(far_buf, &lfap_far->ingress_port);
    if (lfap_far->egress_port.type)
        lfap_add_IE(far_buf, &lfap_far->egress_port);
    if (lfap_far->vlan_id.type)
        lfap_add_IE(far_buf, &lfap_far->vlan_id);
    if (lfap_far->next_hop_addr.type)
        lfap_add_IE(far_buf, &lfap_far->next_hop_addr);
    if (lfap_far->dest_AS.type)
        lfap_add_IE(far_buf, &lfap_far->dest_AS); 


    for( i = 0; i < LFAP_MAX_LABELS; i++) {
      if (lfap_far->inMplsLabel[i].type)
        lfap_add_IE(far_buf, &lfap_far->inMplsLabel[i]);
      if (lfap_far->outMplsLabel[i].type)
        lfap_add_IE(far_buf, &lfap_far->outMplsLabel[i]);
    }

    /* We may have several different flavors of Class of Service.
       Cycle through them and add the ones that have a value.
    */
    for( type = 1; type <= LFAP_COS_MAX_TYPES; type++) {

      /* If the value is the same for received and transmitted, we can
         save a little space on the wire by setting the R/T bits in one IE
      */
      if ( lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED].type &&
           lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED].type &&
            (lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED].CoS.val ==
             lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED].CoS.val) ) {

          /* The R bit should already be set, now set the T bit */
          /* Also, keep us from sending the other IE           */
          lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED].CoS.RT |= LFAP_COS_TRANSMITTED;
          lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED].type = 0;
        }


      if (lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED].type)
        lfap_add_IE(far_buf, &lfap_far->CoS_Arr[type][LFAP_COS_RECEIVED]);

      if (lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED].type)
        lfap_add_IE(far_buf, &lfap_far->CoS_Arr[type][LFAP_COS_TRANSMITTED]);

    }
    
    for(i = 0; i < LFAP_MAX_FECs; i++) {
      if (lfap_far->ldpFEC[i].type)
        lfap_add_IE(far_buf, &lfap_far->ldpFEC[i]);
    }

    return lfap_end_message(far_buf);
}

/******************************************************************************/
/*                                                                            */
/*  Function: verifyLfapFAR                                                   */
/*                                                                            */
/*  Verifies which IEs of the received LFAP_FAR message are encoded.          */
/*  LFAP_FAR_INVALID is returned if the mandatory IEs are not encoded,        */
/*  otherwise the returned value indicates which IEs are encoded in the       */
/*  message.                                                                  */
/*                                                                            */
/******************************************************************************/
lfapui32_t verifyLfapFAR (LFAP_FAR* lfap_far)
{
    lfapui32_t verify = LFAP_FAR_INVALID;
    lfapui32_t t;

    if (lfap_far->flow_id.type && lfap_far->time.type )
    {
        verify = LFAP_FAR_VALID;
        if (lfap_far->src_addr.type)
           verify += LFAP_FAR_SRC_ADDR;
        if (lfap_far->dest_addr.type)
           verify += LFAP_FAR_DEST_ADDR;
        if (lfap_far->src_port.type)
            verify += LFAP_FAR_SRC_PORT;
        if (lfap_far->protocol.type)
            verify += LFAP_FAR_PROTOCOL_ID;
        if (lfap_far->client_data.type)
            verify += LFAP_FAR_CLIENT_DATA;
        if (lfap_far->src_cce_addr.type)
            verify += LFAP_FAR_SOURCE_CCE_ADDR;
        if (lfap_far->qualifier.type)
            verify += LFAP_FAR_QUALIFIER;
        if (lfap_far->src_AS.type)
            verify += LFAP_FAR_SRC_AS;
        if (lfap_far->ingress_port.type)
            verify += LFAP_FAR_INGRESS_PORT;
        if (lfap_far->egress_port.type)
            verify += LFAP_FAR_EGRESS_PORT;
        if (lfap_far->vlan_id.type)
            verify += LFAP_FAR_VLAN_ID;
        if (lfap_far->inMplsLabel[0].type)
            verify += LFAP_FAR_IN_MPLS_LABEL;
        if (lfap_far->outMplsLabel[0].type)
            verify += LFAP_FAR_OUT_MPLS_LABEL;
        if (lfap_far->ldpFEC[0].type)
            verify += LFAP_FAR_LDP_FEC;
        if (lfap_far->next_hop_addr.type)
            verify += LFAP_FAR_NEXT_HOP_ADDR;
        if (lfap_far->dest_AS.type)
            verify += LFAP_FAR_DEST_AS;

        for( t = 1; t <= LFAP_COS_MAX_TYPES; t++) {
          if ( lfap_far->CoS_Arr[t][LFAP_COS_RECEIVED].type ||
               lfap_far->CoS_Arr[t][LFAP_COS_TRANSMITTED].type ) {
            verify += LFAP_FAR_COS;
            break;
          }
        }
    }

    return verify;
}

/******************************************************************************/
/*                                                                            */
/*  Function: getDestPortLfapFAR                                              */
/*                                                                            */
/*  Get the TCP/UDP destination port.                                         */
/*                                                                            */
/*   Return Value of LFAP_FAILURE means there is no dest port for this flow.  */
/******************************************************************************/
lfapui32_t getDestPortLfapFAR (LFAP_FAR* lfap_far, lfapui16_t *port)
{
  if ( lfap_far->protocolListCnt <= LFAP_PROTOCOL_L4 ||
       lfap_far->protocolList[LFAP_PROTOCOL_L1] != LFAP_ETHER ||
       lfap_far->protocolList[LFAP_PROTOCOL_L2] != LFAP_IP ||
       lfap_far->protocolList[LFAP_PROTOCOL_L4] > 0xFFFF )
    return LFAP_FAILURE;


  *port = (lfapui16_t)lfap_far->protocolList[LFAP_PROTOCOL_L4];
  return LFAP_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/*  Function: getIPprotocolLfapFAR                                            */
/*                                                                            */
/*  Get the IP protocol (e.g. TCP/UDP)                                        */
/*                                                                            */
/*   Return Value of LFAP_FAILURE means there is no IP protocol this flow.    */
/******************************************************************************/
lfapui32_t getIPprotocolLfapFAR (LFAP_FAR* lfap_far, lfapui8_t *proto)
{
  if ( lfap_far->protocolListCnt <= LFAP_PROTOCOL_L3 ||
       lfap_far->protocolList[LFAP_PROTOCOL_L1] != LFAP_ETHER ||
       lfap_far->protocolList[LFAP_PROTOCOL_L2] != LFAP_IP ||
       lfap_far->protocolList[LFAP_PROTOCOL_L3] > 0xFF )
    return LFAP_FAILURE;

  *proto = (lfapui8_t)lfap_far->protocolList[LFAP_PROTOCOL_L3];
  return LFAP_SUCCESS;
}


/******************************************************************************/
/*                                                                            */
/*  Function: getFlowTypeLfapFAR                                              */
/*                                                                            */
/*  Get the flow type (e.g. IP or MPLS                                        */
/*                                                                            */
/*   Return Value is always LFAP_SUCCESS.                                     */
/******************************************************************************/
lfapui32_t getFlowTypeLfapFAR (LFAP_FAR* lfap_far, lfapui32_t *type)
{
  *type = LFAP_IP; /* IP is the default */

  if ( lfap_far->protocolListCnt > LFAP_PROTOCOL_L2 &&
       lfap_far->protocolList[LFAP_PROTOCOL_L1] == LFAP_ETHER )
    *type = lfap_far->protocolList[LFAP_PROTOCOL_L2];

  return LFAP_SUCCESS;
}
