/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_string.h"
#endif
/* End of auto-generated includes */

#include "LfapFUN.h"
#include "LfapMsg.h"


/******************************************************************************/
/*                                                                            */
/*  Function: resetLfapFUN                                                    */
/*                                                                            */
/*  Resets the internal IEs of the LFAP_FUN message structure to 0.           */
/*                                                                            */
/******************************************************************************/
void resetLfapFUN (LFAP_FUN* fun, lfapui8_t resetFlags)
{
    lfap_memset(&(fun->flow_state), 0, LFAP_IE_SIZE);
    lfap_memset(&(fun->byte_count), 0, LFAP_IE_SIZE);
    lfap_memset(&(fun->packet_count), 0, LFAP_IE_SIZE);
    lfap_memset(&(fun->flow_id), 0, LFAP_IE_SIZE);
    lfap_memset(&(fun->time), 0, LFAP_IE_SIZE);
    if (resetFlags == LFAP_RESET_FLAGS) fun->encode_flag = LFAP_NOT_ENCODED;
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapFUN                                                     */
/*                                                                            */
/*  Initializes the internals of the LFAP_FUN message structure.              */
/*                                                                            */
/******************************************************************************/
void initLfapFUN (LFAP_FUN* fun)
{
    resetLfapFUN(fun, LFAP_RESET_FLAGS);
    lfap_buffer_init(&(fun->lfap_msg), 0, 0);

    fun->encode_flag = LFAP_NOT_ENCODED;

    fun->hdr.opcode = LFAP_FUN_OPCODE;
    fun->hdr.status = LFAP_MSG_SUCCESS;
    fun->hdr.msgid = getNextLfapMsgId();
}

/******************************************************************************/
/*                                                                            */
/*  Function: initLfapFUNflowId                                               */
/*                                                                            */
/*  Initializes the internals of the LFAP_FUN message structure, along with   */
/*  setting the flow id IE (a mandatory IE for an LFAP_FUN message).          */
/*                                                                            */
/******************************************************************************/
void initLfapFUNflowId (LFAP_FUN* fun, lfapui32_t fas_id_hi,
                        lfapui32_t fas_id_lo, lfapui32_t cce_id)
{
    initLfapFUN(fun);
    setFlowIdLfapFUN(fun, fas_id_hi, fas_id_lo, cce_id);
}

/******************************************************************************/
/*                                                                            */
/*  Function: initRcvdLfapFUN                                                 */
/*                                                                            */
/*  Initializes the internals of the LFAP_FUN message structure, from a       */
/*  received message.  LFAP_SUCCESS is returned upon success, LFAP_ERROR is   */
/*  returned if there is an error. lfap_buffer_get_error(&fun->lfap_msg) can  */
/*  be used if there is an error.                                             */
/*                                                                            */
/******************************************************************************/
int initRcvdLfapFUN (LFAP_FUN* fun, void* buf, int buf_len)
{
  LFAPIE ie;
  int get_header_val = LFAP_SUCCESS;
  
  resetLfapFUN(fun, LFAP_RESET_FLAGS);
  
  lfap_buffer_init(&fun->lfap_msg, buf, buf_len);
  lfap_buffer_insert(&fun->lfap_msg, buf_len);
  
  get_header_val = lfap_get_header(&fun->lfap_msg, &fun->hdr);
  if (get_header_val == LFAP_SUCCESS) {
    lfap_memset(&ie, 0, LFAP_IE_SIZE);
    while (lfap_get_IE(&fun->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
      setComponentLfapFUN(fun, &ie);
  }
  
  return get_header_val;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setComponentLfapFUN                                             */
/*                                                                            */
/*  Sets one of the internal IEs of the LFAP_FUN message structure.           */
/*                                                                            */
/******************************************************************************/
int setComponentLfapFUN (LFAP_FUN* fun, LFAPIE* ie)
{
    int rc = LFAP_SUCCESS;

    switch (ie->type) {
        case LFAP_BYTES_DELTA:
            fun->byte_count = *ie;
            break;
        case LFAP_PACKETS_DELTA:
            fun->packet_count = *ie;
            break;
        case LFAP_BYTES_TOTAL:
            fun->byte_count = *ie;
            break;
        case LFAP_PACKETS_TOTAL:
            fun->packet_count = *ie;
            break;
        case LFAP_FLOW_ID:
            fun->flow_id = *ie;
            break;
        case LFAP_FLOW_STATE:
            fun->flow_state = *ie;
            break;
        case LFAP_TIME:
            fun->time = *ie;
            break;
        default:
            rc = LFAP_ERROR;
            break;
    }
    if (rc == LFAP_SUCCESS)
        fun->encode_flag |= LFAP_ENCODE_IE;

    return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setFlowIdLfapFUN                                                */
/*                                                                            */
/*  Sets the flow_id IE of the LFAP_FUN message structure.                    */
/*                                                                            */
/******************************************************************************/
void setFlowIdLfapFUN (LFAP_FUN* fun, lfapui32_t fas_id_hi,
                       lfapui32_t fas_id_lo, lfapui32_t cce_id)
{
  fun->flow_id.fid_ie.type = LFAP_FLOW_ID;
  fun->flow_id.fid_ie.fas_id[0] = fas_id_hi;
  fun->flow_id.fid_ie.fas_id[1] = fas_id_lo;
  fun->flow_id.fid_ie.cce_id = cce_id;

  fun->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setStateLfapFUN                                                 */
/*                                                                            */
/*  Sets the flow_state IE of the LFAP_FUN message structure.                 */
/*                                                                            */
/******************************************************************************/
void setStateLfapFUN (LFAP_FUN* fun, lfapui32_t state)
{
    fun->flow_state.rsvd_ui8.type = LFAP_FLOW_STATE;
    fun->flow_state.rsvd_ui8.reserved = 0;
    fun->flow_state.rsvd_ui8.val = state;

    fun->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setTimeLfapFUN                                                  */
/*                                                                            */
/*  Sets the time IE of the LFAP_FUN message structure.                       */
/*                                                                            */
/******************************************************************************/
void setTimeLfapFUN (LFAP_FUN* fun, lfapui32_t time)
{
  fun->time.i32_ie.type = LFAP_TIME;
  fun->time.i32_ie.val = time;

  fun->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setByteCountLfapFUN                                             */
/*                                                                            */
/*  Sets the byte_count IE of the LFAP_FUN message structure.                 */
/*                                                                            */
/******************************************************************************/
void setByteCountLfapFUN (LFAP_FUN* fun, 
                          lfapui16_t type, 
                          lfapui64_t rcvd,
                          lfapui64_t sent)
{
    fun->byte_count.dataCount_ie.type = type;
    fun->byte_count.dataCount_ie.dc_rcvd.hi_val = rcvd >> 32;
    fun->byte_count.dataCount_ie.dc_rcvd.lo_val = rcvd;
    fun->byte_count.dataCount_ie.dc_sent.hi_val = sent >> 32;
    fun->byte_count.dataCount_ie.dc_sent.lo_val = sent;

    fun->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setPacketCountLfapFUN                                           */
/*                                                                            */
/*  Sets the packet_count IE of the LFAP_FUN message structure.               */
/*                                                                            */
/******************************************************************************/
void setPacketCountLfapFUN (LFAP_FUN* fun, 
                            lfapui16_t type, 
                            lfapui64_t rcvd,
                            lfapui64_t sent)
{
    fun->packet_count.dataCount_ie.type = type;
    fun->packet_count.dataCount_ie.dc_rcvd.hi_val = rcvd >> 32;
    fun->packet_count.dataCount_ie.dc_rcvd.lo_val = rcvd;
    fun->packet_count.dataCount_ie.dc_sent.hi_val = sent >> 32;
    fun->packet_count.dataCount_ie.dc_sent.lo_val = sent;

    fun->encode_flag |= LFAP_ENCODE_IE;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeIEsLfapFUN                                                */
/*                                                                            */
/*  Encodes FUN IE's to the fun_buf.                                          */
/*                                                                            */
/******************************************************************************/
static void encodeIEsLfapFUN (LFAP_FUN* fun)
{
  LFAPBuffer* fun_buf = &(fun->lfap_msg);

    if (fun->flow_id.type)
        lfap_add_IE(fun_buf, &fun->flow_id);
    if (fun->time.type)
        lfap_add_IE(fun_buf, &fun->time);
    if (fun->flow_state.type)
        lfap_add_IE(fun_buf, &fun->flow_state);
    if (fun->byte_count.type)
        lfap_add_IE(fun_buf, &fun->byte_count);
    if (fun->packet_count.type)
        lfap_add_IE(fun_buf, &fun->packet_count);
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeFixedLfapFUN                                              */
/*                                                                            */
/*  Encodes the fixed portion of the LFAP_FUN message by placing the IEs into */
/*  the fixed portion of the LFAPBuffer.  If fixed information is desired to  */
/*  be encoded, this must be called before any other encode routine is        */
/*  called.  The IEs of the LFAP_FUN message are reset after the encoding is  */
/*  complete.                                                                 */
/*                                                                            */
/******************************************************************************/
int encodeFixedLfapFUN (LFAP_FUN* fun)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* fun_buf = &(fun->lfap_msg);
  
  if ((!((fun->encode_flag & LFAP_ENCODE_COMPLETE) ||
         (fun->encode_flag & LFAP_ENCODE_RECORD))) &&
      (fun->encode_flag & LFAP_ENCODE_IE)) {
    lfap_begin_message(fun_buf, &fun->hdr);
    lfap_begin_multiple(fun_buf);
    
    lfap_begin_fixed(fun_buf);
    encodeIEsLfapFUN(fun);
    rc = lfap_end_fixed(fun_buf);
    
    fun->encode_flag &= ~LFAP_ENCODE_IE;
    fun->encode_flag |= LFAP_ENCODE_FIXED;
    resetLfapFUN(fun, LFAP_NO_RESET_FLAGS);
  }
  
  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeRecordLfapFUN                                             */
/*                                                                            */
/*  Encodes a single record of the LFAP_FUN message by placing the IEs into   */
/*  the LFAPBuffer.                                                           */
/*                                                                            */
/******************************************************************************/
int encodeRecordLfapFUN (LFAP_FUN* fun)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* fun_buf = &(fun->lfap_msg);
  
  if ((!(fun->encode_flag & LFAP_ENCODE_COMPLETE)) &&
      (fun->encode_flag & LFAP_ENCODE_IE)) {
    if (!(fun->encode_flag & LFAP_ENCODE_RECORD) &&
        !(fun->encode_flag & LFAP_ENCODE_FIXED)) { 
      lfap_begin_message(fun_buf, &fun->hdr);
      lfap_begin_multiple(fun_buf);
    }
    lfap_begin_record(fun_buf);
    encodeIEsLfapFUN(fun);
    rc = lfap_end_record(fun_buf);
    
    fun->encode_flag &= ~LFAP_ENCODE_IE;
    fun->encode_flag |= LFAP_ENCODE_RECORD;
    resetLfapFUN(fun, LFAP_NO_RESET_FLAGS);
  }
  
  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: encodeCompletedLfapFUN                                          */
/*                                                                            */
/*  Finishes the encoding process of the LFAP_FUN message into the            */
/*  LFAPBuffer.  The LFAP_FUN message (in the form of the LFAPBuffer) is      */
/*  ready to be sent.                                                         */
/*                                                                            */
/******************************************************************************/
int encodeCompletedLfapFUN (LFAP_FUN* fun)
{
  int rc = LFAP_ERROR;
  LFAPBuffer* fun_buf = &(fun->lfap_msg);
  
  if (!(fun->encode_flag & LFAP_ENCODE_COMPLETE)) {
    if (fun->encode_flag & LFAP_ENCODE_IE) {
      if ((fun->encode_flag & LFAP_ENCODE_RECORD) ||
          (fun->encode_flag & LFAP_ENCODE_FIXED)) {
        lfap_begin_record(fun_buf);
        encodeIEsLfapFUN(fun);
        lfap_end_record(fun_buf);
        lfap_end_multiple(fun_buf);
      }
      else {
        lfap_begin_message(fun_buf, &fun->hdr);
        encodeIEsLfapFUN(fun);
      }
      rc = lfap_end_message(fun_buf);
      fun->encode_flag |= LFAP_ENCODE_COMPLETE;
    }
    else if (fun->encode_flag & LFAP_ENCODE_RECORD) {
      lfap_end_multiple(fun_buf);
      rc = lfap_end_message(fun_buf);
      fun->encode_flag |= LFAP_ENCODE_COMPLETE;
    }
  }
  
  return rc;
}

/******************************************************************************/
/*                                                                            */
/*  Function: decodeNextRecordLfapFUN                                         */
/*                                                                            */
/*  Decodes the next record in the multiple record LFAP_FUN message.  The IEs */
/*  of the record are placed into the IEs of the LFAP_FUN message.  If there  */
/*  are no more records, the LFAP_FUN's IEs are reset and LFAP_EMPTY_BUFFER   */
/*  is returned, else LFAP_SUCCESS is returned.                               */
/*                                                                            */
/******************************************************************************/
int decodeNextRecordLfapFUN (LFAP_FUN* fun)
{
  LFAPIE ie;
  int msg_decoded = lfap_next_record(&fun->lfap_msg);
  
  resetLfapFUN(fun, LFAP_NO_RESET_FLAGS);
  
  if (msg_decoded != LFAP_EMPTY_BUFFER) {
    lfap_memset(&ie, 0, LFAP_IE_SIZE);
    while (lfap_get_IE(&fun->lfap_msg, &ie) != LFAP_EMPTY_BUFFER)
      setComponentLfapFUN(fun, &ie);
  }
  return msg_decoded;
}

/******************************************************************************/
/*                                                                            */
/*  Function: verifyLfapFUN                                                   */
/*                                                                            */
/*  Verifies which IEs of the received LFAP_FUN message are encoded.          */
/*  LFAP_FUN_INVALID is returned if the mandatory IEs are not encoded,        */
/*  otherwise the returned value indicates which IEs are encoded in the       */
/*  message.                                                                  */
/*                                                                            */
/******************************************************************************/
lfapui8_t verifyLfapFUN (LFAP_FUN* fun)
{
    lfapui8_t verify = LFAP_FUN_INVALID;

    if (fun->time.type && fun->flow_id.type) {
        if (fun->flow_state.type ||
            fun->byte_count.type ||
            fun->packet_count.type) {
            verify = LFAP_FUN_VALID;
            if (fun->flow_state.type)
                verify += LFAP_FUN_FLOW_STATE;
            if (fun->byte_count.type)
                verify += LFAP_FUN_BYTE_COUNT;
            if (fun->packet_count.type)
                verify += LFAP_FUN_PACKET_COUNT;
        }
    }

    return verify;
}
