/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* 
 * WARNING WARNING
 *
 * since there are currently no clients on a windows platform
 * and since we don't want this file included when the windows
 * server is built under Visual C++, put in this ifdef
*/
#ifndef _WIN32

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#include "mls_syscalls.h"
#include "err_api.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "cdefs.h"
#include "nucleus.h"
#include "ni_api.h" 
#include "mls_in.h"
#include "mls_resolv.h"
#include "mls_netdb.h"
#include "mls_reent.h"
#include "mls_stdio.h"
#include "mls_stdlib.h"
#include "mls_string.h"
#include "mls_select.h"
#include "byteorder.h"
#include "mls_syscalls.h"
#include "err_api.h"
#include "lfap.emd.h"
#endif

/* End of auto-generated includes */

#include "LfapMsg.h"
#include "LfapFlowMgr.h"
#include "LfapPrivate.h"
#include "LfapMsgQueue.h"
#include "LfapCCEos.h"
#include "LfapCR.h"
#include "LfapKA.h"
#include "LfapDR.h"
#include "LfapCAN.h"
#include "LfapCRN.h"
#include "LfapFER.h"

#ifdef CS_DEBUG
#include "LfapDebug.h"
#endif

#define CLEAR 1

/* GARBAGE_COLLECTOR */
#define LFAP_GARBAGE_COLLECT_NUM_FLOWS	100   /* flows per pass */
#define LFAP_GARBAGE_COLLECT_INTERVAL	10000 /* 10 seconds */


/* set lfap_debug to desired level at runtime to have printouts, or 0 to not,
   as long as the printouts were compiled in (i.e. CS_DEBUG was defined) */
int lfap_debug = LFAP_DBG_ERRORS;

/* no connect/send to FAS - for performance testing */
int lfap_debug_fas = 1;

/* SSR error facility */
extern int lfap_efacid;

LFAPConnState_t LFAP_conn_state = LFAP_INITIAL_STATE;
#ifdef pac_remove

/* flag indicating if the CCE is connected to the LFAP Server, and if the
   LFAP Server is ready and willing to receive messages */
lfapui16_t LFAP_connected;
int LFAP_conn_status;


/* flag indicating if in the process of connecting to a server */
static int LFAP_trying_to_connect;

#endif

/* list of servers and pointer to the node of the current server */
static lfapui32_t LFAP_server;
static LFAPList LFAP_server_list;

/* Server can indicate to the CCE to change to this new server */
static lfapui32_t LFAP_new_server;

/* time before retrying to connect to the servers in the server list */
static lfapui32_t LFAP_server_retry_interval;

/* local ids (to this file) that hold the server ids as well as the
   current flow id (LFAP_cce_id) */
static lfapui32_t LFAP_cce_id;
static lfapui32_t LFAP_fas_hi;
static lfapui32_t LFAP_fas_lo;

/* global variables for the statistics and FUN and FAR messages */
static lfapui16_t LFAP_batch_size;
static lfapui32_t LFAP_poll_interval;
static lfapui16_t LFAP_bytes_stats;
static lfapui16_t LFAP_packets_stats;
static lfapui32_t LFAP_batch_interval;

/* Error structure for keeping the last error */
static LFAPError LFAP_error;

/* LFAP API Statistics */
LFAPApiStats* LFAP_stats;
LFAPApiStats* LFAP_curr_stats;

/* Message counter to test to see if the CCE has lost connection to the FAS,
   but the LFAP API has not been notified (i.e. network down, but have not
   actually lost the connection */
static lfapui16_t LFAP_msgs_rcvd;
static lfapui32_t LFAP_lost_conn_interval;
static lfapui16_t LFAP_lost_conn_count = 0;

/* timers */
lfapui16_t LFAP_lost_conn_timer;
lfapui16_t LFAP_connection_timer;
lfapui16_t LFAP_flow_update_timer;
lfapui16_t LFAP_batch_interval_timer;
lfapui16_t LFAP_garbage_collector_timer; /* GARBAGE_COLELCTOR */

/* reporting on flows */
static LFAP_AR LFAP_report_flows_ar;
static lfapui16_t LFAP_report_on_flows;
static LFAP_ARA LFAP_report_flows_ara;

/* sending vector, for writev() */
static lfapui16_t LFAP_send_vector_len;
static lfapui16_t LFAP_send_vector_entries;
static LFAPMsgHolder LFAP_send_vector[LFAP_MAX_MSGS_IN_SEND_VECTOR];
static LFAPNode* LFAP_send_vector_nodes[LFAP_MAX_MSGS_IN_SEND_VECTOR];

/* sending a single message, for write */
static lfapui16_t LFAP_send_message_len = 0;
static char *LFAP_send_message = 0;
static LFAPNode* LFAP_send_message_node = 0;

/* reporting on flows */
LFAPList LFAP_stats_list;            /* list of flows for stat reporting */
static int LFAP_stats_num_flows;     /* number of flows in stats list */
static LFAPNode* LFAP_gcoll_flow;    /* GARBAGE_COLELCTOR, remove flows */
static LFAPNode* LFAP_timer_flow;    /* timer callback pointer to the list */
static lfapui32_t LFAP_fun_update_interval;

/* max send queue size */
lfapui32_t LFAP_max_send_queue_size;

/* if connecting and at the end of server list, retry at beginning or not */
static lfapui16_t LFAP_retry_server_list;


/* CCE functions */
extern lfapui32_t LFAPGetCurTime (void);
extern lfapui32_t LFAPGetSysUptime (void);
extern void LFAPSynchronize (void);

extern void LFAPGetStatistics (lfapui32_t flow_id, LFAPFlowStats* flow_stats);

extern int LFAPSendVector (LFAPMsgHolder* vec,
                           lfapui16_t vec_len,
                           lfapui16_t total_len,
                           lfapui16_t lost_conn_count);
extern int LFAPConnect (lfapui32_t server_ip);
extern int LFAPDisconnect (lfapui32_t server_ip);

extern void LFAPSetSystemTimer (const LFAP_CB_ENTRYPOINT lfap_cb_entry,
                                lfapui32_t return_handle,
                                lfapui32_t time_val);
extern void lfap_init_encode();
extern void lfap_init_decode();

extern int LFAPGarbageCollector (void); /* GARBAGE_COLELCTOR */
extern int LFAPDoesFlowExist (lfapui32_t flow_id); /* GARBAGE_COLELCTOR */

/* SSR next 2 are for adding source AS when flow is sent instead of when 
 *     LFAP message is created. This is because we want to pay the
 *     CPU cost on the LFAP task not the flow set up task.
 */
extern lfapui32_t LFAPGetSrcAS (lfapui32_t src_addr); 
extern int lfap_send_ies;

typedef enum { High = 1, Low } MsgPri_t;

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPMsgPriority_                                                */
/*                                                                            */
/*  Return Priority of LFAP Message                                           */
/*                                                                            */
/******************************************************************************/
/* inline  */
static MsgPri_t LFAPMsgPriority_ (void *msg, int len)
{
    /*  High Priority Messages */
    switch (lfap_msg_get_opcode(msg)) {
      case LFAP_AR_OPCODE: 
        {
          LFAP_AR ar;
          initRcvdLfapAR(&ar, msg, len);
          if ( ar.command.rsvd_ui8.val == LFAP_RETURN_INDICATED_FLOWS) 
            break;
        }
      case LFAP_ARA_OPCODE:
      case LFAP_VR_OPCODE:
      case LFAP_VRA_OPCODE:
      case LFAP_CR_OPCODE:
      case LFAP_CAN_OPCODE:
      case LFAP_CRN_OPCODE:
      case LFAP_FER_OPCODE:
      case LFAP_DR_OPCODE:
        return High;
    }
    
    return Low;
}
/* nline */
static lfapui16_t addMsgToLfapSendQueue (void* buf, unsigned int len, int required)
{
  lfapui16_t status;
  
  MsgPri_t pri = LFAPMsgPriority_(buf, len);
  switch (pri) {
    case High:                
      status = addMsgToLfapHighSendQueue(buf, len, required);
      break;
    case Low:
    default:          
      status = addMsgToLfapRegSendQueue(buf, len, required);
  }  

  if (!status) {
    /* API Stats */
    if (LFAP_curr_stats) {

      LFAP_curr_stats->dropped_send_msgs++;
      /* PAC_REMOVE      if (LFAP_connected == LFAP_CONNECTION_ESTABLISHED) */
      if (LFAP_conn_state == LFAP_SESSION_ESTABLISHED)
        LFAP_curr_stats->dropped_send_msgs_conn++;

      switch (lfap_msg_get_opcode(buf)) {
        case LFAP_VR_OPCODE:
          LFAP_curr_stats->dropped_vr_msgs++;
          break;
        case LFAP_VRA_OPCODE:
          LFAP_curr_stats->dropped_vra_msgs++;
          break;
        case LFAP_FAR_OPCODE:
          LFAP_curr_stats->dropped_far_msgs++;
          break;
        case LFAP_FUN_OPCODE:
          LFAP_curr_stats->dropped_fun_msgs++;
          break;
        case LFAP_AR_OPCODE:
          LFAP_curr_stats->dropped_ar_msgs++;
          break;
        case LFAP_ARA_OPCODE:
          LFAP_curr_stats->dropped_ara_msgs++;
          break;
        case LFAP_CR_OPCODE:
          LFAP_curr_stats->dropped_cr_msgs++;
          break;
        case LFAP_CAN_OPCODE:
          LFAP_curr_stats->dropped_can_msgs++;
          break;
        case LFAP_CRN_OPCODE:
          LFAP_curr_stats->dropped_crn_msgs++;
          break;
        case LFAP_FER_OPCODE:
          LFAP_curr_stats->dropped_crn_msgs++;
          break;
        case LFAP_KA_OPCODE:
          LFAP_curr_stats->dropped_ka_msgs++;
          break;
        case LFAP_DR_OPCODE:
          LFAP_curr_stats->dropped_dr_msgs++;
          break;
      }
    }
  }
  return status;
}



/******************************************************************************/
/*                                                                            */
/*  Function: initializeLfapFlowMgrGlobals_                                   */
/*                                                                            */
/*  Initializes the global variables of the LFAP Flow Manager.                */
/*                                                                            */
/******************************************************************************/
static void initializeLfapFlowMgrGlobals_ (void)
{
  /* PAC_REMOVE    LFAP_connected = LFAP_CONNECTION_LOST;*/
  /* PAC_REMOVE    LFAP_trying_to_connect = LFAP_NOT_TRYING_TO_CONNECT;*/
    LFAP_conn_state = LFAP_INITIAL_STATE;
    LFAP_server = 0;
    LFAP_new_server = 0;

    LFAP_cce_id = 0;
    LFAP_fas_hi = 0;
    LFAP_fas_lo = 0;

    LFAP_stats = 0;
    LFAP_curr_stats = 0;

    LFAP_msgs_rcvd = 0;

    LFAP_connection_timer = LFAP_TIMER_EXPIRED;
    LFAP_lost_conn_timer = LFAP_TIMER_INACTIVE;
    LFAP_flow_update_timer = LFAP_TIMER_INACTIVE;
    LFAP_batch_interval_timer = LFAP_TIMER_INACTIVE;
    LFAP_garbage_collector_timer = LFAP_TIMER_EXPIRED; /* GARBAGE_COLELCTOR */

    LFAP_report_on_flows = LFAP_NOT_REPORTING_FLOWS;

    LFAP_send_vector_len = 0;
    LFAP_send_vector_entries = 0;

    LFAP_error.time = 0;
    LFAP_error.error = LFAP_NO_ERROR_OCCURRED;
    /* PAC_REMOVE    LFAP_conn_status = LFAP_CONNECTION_LOST;*/

    LFAP_gcoll_flow = 0; /* GARBAGE_COLELCTOR */
    LFAP_timer_flow = 0;
    LFAP_stats_num_flows = 0;
    LFAP_fun_update_interval = LFAP_DEFAULT_POLL_INTERVAL;

    LFAP_retry_server_list = LFAP_RETRY_IF_LIST_END;
    LFAP_send_message_len = 0;
    LFAP_send_message = 0;
    LFAP_send_message_node = 0;

}

/******************************************************************************/
/*                                                                            */
/*  Function: setLfapStatsType_                                               */
/*                                                                            */
/*  Sets the statistics type (delta or total) for what is reported back to    */
/*  the server in LFAP_FUN messages.                                          */
/*                                                                            */
/******************************************************************************/
static void setLfapStatsType_ (lfapui16_t stats_type)
{
    if (stats_type == LFAP_TOTAL_STATS) {
        LFAP_bytes_stats = LFAP_BYTES_TOTAL;
        LFAP_packets_stats = LFAP_PACKETS_TOTAL;
    }
    else {
        LFAP_bytes_stats = LFAP_BYTES_DELTA;
        LFAP_packets_stats = LFAP_PACKETS_DELTA;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: initializeLfapFlowStats_                                        */
/*                                                                            */
/*  Initializes all of the members of the LFAPFlowStats structure.            */
/*                                                                            */
/******************************************************************************/
static void initializeLfapFlowStats_ (LFAPFlowStats* flow_stats)
{
    flow_stats->type = LFAP_NO_STATS_AVAILABLE;
    flow_stats->byte_rcvd = 0;
    flow_stats->byte_sent = 0;

    flow_stats->packet_rcvd = 0;
    flow_stats->packet_sent = 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: initializeLfapFlowInfo                                          */
/*                                                                            */
/*  Initializes all of the members of the LFAPFlowInfo structure.             */
/*                                                                            */
/******************************************************************************/
void initializeLfapFlowInfo (LFAPFlowInfo* flow_info)
{
  int type;

    lfap_memset(flow_info, 0, sizeof(LFAPFlowInfo));
    flow_info->src_as = LFAP_AS_NUM_INVALID;
    flow_info->dest_as = LFAP_AS_NUM_INVALID;
    flow_info->cxn_priority = LFAP_QUALIFIER_INVALID;
    flow_info->checkpoint = LFAP_QUALIFIER_INVALID;
    flow_info->src_vlan_id = LFAP_NO_VLAN_ID;
    flow_info->dst_vlan_id = LFAP_NO_VLAN_ID;

    for( type = 1; type <= LFAP_COS_MAX_TYPES; type++) {
      flow_info->CoS_Arr[type][LFAP_COS_RECEIVED] = LFAP_COS_INVALID;
      flow_info->CoS_Arr[type][LFAP_COS_TRANSMITTED] = LFAP_COS_INVALID;     
    }

                /* PAC_REMOVE     for (i = 0; i < LFAP_MAX_ADDR_LEN; i++) { */
                /* PAC_REMOVE         flow_info->src_addr[i] = 0; */
                /* PAC_REMOVE         flow_info->dest_addr[i] = 0; */
                /* PAC_REMOVE         flow_info->src_cce_addr[i] = 0; */
                /* PAC_REMOVE     } */
                /* PAC_REMOVE  */
                /* PAC_REMOVE     flow_info->src_fam = 0; */
                /* PAC_REMOVE     flow_info->src_len = 0; */
                /* PAC_REMOVE  */
                /* PAC_REMOVE     flow_info->dest_fam = 0; */
                /* PAC_REMOVE     flow_info->dest_len = 0; */
                /* PAC_REMOVE  */
                /* PAC_REMOVE     flow_info->src_cce_fam = 0; */
                /* PAC_REMOVE     flow_info->src_cce_len = 0; */
                /* PAC_REMOVE  */
                /* PAC_REMOVE     flow_info->src_port = 0; */
                /* PAC_REMOVE     flow_info->src_port_type = 0; */
                /* PAC_REMOVE  */
                /* PAC_REMOVE     for (i = 0; i < LFAP_MAX_PROTOCOL_LEN; i++) */
                /* PAC_REMOVE         flow_info->protocol[i] = 0; */
                /* PAC_REMOVE  */
                /* PAC_REMOVE     flow_info->client_data = 0; */
                /* PAC_REMOVE     flow_info->client_data_len = 0; */
    /* PAC_REMOVE     flow_info->ingress_port = 0; */
    /* PAC_REMOVE     flow_info->egress_port = 0; */

}

/******************************************************************************/
/*                                                                            */
/*  Function: initializeLfapApiStats_                                         */
/*                                                                            */
/*  Initializes all of the members of the LFAPApiStats structure.             */
/*                                                                            */
/******************************************************************************/
static void initializeLfapApiStats_ (LFAPApiStats* api_stats,
                                     lfapui16_t num_servers,
                                     lfapui32_t this_server)
{

  memset(api_stats, 0, sizeof(LFAPApiStats));
  api_stats->num_servers = num_servers;
  api_stats->this_server = this_server;
  api_stats->this_server_in_list = 1;

#if 0
    api_stats->up_time = 0;
    api_stats->conn_time = 0;

    api_stats->num_servers = num_servers;
    api_stats->this_server = this_server;
    api_stats->this_server_in_list = 1;

    api_stats->retry_conn_success = 0;
    api_stats->retry_conn_failure = 0;

    api_stats->total_bytes_sent = 0;
    api_stats->total_bytes_rcvd = 0;

    api_stats->msgs_sent_success = 0;
    api_stats->msgs_sent_failure = 0;

    api_stats->msgs_rcvd_corrupt = 0;
    api_stats->msgs_rcvd_error = 0;
    api_stats->msgs_rcvd_success = 0;

    api_stats->msgs_send_buf = 0;
    api_stats->msgs_rcvd_buf = 0;

    api_stats->msg_vr_sent = 0;
    api_stats->msg_vra_rcvd = 0;
    api_stats->msg_ar_sent = 0;
    api_stats->msg_ar_rcvd = 0;
    api_stats->msg_ara_sent = 0;
    api_stats->msg_ara_rcvd = 0;
    api_stats->msg_far_sent = 0;
    api_stats->msg_fun_sent = 0;
    api_stats->msg_unk_rcvd = 0;

    api_stats->dropped_send_msgs = 0;
    api_stats->dropped_send_msgs_conn = 0;
    api_stats->dropped_vr_msgs = 0;
    api_stats->dropped_ar_msgs = 0;
    api_stats->dropped_ara_msgs = 0;
    api_stats->dropped_far_msgs = 0;
    api_stats->dropped_fun_i_msgs = 0;
    api_stats->dropped_fun_msgs = 0;
    api_stats->dropped_rcvd_msgs = 0;

    api_stats->lost_bytes_sent = 0;
    api_stats->lost_packets_sent = 0;
    api_stats->lost_bytes_rcvd = 0;
    api_stats->lost_packets_rcvd = 0;

    api_stats->flow_setups = 0;
    api_stats->flow_teardowns = 0;
    api_stats->active_flows = 0;

    api_stats->peak_msgs_send_buf = 0;
    api_stats->peak_active_flows = 0;
#endif
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPConnectionTimerExpired_                                     */
/*                                                                            */
/*  Timer callback function that indicates the LFAP connection timer has      */
/*  expired.                                                                  */
/*                                                                            */
/******************************************************************************/
static int LFAPConnectionTimerExpired_ (void* context)
{
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_TIMERS)
        fprintf(stderr, "LFAPConnectionTimerExpired_\n");
    #endif

    LFAP_connection_timer = LFAP_TIMER_EXPIRED;
    return 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPFlowUpdateTimerExpired_                                     */
/*                                                                            */
/*  Timer callback function that indicates the LFAP Flow update timer has     */
/*  expired.                                                                  */
/*                                                                            */
/******************************************************************************/
int LFAPFlowUpdateTimerExpired_ (void* context)
{
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_TIMERS)
        fprintf(stderr, "LFAPFlowUpdateTimerExpired_\n");
    #endif

    LFAP_flow_update_timer = LFAP_TIMER_EXPIRED;
    return 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPLostConnTimerExpired_                                       */
/*                                                                            */
/*  Timer callback function that indicates the LFAP lost connection timer has */
/*  expired.                                                                  */
/*                                                                            */
/******************************************************************************/
static int LFAPLostConnTimerExpired_ (void* context)
{
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_TIMERS)
        fprintf(stderr, "LFAPLostConnTimerExpired_\n");
    #endif

    LFAP_lost_conn_timer = LFAP_TIMER_EXPIRED;
    return 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPBatchIntervalExpired_                                       */
/*                                                                            */
/*  Timer callback function that indicates the LFAP batch interval timer has  */
/*  expired.                                                                  */
/*                                                                            */
/******************************************************************************/
static int LFAPBatchIntervalExpired_ (void* context)
{
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_TIMERS)
        fprintf(stderr, " LFAPBatchIntervalExpired_...\n");
    #endif

    LFAP_batch_interval_timer = LFAP_TIMER_EXPIRED;
    return 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGarbageCollectorIntervalExpired_                            */
/*                                                                            */
/*  GARBAGE_COLELCTOR timer callback function that indicates the LFAP garbage */
/*  collector timer has expired.                                              */
/*                                                                            */
/******************************************************************************/
static int LFAPGarbageCollectorIntervalExpired_ (void* context)
{
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_TIMERS)
            fprintf(stderr, " LFAPGarbageCollectorIntervalExpired_...\n");
    #endif

    LFAP_garbage_collector_timer = LFAP_TIMER_EXPIRED;
    return 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: setLfapApiStats_                                                */
/*                                                                            */
/*  Sets the LFAP_curr_stats pointer to the appropriate LFAPApiStats          */
/*  structure in the LFAP_stats array according to the server.                */
/*                                                                            */
/******************************************************************************/
static void setLfapApiStats_ (lfapui32_t server)
{
    lfapui16_t stat = 0;
    lfapui16_t num_stats = 0;
    LFAPApiStats* prev_stats = LFAP_curr_stats;

    LFAP_curr_stats = 0;

    if (LFAP_stats) {
        num_stats = LFAP_stats->num_servers;
        while (stat < num_stats) {
            if (LFAP_stats[stat].this_server == server) {
                LFAP_curr_stats = &LFAP_stats[stat];
                if (prev_stats && (LFAP_curr_stats != prev_stats)) {
                    LFAP_curr_stats->active_flows = prev_stats->active_flows;
                    prev_stats->active_flows = 0;
                    prev_stats->msgs_send_buf = 0;
                    prev_stats->msgs_rcvd_buf = 0;
                }
                break;
            }
            stat++;
        }
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: serverInLfapApiStats_                                           */
/*                                                                            */
/*  Returns 1 if the server is represented in the LFAP_stats, else returns 0. */
/*                                                                            */
/******************************************************************************/
static int serverInLfapApiStats_ (lfapui32_t server)
{
    lfapui16_t stat = 0;
    lfapui16_t server_in_stats = 0;
    lfapui16_t num_stats = LFAP_stats->num_servers;

    while (stat < num_stats) {
        if (LFAP_stats[stat].this_server == server) {
            server_in_stats = 1;
            break;
        }
        stat++;
    }

    return server_in_stats;
}

/******************************************************************************/
/*                                                                            */
/*  Function: copyLfapApiStats_                                               */
/*                                                                            */
/*  Copies the statistics from the 'from_stats' LFAPApiStats to the           */
/*  'to_stats' LFAPApiStats.                                                  */
/*                                                                            */
/******************************************************************************/
static void copyLfapApiStats_ (LFAPApiStats* to_stats, LFAPApiStats* from_stats)
{
    lfapui32_t server = 0;
    LFAPListIterator iter;
    lfapui16_t server_in_list = 0;
    initLfapIterList(&iter, &LFAP_server_list);
    while ((server = (lfapui32_t) incLfapIter(&iter))) {
        if (server == from_stats->this_server) {
            server_in_list = 1;
            break;
        }
    }

    memcpy(to_stats, from_stats, sizeof(LFAPApiStats));
    to_stats->this_server_in_list = server_in_list;
    
#if 0
    to_stats->up_time = from_stats->up_time;
    to_stats->conn_time = from_stats->conn_time;

    to_stats->num_servers = from_stats->num_servers;
    to_stats->this_server = from_stats->this_server;
    to_stats->this_server_in_list = server_in_list;

    to_stats->retry_conn_success = from_stats->retry_conn_success;
    to_stats->retry_conn_failure = from_stats->retry_conn_failure;

    to_stats->total_bytes_sent = from_stats->total_bytes_sent;
    to_stats->total_bytes_rcvd =from_stats->total_bytes_rcvd;

    to_stats->msgs_sent_success = from_stats->msgs_sent_success;
    to_stats->msgs_sent_failure = from_stats->msgs_sent_failure;

    to_stats->msgs_rcvd_corrupt = from_stats->msgs_rcvd_corrupt;
    to_stats->msgs_rcvd_error = from_stats->msgs_rcvd_error;
    to_stats->msgs_rcvd_success = from_stats->msgs_rcvd_success;

    to_stats->msgs_send_buf = from_stats->msgs_send_buf;
    to_stats->msgs_rcvd_buf = from_stats->msgs_rcvd_buf;

    to_stats->msg_vr_sent = from_stats->msg_vr_sent;
    to_stats->msg_vra_rcvd = from_stats->msg_vra_rcvd;
    to_stats->msg_ar_sent = from_stats->msg_ar_sent;
    to_stats->msg_ar_rcvd = from_stats->msg_ar_rcvd;
    to_stats->msg_ara_sent = from_stats->msg_ara_sent;
    to_stats->msg_ara_rcvd = from_stats->msg_ara_rcvd;
    to_stats->msg_far_sent = from_stats->msg_far_sent;
    to_stats->msg_fun_sent = from_stats->msg_fun_sent;
    to_stats->msg_unk_rcvd = from_stats->msg_unk_rcvd;

    to_stats->dropped_send_msgs = from_stats->dropped_send_msgs;
    to_stats->dropped_send_msgs_conn = from_stats->dropped_send_msgs_conn;
    to_stats->dropped_vr_msgs = from_stats->dropped_vr_msgs;
    to_stats->dropped_ar_msgs = from_stats->dropped_ar_msgs;
    to_stats->dropped_ara_msgs = from_stats->dropped_ara_msgs;
    to_stats->dropped_far_msgs = from_stats->dropped_far_msgs;
    to_stats->dropped_fun_i_msgs = from_stats->dropped_fun_i_msgs;
    to_stats->dropped_fun_msgs = from_stats->dropped_fun_msgs;
    to_stats->dropped_rcvd_msgs = from_stats->dropped_rcvd_msgs;

    to_stats->lost_bytes_sent = from_stats->lost_bytes_sent;
    to_stats->lost_packets_sent = from_stats->lost_packets_sent;
    to_stats->lost_bytes_rcvd = from_stats->lost_bytes_rcvd;
    to_stats->lost_packets_rcvd = from_stats->lost_packets_rcvd;

    to_stats->flow_setups = from_stats->flow_setups;
    to_stats->flow_teardowns = from_stats->flow_teardowns;
    to_stats->active_flows = from_stats->active_flows;

    to_stats->peak_msgs_send_buf = from_stats->peak_msgs_send_buf;
    to_stats->peak_active_flows = from_stats->peak_active_flows;
#endif
}

/******************************************************************************/
/*                                                                            */
/*  Function: updateLfapApiStats_                                             */
/*                                                                            */
/*  Updates the array of LFAPApiStats structures.                             */
/*                                                                            */
/******************************************************************************/
static void updateLfapApiStats_ (void)
{
    lfapui16_t stat = 0;
    LFAPList new_servers;
    LFAPListIterator iter;
    lfapui32_t server = 0;
    LFAPApiStats* old_stats;
    lfapui16_t server_in_list;
    lfapui16_t num_servers = 0;
    lfapui16_t curr_num_servers;

    num_servers = lfapListEntries(&LFAP_server_list);

    if (!LFAP_stats) {
        LFAP_stats = (LFAPApiStats*)
                     lfap_malloc(num_servers * SIZEOF_LFAP_API_STATS);

        initLfapIterList(&iter, &LFAP_server_list);
        while ((server = (lfapui32_t) incLfapIter(&iter)))
            initializeLfapApiStats_(&LFAP_stats[stat++], num_servers, server);
    }
    else {
        LFAP_curr_stats = 0;
        curr_num_servers = LFAP_stats->num_servers;
        initLfapList(&new_servers);

        initLfapIterList(&iter, &LFAP_server_list);
        while ((server = (lfapui32_t) incLfapIter(&iter))) {
            if (!serverInLfapApiStats_(server))
                lfapListAppend(&new_servers, (void*) server);
        }

        if ((num_servers = lfapListEntries(&new_servers))) {
            num_servers += curr_num_servers;
            old_stats = LFAP_stats;
            LFAP_stats = (LFAPApiStats*)
                         lfap_malloc(num_servers*SIZEOF_LFAP_API_STATS);
            while (stat < curr_num_servers) {
                old_stats[stat].num_servers = num_servers;
                copyLfapApiStats_(&LFAP_stats[stat], &old_stats[stat]);
                stat++;
            }
            lfap_free(old_stats);
            initLfapIterList(&iter, &new_servers);
            while ((server = (lfapui32_t) incLfapIter(&iter)))
                initializeLfapApiStats_(&LFAP_stats[stat++],num_servers,server);
            clearLfapList(&new_servers);
        }
        else {
            /* need to determine if any servers have been removed */
            while (stat < curr_num_servers) {
                server_in_list = 0;

                initLfapIterList(&iter, &LFAP_server_list);
                while ((server = (lfapui32_t) incLfapIter(&iter))) {
                    if (server == LFAP_stats[stat].this_server) {
                        server_in_list = 1;
                        break;
                    }
                }
                LFAP_stats[stat].this_server_in_list = server_in_list;
                stat++;
            }
        }
    }

    /* setting LFAP_curr_stats pointer to the appropriate stats */
    if (LFAP_server) setLfapApiStats_(LFAP_server);
}

/******************************************************************************/
/*                                                                            */
/*  Function: setLfapFUNStats_                                                */
/*                                                                            */
/*  Sets the statistics of a LFAP_FUN message.                                */
/*                                                                            */
/******************************************************************************/
static void setLfapFUNStats_ (LFAP_FUN* fun, LFAPFlowStats* flow_stats)
{
    if (flow_stats) {
        if (flow_stats->type & LFAP_BYTES_STATS) {
            setByteCountLfapFUN(fun, LFAP_bytes_stats,
                                flow_stats->byte_rcvd,
                                flow_stats->byte_sent);
        }
        if (flow_stats->type & LFAP_PACKETS_STATS) {
            setPacketCountLfapFUN(fun, LFAP_packets_stats,
                                  flow_stats->packet_rcvd,
                                  flow_stats->packet_sent);
        }
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: createLfapFUNMsg_                                               */
/*                                                                            */
/*  Creates a single record of the LFAP_FUN message to be sent to the server  */
/*  for the flow flow_id.  Returns 1 if successful, 0 if unsuccessful.        */
/*                                                                            */
/******************************************************************************/
static int createLfapFUNMsg_ (LFAP_FUN* fun, lfapui32_t flow_id)
{
    int fun_successful = 1;
    LFAPFlowStats flow_stats;
    initializeLfapFlowStats_(&flow_stats);

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "createLfapFUNMsg_: %lu\n", flow_id);
    #endif

    LFAPGetStatistics(flow_id, &flow_stats);
    if (flow_stats.type != LFAP_NO_STATS_AVAILABLE) {
        setFlowIdLfapFUN(fun, LFAP_fas_hi, LFAP_fas_lo, flow_id);
        setLfapFUNStats_(fun, &flow_stats);
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(fun, 0, LFAP_SEND);
        #endif
        encodeRecordLfapFUN(fun);
    }
    else
        fun_successful = 0;

    return fun_successful;
}

/******************************************************************************/
/*                                                                            */
/*  Function: removeFlowLfapStatsList_                                        */
/*                                                                            */
/*  Removes the flow_id from the LFAP_stats_list.                             */
/*                                                                            */
/******************************************************************************/
/* GARBAGE_COLELCTOR - not needed since the algorithm was changed to use a garbage
       collection algorithm. note - the previous alogrithm searched
       the entire linked list looking for the flow to delete. This
       was too slow.
static void removeFlowLfapStatsList_ (lfapui32_t flow_id)
{
    LFAPNode* prev;
    LFAPNode* flow;
    LFAPNode* next;

    if ((prev = previousLfapNode(&LFAP_stats_list, (const void*) flow_id))) {
        flow = nextLfapNode(prev);
        next = nextLfapNode(flow);
        if (LFAP_timer_flow == flow)
            LFAP_timer_flow = next;

        removeNextLfapNode(&LFAP_stats_list, prev);
        LFAP_stats_num_flows = lfapListEntries(&LFAP_stats_list);
    }
    #ifdef CS_DEBUG
    else if (lfap_debug & LFAP_DBG_WARNINGS) {
        fprintf(stderr, " LFAP Warning: Could not find the flow in");
        fprintf(stderr, " the timer list:  %lu\n\n", flow_id);
    }
    #endif
} */

/******************************************************************************/
/*                                                                            */
/*  Function: setLfapFunUpdateInterval_                                       */
/*                                                                            */
/*  Sets the global LFAP_fun_update_interval variable to the appropriate      */
/*  time.                                                                     */
/*                                                                            */
/******************************************************************************/
static void setLfapFunUpdateInterval_ (void)
{
    LFAP_stats_num_flows = lfapListEntries(&LFAP_stats_list);
    if (!LFAP_stats_num_flows) return;

    if (LFAP_stats_num_flows%LFAP_batch_size)
        LFAP_fun_update_interval =
        LFAP_poll_interval/(((int)LFAP_stats_num_flows/LFAP_batch_size)+1);
    else
        LFAP_fun_update_interval =
        LFAP_poll_interval/((int) LFAP_stats_num_flows/LFAP_batch_size);

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "setLfapFunUpdateInterval_(), setting to: %lu\n",
                         LFAP_fun_update_interval);
    #endif
}

/******************************************************************************/
/*                                                                            */
/*  Function: setLfapCheckpointTimer_                                         */
/*                                                                            */
/*  Adds the flow_id into the LFAP_stats_list, thereby having the flow be     */
/*  reported on in LFAP_FUN messages.                                         */
/*                                                                            */
/******************************************************************************/
static void setLfapCheckpointTimer_ (lfapui32_t flow_id)
{
    const LFAP_CB_ENTRYPOINT cb_entry = LFAPFlowUpdateTimerExpired_;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "SetCheckpointTimer() for flow: %lu for %lu\n",
                         flow_id, LFAP_poll_interval);
    #endif

    lfapListAppend(&LFAP_stats_list, (void*) flow_id);
    setLfapFunUpdateInterval_();

    if (LFAP_flow_update_timer == LFAP_TIMER_INACTIVE) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_TIMERS)
            fprintf(stderr, "  SetSystemTimer() for: %lu\n",
                               LFAP_fun_update_interval);
        #endif
        LFAPSetSystemTimer(cb_entry, 0, LFAP_fun_update_interval);
        LFAP_flow_update_timer = LFAP_TIMER_WAITING;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: deleteLfapCheckpointTimer_                                      */
/*                                                                            */
/*  Removes the flow_id from the LFAP_stats_list, thereby having the flow not */
/*  be reported on in LFAP_FUN messages anymore.                              */
/*                                                                            */
/*******************************************************************************
 GARBAGE_COLELCTOR - not needed since the algorithm was changed to use a garbage
       collection algorithm. note - the previous alogrithm searched
       the entire linked list looking for the flow to delete. This
       was too slow.
static void deleteLfapCheckpointTimer_ (lfapui32_t flow_id)
{
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr,"deleteLfapCheckpointTimer_() for flow: %lu \n",flow_id);
    #endif

    removeFlowLfapStatsList_(flow_id);
    setLfapFunUpdateInterval_();
} */

/******************************************************************************/
/*                                                                            */
/*  Function: sendLfapFlowUpdates_                                            */
/*                                                                            */
/*  Callback for the system timer, for reporting on flows.                    */
/*                                                                            */
/******************************************************************************/
static void sendLfapFlowUpdates_ (void)
{
    LFAP_FUN fun;
    LFAPNode* tail;
    lfapui32_t flow_id;
    int flow_count = 0;
    const LFAP_CB_ENTRYPOINT cb_entry = LFAPFlowUpdateTimerExpired_;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_TIMERS) {
        fprintf(stderr, "sendLfapFlowUpdates_() ....\n");
        fprintf(stderr, "    LFAP_batch_size       = %d\n", LFAP_batch_size);
        fprintf(stderr,"    LFAP_stats_num_flows  = %d\n",LFAP_stats_num_flows);
    }
    #endif

    /* PAC_REMOVE    if (LFAP_connected == LFAP_CONNECTION_LOST) {*/
    if ( LFAP_conn_state != LFAP_SESSION_ESTABLISHED ) {
        /* The FAS is down, don't bother asking the CCE for stats,
           because if the CCE is keeping delta statistics, the data
           will be lost.  Instead wait for the next timer to see
           if the FAS is up again */

        /* reset timer */
        LFAPSetSystemTimer(cb_entry, 0, LFAP_fun_update_interval);
        LFAP_flow_update_timer = LFAP_TIMER_WAITING;
    }
    else if (LFAP_stats_num_flows) {
        /* reset timer */
        LFAPSetSystemTimer(cb_entry, 0, LFAP_fun_update_interval);
        LFAP_flow_update_timer = LFAP_TIMER_WAITING;

        /* initialize the FUN msg */
        initLfapFUN(&fun);
        setTimeLfapFUN(&fun, LFAPGetSysUptime());
        setStateLfapFUN(&fun, LFAP_ACTIVE);
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(&fun, 0, LFAP_SEND);
        #endif
        encodeFixedLfapFUN(&fun);

        tail = tailLfapNode(&LFAP_stats_list);

        if (!LFAP_timer_flow || (LFAP_timer_flow == tail))
            LFAP_timer_flow = firstLfapNode(&LFAP_stats_list);

        while (flow_count < LFAP_batch_size) {
            if (LFAP_timer_flow == tail) {
                LFAP_timer_flow = firstLfapNode(&LFAP_stats_list);
                break;
            }

            flow_id = (lfapui32_t) lfapNodeInfo(LFAP_timer_flow);
            /* GARBAGE_COLELCTOR - since flows are deleted from the hardware but
             *        sre not immediately deleted from the stats list
             *        (the garabage collector does the actual delete)
             *        we need to make sure the flow is still valid befire
             *        we try to get stats on it.
             */
            if (LFAPDoesFlowExist(flow_id)) {
                #ifdef CS_DEBUG
                if (lfap_debug & LFAP_DBG_FUNCTIONS) {
                    fprintf(stderr, "  Flow %lu exists, ", flow_id);
                    fprintf(stderr, "going to send a FUN for it\n");
                }
                #endif
 
                if (createLfapFUNMsg_(&fun, flow_id))
                    flow_count++;
            }

            LFAP_timer_flow = nextLfapNode(LFAP_timer_flow);
        }

        encodeCompletedLfapFUN(&fun);
        if (flow_count) {
            if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&fun), getLfapMsgLength(&fun), LFAP_OPT_MSG))
                freeLfapMsg(&fun);
        }
        else
            freeLfapMsg(&fun);
    }
    else {
        LFAP_timer_flow = 0;
        LFAP_flow_update_timer = LFAP_TIMER_INACTIVE;

        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_FUNCTIONS)
            fprintf(stderr, "   NOT SENDING A FUN, no FLOWS\n");
        #endif
    }

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_TIMERS) {
        if (LFAP_stats_num_flows)
            fprintf(stderr, "  LFAPSetSystemTimer() for: %lu\n",
                               LFAP_fun_update_interval);
    }
    #endif
}

/******************************************************************************/
/*                                                                            */
/*  Function: reportLfapFlows_                                                */
/*                                                                            */
/*  Sends up statistics for a given set of flows.  The server has asked for   */
/*  reports on flows indicated in the AR message.                             */
/*                                                                            */
/******************************************************************************/
static lfapui16_t reportLfapFlows_ (void)
{
    LFAPIE ie;
    LFAP_FUN fun;
    lfapui32_t flow_id;
    int flow_count = 0;
    lfapui16_t reporting = LFAP_REPORT_INDICATED_FLOWS;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, "reportLfapFlows_: ");
        if (LFAP_report_on_flows == LFAP_REPORT_INDICATED_FLOWS)
            fprintf(stderr, "INDICATED flows\n");
        else
            fprintf(stderr, "NO flows????\n");
    }
    #endif

    initLfapFUN(&fun);
    setTimeLfapFUN(&fun, LFAPGetSysUptime());
    setStateLfapFUN(&fun, LFAP_ACTIVE);
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_MESSAGES)
        PRINT_LFAP_MESSAGE(&fun, 0, LFAP_SEND);
    #endif
    encodeFixedLfapFUN(&fun);

    while (flow_count < LFAP_batch_size) {
        if (verifyLfapAR(&LFAP_report_flows_ar) &&
        (LFAP_report_flows_ar.command.rsvd_ui8.val == LFAP_RETURN_INDICATED_FLOWS)) {
            flow_id = LFAP_report_flows_ar.flow_id.fid_ie.cce_id;
            if (!createLfapFUNMsg_(&fun, flow_id)) {
                setFlowIdLfapARA(&LFAP_report_flows_ara, LFAP_fas_hi,
                                 LFAP_fas_lo, flow_id);
                setFailCodeLfapARA(&LFAP_report_flows_ara, LFAP_NO_SUCH_FLOW);
                #ifdef CS_DEBUG
                if (lfap_debug & LFAP_DBG_MESSAGES)
                    PRINT_LFAP_MESSAGE(&LFAP_report_flows_ara, 0, LFAP_SEND);
                #endif
                encodeRecordLfapARA(&LFAP_report_flows_ara);
            }
            else
                flow_count++;
        }
        else {
            setLfapMsgStatus(&LFAP_report_flows_ara, LFAP_MSG_CORRUPTED);
            reporting = LFAP_NOT_REPORTING_FLOWS;
            break;
        }

        if (lfap_next_record(&LFAP_report_flows_ar.lfap_msg) == LFAP_ERROR) {
            reporting = LFAP_NOT_REPORTING_FLOWS;
            break;
        }

        lfap_memset(&ie, 0, LFAP_IE_SIZE);
        resetLfapAR(&LFAP_report_flows_ar, LFAP_NO_RESET_FLAGS);

        while (lfap_get_IE(&LFAP_report_flows_ar.lfap_msg, &ie) != LFAP_ERROR)
            setComponentLfapAR(&LFAP_report_flows_ar, &ie);

        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(&LFAP_report_flows_ar, 0, LFAP_RECV);
        #endif
    }

    encodeCompletedLfapFUN(&fun);
    if (flow_count) {
        if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&fun), getLfapMsgLength(&fun), LFAP_OPT_MSG))
            freeLfapMsg(&fun);
    }
    else
        freeLfapMsg(&fun);

    /* finished? */
    if (reporting == LFAP_NOT_REPORTING_FLOWS) {
        freeLfapMsg(&LFAP_report_flows_ar);

        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(&LFAP_report_flows_ara, 0, LFAP_SEND);
        #endif
        encodeCompletedLfapARA(&LFAP_report_flows_ara);
        if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&LFAP_report_flows_ara),
                                   getLfapMsgLength(&LFAP_report_flows_ara),
                                   LFAP_REQ_MSG)) {
            freeLfapMsg(&LFAP_report_flows_ara);
        }

        LFAP_report_on_flows = LFAP_NOT_REPORTING_FLOWS;
    }

    return reporting;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSendMessage_                                                */
/*                                                                            */
/*  Sends the message to the server (by calling the CCE's LFAPSendMessage     */
/*  function).  Also cleans up the message's memory buffer.  Returns          */
/*  LFAP_SEND_SUCCESS if successfully sent message, LFAP_SEND_FAILURE if the  */
/*  message was not sent. Clear causes any partially sent messages to be      */
/*  cleared.                                                                  */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
/* PAC_REMOVEstatic int LFAPSendMessage_ (void* msg, lfapui16_t msg_len) */
static int LFAPSendMessage_ (int clear) {

    int status;

    if (lfap_debug_fas)
      status = LFAPSendMessage(LFAP_send_message, LFAP_send_message_len, clear);
    else
      status = LFAP_SEND_SUCCESS;

    if (LFAP_curr_stats) {
        if (status == LFAP_SEND_SUCCESS && !clear) {
            LFAP_curr_stats->msgs_sent_success++;
            LFAP_curr_stats->total_bytes_sent += LFAP_send_message_len;
            switch(lfap_msg_get_opcode(LFAP_send_message)) {
                case LFAP_VR_OPCODE:
                    LFAP_curr_stats->msg_vr_sent++;
                    break;
                case LFAP_AR_OPCODE:
                    LFAP_curr_stats->msg_ar_sent++;
                    break;
                case LFAP_ARA_OPCODE:
                    LFAP_curr_stats->msg_ara_sent++;
                    break;
                case LFAP_FAR_OPCODE:
                    LFAP_curr_stats->msg_far_sent++;
                    break;
                case LFAP_FUN_OPCODE:
                    LFAP_curr_stats->msg_fun_sent++;
                    break;
                case LFAP_CR_OPCODE:
                    LFAP_curr_stats->msg_cr_sent++;
                    break;
            }
        }
        else
            LFAP_curr_stats->msgs_sent_failure++;
    }

    if ( status == LFAP_SEND_SUCCESS )
    {
      LFAPNode* node;
      LFAPSendMsg* msg;

      node = LFAP_send_message_node;
      msg = (LFAPSendMsg*) lfapNodeInfo(node);
      lfap_free(msg->buf);
      msg->buf = 0; msg->len = 0;
      returnLfapSendMsg(node);
      LFAP_send_message = 0;
      LFAP_send_message_len = 0;
      LFAP_send_message_node = 0;
    }
    
    return status;
}

/******************************************************************************/
/*                                                                            */
/*  Function: requestLfapFlowPrefix_                                          */
/*                                                                            */
/*  Requests the flow prefix from the server if one is not already set.       */
/*                                                                            */
/******************************************************************************/
static void requestLfapFlowPrefix_ (void)
{
    LFAP_AR ar;

    if (!(LFAP_fas_hi || LFAP_fas_lo)) {
        initLfapARcommand(&ar, LFAP_RETURN_FLOW_PREFIX);
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(&ar, 0, LFAP_SEND);
        #endif
        encodeCompletedLfapAR(&ar);
        if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&ar), getLfapMsgLength(&ar), LFAP_REQ_MSG))
            freeLfapMsg(&ar);
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: updateLfapFlowPrefix_                                           */
/*                                                                            */
/*  Places the server flow prefix into the global LFAP_fas_hi and LFAP_fas_lo */
/*  variables.                                                                */
/*                                                                            */
/******************************************************************************/
void updateLfapFlowPrefix_ (lfapui32_t fas_id_hi, lfapui32_t fas_id_lo)
{
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, "updateLfapFlowPrefix_:\n");
        fprintf(stderr, "    old: %lu %lu\n", LFAP_fas_hi, LFAP_fas_lo);
        fprintf(stderr, "    new: %lu %lu\n", fas_id_hi, fas_id_lo);
    }
    #endif

    if (!fas_id_hi && !fas_id_lo) {
        LFAP_error.error = LFAP_INVALID_FLOW_PREFIX_RCVD;
        LFAP_error.time = LFAPGetSysUptime();
    }

    LFAP_fas_hi = fas_id_hi;
    LFAP_fas_lo = fas_id_lo;
}

/******************************************************************************/
/*                                                                            */
/*  Function: updateLfapServerList_                                           */
/*                                                                            */
/*  Updates the list of servers with a new list of server IP addresses.  If   */
/*  there are any addresses in the list currently, they are replaced, not     */
/*  added to.  The server IPs are in the form of lfapui32_t.  Returns 1 if    */
/*  the list has changed, returns 0 if the list is unchanged.                 */
/*                                                                            */
/******************************************************************************/
static int updateLfapServerList_ (lfapui32_t* server_list,
                                  lfapui16_t server_list_len)
{
    int list_changed = 0;
    lfapui16_t list_i = 0;
    LFAPListIterator iter;

    if (lfapListEntries(&LFAP_server_list) == server_list_len) {
        initLfapIterList(&iter, &LFAP_server_list);
        for (list_i = 0; list_i < server_list_len; list_i++) {
            if (((lfapui32_t)incLfapIter(&iter)) != server_list[list_i]) {
                list_changed = 1;
                break;
            }
        }
    }
    else {
        list_changed = 1;
    }

    if (list_changed) {
        clearLfapList(&LFAP_server_list);
        for (list_i = 0; list_i < server_list_len; list_i++)
            lfapListAppend(&LFAP_server_list, (void*) server_list[list_i]);
    }

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "updateLfapServerList_  changed: %d\n", list_changed);
    #endif

    return list_changed;
}

/******************************************************************************/
/*                                                                            */
/*  Function: updateLfapServer_                                               */
/*                                                                            */
/*  Updates the list of servers, given a single server address.  The new      */
/*  server address may or may not be in the list of servers.  If not, it is   */
/*  added.                                                                    */
/*                                                                            */
/******************************************************************************/
static void updateLfapServer_ (lfapui32_t new_server)
{
    LFAP_server = new_server;

    if (!lfapListFindRef(&LFAP_server_list, (void*) new_server)) {
        lfapListAppend(&LFAP_server_list, (void*) new_server);
        updateLfapApiStats_();
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: indicateLfapVersion_                                            */
/*                                                                            */
/*  Sends the CCE's LFAP version request message to the server.               */
/*                                                                            */
/******************************************************************************/
static void indicateLfapVersion_ (void)
{
    LFAP_VR vr;
    
    initLfapVR(&vr);
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_MESSAGES)
        PRINT_LFAP_MESSAGE(&vr, 0, LFAP_SEND);
    #endif
    encodeCompletedLfapVR(&vr);
    if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&vr), getLfapMsgLength(&vr), LFAP_REQ_MSG))
      freeLfapMsg(&vr);
    else
      LFAP_conn_state = LFAP_PROTOCOL_NEGOTIATION;
}


/******************************************************************************/
/*                                                                            */
/*  Function: removeLfapVersion_                                              */
/*                                                                            */
/*  Removes the CCE's LFAP version request message from the high send queue.  */
/*                                                                            */
/******************************************************************************/
/* PAC_REMOVE static void removeLfapVersion_ (void) */
/* PAC_REMOVE { */
/* PAC_REMOVE     LFAPNode* node; */
/* PAC_REMOVE     LFAPSendMsg* msg; */
/* PAC_REMOVE     int entries = msgsInLfapHighSendQueue(); */
/* PAC_REMOVE  */
/* PAC_REMOVE     while (entries) { */
/* PAC_REMOVE         if ((node = getMsgFromLfapHighSendQueue())) { */
/* PAC_REMOVE             msg = (LFAPSendMsg*) lfapNodeInfo(node); */
/* PAC_REMOVE             if (lfap_msg_get_opcode(msg->buf) == LFAP_VR_OPCODE) */
/* PAC_REMOVE                 lfap_free(msg->buf); */
/* PAC_REMOVE             PAC_REMOVE else */
/* PAC_REMOVE                 addMsgToLfapHighSendQueue(msg->buf, msg->len, LFAP_REQ_MSG); */
/* PAC_REMOVE             msg->buf = 0; msg->len = 0; */
/* PAC_REMOVE             returnLfapSendMsg(node); */
/* PAC_REMOVE         } */
/* PAC_REMOVE         entries--; */
/* PAC_REMOVE     } */
/* PAC_REMOVE } */

/******************************************************************************/
/*                                                                            */
/*  Function: removeLfapMsgs   _                                              */
/*                                                                            */
/*  Removes the all the messages from the high send and receive queues.       */
/*  Since all the high priority messages have to do with session              */
/*  establishment, we need to ger rid of them or we'll mess up the next       */
/*  session negotiation. The only one we could have used was the ARA          */
/*  flow prefix. No big deal, we'll get it next time and this keeps the       */
/*  code nice and simple.                                                     */
/*                                                                            */
/******************************************************************************/
static void removeLfapMsgs_ (void)
{
    LFAPNode* node;
    LFAPSendMsg* smsg;
    LFAPRcvdMsg* rmsg;
    int entries;

    entries = msgsInLfapHighSendQueue();
    while (entries) {
        if ((node = getMsgFromLfapHighSendQueue())) {
            smsg = (LFAPSendMsg*) lfapNodeInfo(node);
            lfap_free(smsg->buf);

            smsg->buf = 0; smsg->len = 0;
            returnLfapSendMsg(node);
        }
        entries--;
    }

    /* Now clear any partially sent messages */
    if ( LFAP_send_message != 0 )
      LFAPSendMessage_(CLEAR);

    entries = msgsInLfapHighRcvdQueue();
    while (entries) {
        if ((node = getMsgFromLfapHighRcvdQueue())) {
            rmsg = (LFAPRcvdMsg*) lfapNodeInfo(node);
            lfap_free(rmsg->buf);

            rmsg->buf = 0; rmsg->len = 0;
            returnLfapSendMsg(node);
        }
        entries--;
    }

}

/******************************************************************************/
/*                                                                            */
/*  Function: packLfapServerListAR_                                             */
/*                                                                            */
/*  Places the LFAP_server_list into the LFAP_AR message to be sent to the    */
/*  server.  If there are no entries in the list, nothing is done to the AR   */
/*  message.  The number of entries in the list is returned.                  */
/*                                                                            */
/******************************************************************************/
static int packLfapServerListAR_ (LFAP_AR* ar)
{
    char ip[4];
    int entries;
    LFAPListIterator iter;
    lfapui32_t server_ip = 0;

    entries = lfapListEntries(&LFAP_server_list);

    if (entries) {
        setCommandLfapAR(ar, LFAP_LIST_OF_FASS);
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(ar, 0, LFAP_SEND);
        #endif
        encodeFixedLfapAR(ar);

        initLfapIterList(&iter, &LFAP_server_list);
        while ((server_ip = (lfapui32_t) incLfapIter(&iter))) {
            ip[0] = (char) (server_ip >> 24) & 0xff;
            ip[1] = (char) (server_ip >> 16) & 0xff;
            ip[2] = (char) (server_ip >>  8) & 0xff;
            ip[3] = (char) (server_ip >>  0) & 0xff;
            setFASIpAddrLfapAR(ar, ip, LFAP_IP_VERSION_4, 4);

            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_MESSAGES)
                PRINT_LFAP_MESSAGE(ar, 0, LFAP_SEND);
            #endif
            encodeRecordLfapAR(ar);
        }
    }

    return entries;
}

/******************************************************************************/
/*                                                                            */
/*  Function: packLfapServerListCR_                                             */
/*                                                                            */
/*  Places the LFAP_server_list into the LFAP_CR message to be sent to the    */
/*  server.  If there are no entries in the list, nothing is done to the CR   */
/*  message.  The number of entries in the list is returned.                  */
/*                                                                            */
/******************************************************************************/
static int packLfapServerListCR_ (LFAP_CR* cr)
{
    char ip[4];
    int entries;
    LFAPListIterator iter;
    lfapui32_t server_ip = 0;

    entries = lfapListEntries(&LFAP_server_list);

    if (entries) {

        encodeFixedLfapCR(cr);
        initLfapIterList(&iter, &LFAP_server_list);
        while ((server_ip = (lfapui32_t) incLfapIter(&iter))) {
            ip[0] = (char) (server_ip >> 24) & 0xff;
            ip[1] = (char) (server_ip >> 16) & 0xff;
            ip[2] = (char) (server_ip >>  8) & 0xff;
            ip[3] = (char) (server_ip >>  0) & 0xff;
            setFASIpAddrLfapCR(cr, ip, LFAP_IP_VERSION_4, 4);

            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_MESSAGES)
                PRINT_LFAP_MESSAGE(cr, 0, LFAP_SEND);
            #endif
            encodeRecordLfapCR(cr);
        }
    }

    return entries;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPConnect_                                                    */
/*                                                                            */
/*  Attempt to make a TCP connection to the current LFAP_server.  If server   */
/*  is specified, a connection is attempted to be made to that server.  If    */
/*  not, the connection is attempted to be made with the LFAP_server.         */
/*                                                                            */
/******************************************************************************/
static int LFAPConnect_ (lfapui32_t server)
{
    #ifdef CS_DEBUG
    char buffer[256];
    #endif
    /*    int connect = LFAP_TCP_CONNECTION_FAILED; */
    int connect = LFAP_FAILURE;

    if (server) {
        setLfapApiStats_(server);

        if (lfap_debug_fas)
            connect = LFAPConnect(server);
        else {
          /*            connect = LFAP_TCP_CONNECTION_ESTABLISHED; */
            connect = LFAP_SUCCESS;
            LFAP_fas_hi = LFAP_fas_lo = 1;
        }

        /* API Stats */
        if (LFAP_curr_stats) {
          /*            if (connect == LFAP_TCP_CONNECTION_ESTABLISHED) { */
          if ( connect == LFAP_SUCCESS ) {
            LFAP_curr_stats->conn_time = LFAPGetSysUptime();
            LFAP_curr_stats->retry_conn_success++;
            }
            else
                LFAP_curr_stats->retry_conn_failure++;
        }
    }

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_CONNECTING) {
        fprintf(stderr, "LFAPConnect_\n");
        fprintf(stderr, "   connecting to:   ");
        sprintf(buffer, "%d.%d.%d.%d",
               (int) (server >> 24) & 0xff,
               (int) (server >> 16) & 0xff,
               (int) (server >>  8) & 0xff,
               (int) (server >>  0) & 0xff);
        fprintf(stderr, "%s\t...", buffer);

        /*        if (connect == LFAP_TCP_CONNECTION_FAILED) */
        if ( connect == LFAP_FAILURE )
          fprintf(stderr, "   Could not connect\n\n");
        /*        else if (connect == LFAP_TCP_CONNECTION_ESTABLISHED) */
        else if (connect == LFAP_SUCCESS)
          fprintf(stderr, "   Connected!\n\n");
        else
            fprintf(stderr, "   ?????????\n\n");
    }
    #endif

    return connect;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPDisconnect_                                                 */
/*                                                                            */
/*  Attempt to disconnect the TCP connection to the current LFAP_Server.  The */
/*  server is used if specified, if not, the connection is attempted to be    */
/*  broken with the LFAP_server.                                              */
/*                                                                            */
/******************************************************************************/
static void LFAPDisconnect_ (lfapui32_t server)
{
    #ifdef CS_DEBUG
    char buffer[256];
    #endif
    lfapui32_t ip = 0;
/*    int disconnect = LFAP_TCP_DISCONNECT_FAILED; */
    int disconnect = LFAP_FAILURE;

    /* stop syncing, sending up stats, etc. */
    LFAP_report_on_flows = LFAP_NOT_REPORTING_FLOWS;

    if (server)
        ip = (lfapui32_t) server;
    else if (LFAP_server)
        ip = LFAP_server;

    if (ip) disconnect = LFAPDisconnect(ip);

        /* PAC_REMOVE   if ((disconnect == LFAP_TCP_DISCONNECT_SUCCESS) && LFAP_curr_stats && */
        if ( (disconnect == LFAP_SUCCESS) &&
            LFAP_curr_stats && 
            LFAP_curr_stats->conn_time) {
        LFAP_curr_stats->up_time = LFAP_curr_stats->up_time +
                                 (LFAPGetSysUptime() - LFAP_curr_stats->conn_time);
        LFAP_curr_stats->conn_time = 0;
    }

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_CONNECTING)
    {
        fprintf(stderr, "LFAPDisconnect_\n");
        fprintf(stderr, "   disconnecting:   ");
        sprintf(buffer, "%d.%d.%d.%d",
               (int) (ip >> 24) & 0xff,
               (int) (ip >> 16) & 0xff,
               (int) (ip >>  8) & 0xff,
               (int) (ip >>  0) & 0xff);
        fprintf(stderr, "%s\t...", buffer);

        /*        if (disconnect == LFAP_TCP_DISCONNECT_FAILED) */
        if (disconnect == LFAP_FAILURE)
            fprintf(stderr, "   Could not disconnect\n\n");
            /*        else if (disconnect == LFAP_TCP_DISCONNECT_SUCCESS) */
        else if (disconnect == LFAP_SUCCESS)
                fprintf(stderr, "   Disconnected!\n\n");
            else
                fprintf(stderr, "   ?????????\n\n");
    }
    #endif
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPLostConnInterval_                                           */
/*                                                                            */
/*  This checks to see if the CCE has lost connection to the server.  This is */
/*  is completed by checking to see if any messages have been received from   */
/*  the server.                                                               */
/*                                                                            */
/******************************************************************************/
static void LFAPLostConnInterval_ (void)
{
    const LFAP_CB_ENTRYPOINT cb = LFAPLostConnTimerExpired_;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, " LFAPLostConnInterval_...\n");
        fprintf(stderr, "      LFAP_msgs_rcvd = %u\n", LFAP_msgs_rcvd);
    }
    #endif
      LFAP_connection_timer = LFAP_TIMER_INACTIVE;


    if (!LFAP_msgs_rcvd && (LFAP_conn_state >= LFAP_TCP_CONNECTION) &&  lfap_debug_fas) {
        LFAP_error.error = LFAP_LOST_CONN_INTERVAL_EXPIRED;
        LFAP_error.time = LFAPGetSysUptime();
        LFAPDisconnect_(LFAP_server);
        LFAPConnectionLost(LFAP_server);
        LFAP_lost_conn_timer = LFAP_TIMER_INACTIVE;
    }
    else {
        LFAP_msgs_rcvd = 0;

        LFAPSetSystemTimer(cb, 0, LFAP_lost_conn_interval);
        LFAP_lost_conn_timer = LFAP_TIMER_WAITING;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPRetryServerList_                                            */
/*                                                                            */
/*  Go through all of the servers in the server list, trying to connect to    */
/*  each one.                                                                 */
/*                                                                            */
/******************************************************************************/
static void LFAPRetryServerList_ (lfapui32_t server)
{
    #ifdef CS_DEBUG
    char buffer[256];
    #endif
    lfapui32_t ip;
    LFAPNode* node;
    /*    int connected = LFAP_TCP_CONNECTION_FAILED; */
    int connected = LFAP_FAILURE;
    int num_servers = lfapListEntries(&LFAP_server_list);
    const LFAP_CB_ENTRYPOINT batch_cb = LFAPBatchIntervalExpired_;
    const LFAP_CB_ENTRYPOINT conn_timer_cb = LFAPConnectionTimerExpired_;

    LFAP_connection_timer = LFAP_TIMER_PROCESSING;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_CONNECTING) {
        fprintf(stderr, " LFAPRetryServerList_...\n");
        fprintf(stderr, "   number of entries: %d\n", num_servers);
        if (LFAP_new_server) {
            fprintf(stderr, "          new server: ");
            sprintf(buffer, "%d.%d.%d.%d",
                   (int) (LFAP_new_server >> 24) & 0xff,
                   (int) (LFAP_new_server >> 16) & 0xff,
                   (int) (LFAP_new_server >>  8) & 0xff,
                   (int) (LFAP_new_server >>  0) & 0xff);
            fprintf(stderr, "%s\n", buffer);
        }
    }
    #endif

    /* PAC_REMOVE    if (LFAP_trying_to_connect == LFAP_ALREADY_TRYING_TO_CONNECT) {*/
    if ( LFAP_conn_state <= LFAP_NO_TCP_CONNECTION)
      LFAP_conn_state = LFAP_ATTEMPTING_TCP_CONNECT;
    else {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_CONNECTING)
            fprintf(stderr, " ** ALREADY CONNECTED OR TRYING TO CONNECT **\n\n");
        #endif
        return;
    }
    /* PAC_REMOVE    else */
    /* PAC_REMOVE      LFAP_trying_to_connect = LFAP_ALREADY_TRYING_TO_CONNECT;*/

    /* PAC_REMOVE    LFAP_conn_status = LFAP_ALREADY_TRYING_TO_CONNECT;*/

    if (num_servers) {
        while (num_servers) {
            if (server) {
                ip = server;
                updateLfapServer_(server);
            }
            else if (LFAP_new_server) {
                ip = LFAP_new_server;
                updateLfapServer_(LFAP_new_server);
                LFAP_new_server = 0;
            }
            else {
                if (!LFAP_server)
                    LFAP_server = (lfapui32_t) lfapListFirst(&LFAP_server_list);
                else if ((node = findLfapNode(&LFAP_server_list, (void*)LFAP_server))) {
                    node = nextLfapNode(node);
                    if (node == tailLfapNode(&LFAP_server_list)) {
                        LFAP_server = (lfapui32_t)
                                      lfapListFirst(&LFAP_server_list);
                        if (LFAP_retry_server_list == LFAP_WAIT_IF_LIST_END) {
                            LFAP_retry_server_list = LFAP_RETRY_IF_LIST_END;
                            break;
                        }
                    }
                    else
                        LFAP_server = (lfapui32_t) lfapNodeInfo(node);
                }
                else
                    LFAP_server = (lfapui32_t) lfapListFirst(&LFAP_server_list);

                ip = LFAP_server;
            }

            connected = LFAPConnect_(ip);

            /*            if (connected == LFAP_TCP_CONNECTION_FAILED) { */
            if ( connected == LFAP_FAILURE ) {
                num_servers--;
                LFAPSynchronize();
            }
            else {
                LFAP_retry_server_list = LFAP_RETRY_IF_LIST_END;
                break;
            }
         }

          /*        if (connected == LFAP_TCP_CONNECTION_FAILED) { */
        if ( connected == LFAP_FAILURE ) {
            LFAP_error.error = LFAP_NO_SERVERS_AVAILABLE;
            LFAP_error.time = LFAPGetSysUptime();

            /* PAC_REMOVE            LFAP_conn_status = LFAP_TCP_CONNECTION_FAILED;*/

            LFAP_connection_timer = LFAP_TIMER_WAITING;
            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_TIMERS) {
                fprintf(stderr, " Setting the server retry timer: ");
                fprintf(stderr, "%lu ms\n", LFAP_server_retry_interval);
            }
            #endif
            LFAPSetSystemTimer(conn_timer_cb, 0, LFAP_server_retry_interval);
        }
        /* PAC_REMOVE         else { */
          /* PAC_REMOVE            LFAP_connected = LFAP_CONNECTION_ESTABLISHED;*/
          /* PAC_REMOVE            LFAP_conn_status = LFAP_TCP_CONNECTED_NO_ACK;*/
        /* PAC_REMOVE             LFAP_connection_timer = LFAP_TIMER_INACTIVE; */

        /* PAC_REMOVE  send up the VR message indicating LFAP version */
        /* PAC_REMOVE             indicateLfapVersion_(); */
        /* PAC_REMOVE         } */
    }
    else {
        LFAP_error.error = LFAP_NO_SERVERS_IN_LIST;
        LFAP_error.time = LFAPGetSysUptime();

        /* PAC_REMOVE        LFAP_conn_status = LFAP_TCP_CONNECTION_FAILED;*/

        LFAP_connection_timer = LFAP_TIMER_WAITING;
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_TIMERS) {
            fprintf(stderr, " Setting the server retry timer: ");
            fprintf(stderr, "%lu ms\n", LFAP_server_retry_interval);
        }
        #endif
        LFAPSetSystemTimer(conn_timer_cb, 0, LFAP_server_retry_interval);
    }

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_CONNECTING) {
        fprintf(stderr, " returning from LFAPRetryServerList_:: ");
/*        if (connected == LFAP_TCP_CONNECTION_FAILED) */
        if (connected == LFAP_FAILURE)
            fprintf(stderr, "FAILED!\n");
        else
            fprintf(stderr, "CONNECTED!\n");
    }
    #endif

    /* start send queue timer */
   /* PAC_REMOVE   if ((connected == LFAP_TCP_CONNECTION_ESTABLISHED) && */

    if ( connected != LFAP_SUCCESS) 
      LFAP_conn_state = LFAP_NO_TCP_CONNECTION;
    else {
      const LFAP_CB_ENTRYPOINT cb = LFAPLostConnTimerExpired_;

      /* send up the VR message indicating LFAP version */
      LFAP_conn_state = LFAP_TCP_CONNECTION;
      indicateLfapVersion_();

      /* don't need to connection timer any more, we have one */
      LFAP_connection_timer = LFAP_TIMER_INACTIVE;


      /* crank up the batch interval if it is not already going */
      if (LFAP_batch_interval_timer == LFAP_TIMER_INACTIVE) {
        LFAPSetSystemTimer(batch_cb, 0, LFAP_batch_interval);
        LFAP_batch_interval_timer = LFAP_TIMER_WAITING;
      }

      /* no that we have a connection, crank up the lost connection
       * timer if it is not already going 
      */
      if ( LFAP_lost_conn_timer == LFAP_TIMER_INACTIVE )
      {
        LFAPSetSystemTimer(cb, 0, LFAP_lost_conn_interval);
        LFAP_lost_conn_timer = LFAP_TIMER_WAITING;
      }
      

    }

    /* PAC_REMOVE    LFAP_trying_to_connect = LFAP_NOT_TRYING_TO_CONNECT;*/
    return;
}

/******************************************************************************/
/*                                                                            */
/*  Function: handleLfapVRA_                                                  */
/*                                                                            */
/*  Performs the appropriate actions for receiving an LFAP_VRA message from   */
/*  the server.                                                               */
/*                                                                            */
/******************************************************************************/
static void handleLfapVRA_ (void* buf, int buf_len)
{
    LFAP_CR cr;
    LFAP_VRA vra;
    lfapui8_t selection = 0;

    initLfapCR(&cr);
    initRcvdLfapVRA(&vra, buf, buf_len);
    /* LFAP_VRA has no ies */

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_MESSAGES)
        PRINT_LFAP_MESSAGE(&vra, 0, LFAP_RECV);
    #endif

    if (lfap_msg_get_version(buf) != LFAP_VERSION) {
      /* PAC_REMOVE         removeLfapVersion_(); */
        LFAPDisconnect_(LFAP_server);
        LFAPConnectionLost(LFAP_server);
        LFAP_retry_server_list = LFAP_WAIT_IF_LIST_END;
    }
    else {
        selection = verifyLfapVRA(&vra);

        /* send list of FASs */
        if (packLfapServerListCR_(&cr)) {
            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_MESSAGES)
                PRINT_LFAP_MESSAGE(&cr, 0, LFAP_SEND);
            #endif
            encodeCompletedLfapCR(&cr);
            if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&cr), getLfapMsgLength(&cr), LFAP_REQ_MSG))
                freeLfapMsg(&cr);
            else
              LFAP_conn_state = LFAP_SECURITY_CHECK;
        }
    }

    freeLfapMsg(&vra);
}

/******************************************************************************/
/*                                                                            */
/*  Function: handleLfapAR_                                                   */
/*                                                                            */
/*  Performs the appropriate actions for receiving an LFAP_AR message from    */
/*  the server.                                                               */
/*                                                                            */
/******************************************************************************/
static void handleLfapAR_ (void* buf, int buf_len)
{
    LFAP_AR ar;
    void* ar_buf;
    LFAP_ARA ara;
    int send_ara = 1;
    int disconnect = 0;
    lfapui8_t selection = 0;
    const LFAP_CB_ENTRYPOINT cb = LFAPLostConnTimerExpired_;

    initRcvdLfapAR(&ar, buf, buf_len);
    initLfapARAfromAR(&ara, &ar);

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_MESSAGES) {
        if ((ar.command.rsvd_ui8.val != LFAP_KEEPALIVE) ||
            (lfap_debug & LFAP_DBG_KEEPALIVE))
            PRINT_LFAP_MESSAGE(&ar, 0, LFAP_RECV);
    }
    #endif

    if ((selection = verifyLfapAR(&ar))) {
        switch(ar.command.rsvd_ui8.val) {
            case LFAP_RETURN_INDICATED_FLOWS:
                if (LFAP_report_on_flows == LFAP_NOT_REPORTING_FLOWS) {
                    ar_buf = (void*) lfap_malloc(buf_len);
                    lfap_memcpy(ar_buf, buf, buf_len);
                    initRcvdLfapAR(&LFAP_report_flows_ar, ar_buf, buf_len);
                    initLfapARAfromAR(&LFAP_report_flows_ara,
                                      &LFAP_report_flows_ar);
                    LFAP_report_on_flows = LFAP_REPORT_INDICATED_FLOWS;
                    send_ara = 0;
                }
                else {
                    setLfapMsgStatus(&ara, LFAP_MSG_CORRUPTED);
                }
                break;
            case LFAP_RETURN_FLOW_PREFIX:
                setFlowPrefixLfapARA(&ara, LFAP_fas_hi, LFAP_fas_lo);
                break;
            case LFAP_CONNECTION_ACCEPTED:
              /* PAC_REMOVE                LFAP_conn_status = LFAP_CONNECTION_ESTABLISHED;*/
                requestLfapFlowPrefix_();
                if (LFAP_lost_conn_timer == LFAP_TIMER_INACTIVE) {
                    LFAPSetSystemTimer(cb, 0, LFAP_lost_conn_interval);
                    LFAP_lost_conn_timer = LFAP_TIMER_WAITING;
                }
                break;
            case LFAP_DISCONNECT:
            case LFAP_CONNECTION_REJECTED:
                disconnect = 1;
                send_ara = 0;
                if (selection & LFAP_AR_FAS_IP_ADDR)
                    LFAP_new_server = *(lfapui32_t*) ar.fas_ip_addr.addr_ie.val;
                break;
            case LFAP_KEEPALIVE:
                send_ara = 0;
                break;
            case LFAP_RETURN_TIME: {
                lfapui32_t ti = LFAPGetSysUptime();
                setTimeLfapARA(&ara, ti);
                }        
                break;
            case LFAP_REQUESTED_IEs:
                break;
            /* API should not receive this */
            case LFAP_LIST_OF_FASS:
            default:
                setLfapMsgStatus(&ara, LFAP_MSG_CORRUPTED);
                break;
        }
    }
    else
        setLfapMsgStatus(&ara, LFAP_MSG_CORRUPTED);

    if (send_ara) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(&ara, 0, LFAP_SEND);
        #endif
        encodeCompletedLfapARA(&ara);
        if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&ara), getLfapMsgLength(&ara), LFAP_REQ_MSG))
            freeLfapMsg(&ara);
    }

    if (disconnect) {
        LFAPDisconnect_(LFAP_server);
        LFAPConnectionLost(LFAP_server);
        LFAP_retry_server_list = LFAP_WAIT_IF_LIST_END;
    }

    freeLfapMsg(&ar);
}

/******************************************************************************/
/*                                                                            */
/*  Function: handleLfapCAN_                                                  */
/*                                                                            */
/******************************************************************************/
static void handleLfapCAN_ (void* buf, int buf_len)
{
    LFAP_CAN can;

    initRcvdLfapCAN(&can, buf, buf_len);
    if (lfap_debug & LFAP_DBG_MESSAGES)
      PRINT_LFAP_MESSAGE(&can, 0, LFAP_RECV);


    /* PAC_REMOVE    LFAP_conn_status = LFAP_CONNECTION_ESTABLISHED;*/
    LFAP_conn_state = LFAP_DATA_EXPORT_NEGOTIATION;
    requestLfapFlowPrefix_();

    freeLfapMsg(&can);
}

/******************************************************************************/
/*                                                                            */
/*  Function: handleLfapCRN_                                                  */
/*                                                                            */
/******************************************************************************/
static void handleLfapCRN_ (void* buf, int buf_len)
{

  LFAP_CRN crn;
  
  initRcvdLfapCRN(&crn, buf, buf_len);
  if (lfap_debug & LFAP_DBG_MESSAGES)
    PRINT_LFAP_MESSAGE(&crn, 0, LFAP_RECV);

  LFAPDisconnect_(LFAP_server);
  /* Note - state set to no connection in LFAPConnectionLost */
  LFAPConnectionLost(LFAP_server);
  LFAP_retry_server_list = LFAP_WAIT_IF_LIST_END;

  freeLfapMsg(&crn);
}

/******************************************************************************/
/*                                                                            */
/*  Function: handleLfapFER_                                                  */
/*                                                                            */
/******************************************************************************/
static void handleLfapFER_ (void* buf, int buf_len)
{

  LFAP_FER fer;
  
  initRcvdLfapFER(&fer, buf, buf_len);
  if (lfap_debug & LFAP_DBG_MESSAGES)
    PRINT_LFAP_MESSAGE(&fer, 0, LFAP_RECV);

  LFAP_conn_state = LFAP_SESSION_ESTABLISHED;
  
  freeLfapMsg(&fer);
}

/******************************************************************************/
/*                                                                            */
/*  Function: handleLfapKA_                                                   */
/*                                                                            */
/******************************************************************************/
static void handleLfapKA_ (void* buf, int buf_len)
{

  LFAP_KA ka;
  
  initRcvdLfapKA(&ka, buf, buf_len);
  if ((lfap_debug &  LFAP_DBG_MESSAGES) && (lfap_debug & LFAP_DBG_KEEPALIVE)) {
    PRINT_LFAP_MESSAGE(&ka, 0, LFAP_RECV);
  }

  freeLfapMsg(&ka);
}


/******************************************************************************/
/*                                                                            */
/*  Function: handleLfapDR_                                                   */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
static void handleLfapDR_ (void* buf, int buf_len)
{
    LFAP_DR dr;
    lfapui8_t selection = 0;

    initRcvdLfapDR(&dr, buf, buf_len);

    if (lfap_debug & LFAP_DBG_MESSAGES)
      PRINT_LFAP_MESSAGE(&dr, 0, LFAP_RECV);

    if ((selection = verifyLfapDR(&dr))) {
      if (selection & LFAP_DR_FAS_IP_ADDR) 
        LFAP_new_server = *(lfapui32_t*) dr.fas_ip_addr.addr_ie.val;
    }
    
    LFAPDisconnect_(LFAP_server);
    LFAPConnectionLost(LFAP_server);
    LFAP_retry_server_list = LFAP_WAIT_IF_LIST_END;

    freeLfapMsg(&dr);
}

/******************************************************************************/
/*                                                                            */
/*  Function: handleLfapARA_                                                  */
/*                                                                            */
/*  Performs the appropriate actions for receiving an LFAP_ARA message from   */
/*  the server.                                                               */
/*                                                                            */
/******************************************************************************/
static void handleLfapARA_ (void* buf, int buf_len)
{
    LFAP_ARA ara;
    LFAPIE flow_prefix;
    lfapui8_t selection = 0;

    initRcvdLfapARA(&ara, buf, buf_len);
    do {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(&ara, 0, LFAP_RECV);
        #endif

        if ((selection = verifyLfapARA(&ara))) {
            if (selection & LFAP_ARA_FLOW_PREFIX) {
                flow_prefix = ara.flow_prefix;
                if (flow_prefix.type == LFAP_FLOW_ID_PREFIX) {
                    updateLfapFlowPrefix_(flow_prefix.pfx.fas_id[0],
                                          flow_prefix.pfx.fas_id[1]);
                }
            }
        }

    } while (decodeNextRecordLfapARA(&ara) != LFAP_EMPTY_BUFFER);

    freeLfapMsg(&ara);
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPDispatchMsg_                                                 */
/*                                                                            */
/*  Call appropriate message handler                                          */
/*                                                                            */
/******************************************************************************/
static void LFAPDispatchMsg_ (LFAPRcvdMsg* msg)
{
    int status = LFAP_SUCCESS;
  
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, " LFAPLFAPDispatchMsg_, opcode: ");
        fprintf(stderr, "%d\n", lfap_msg_get_opcode(msg->buf));
    }
    #endif
    
    switch (lfap_msg_get_opcode(msg->buf)) {
      case LFAP_VRA_OPCODE:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_vra_rcvd++;

        if ( LFAP_conn_state == LFAP_PROTOCOL_NEGOTIATION)
          handleLfapVRA_(msg->buf, msg->len);
        else
          status = LFAP_FAILURE;
        break;

      case LFAP_CAN_OPCODE:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_can_rcvd++;

        if ( LFAP_conn_state == LFAP_SECURITY_CHECK)
          handleLfapCAN_(msg->buf, msg->len);
        else
          status = LFAP_FAILURE;
        break;

      case LFAP_CRN_OPCODE:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_crn_rcvd++;

        if ( LFAP_conn_state == LFAP_SECURITY_CHECK)
          handleLfapCRN_(msg->buf, msg->len);
        else
          status = LFAP_FAILURE;
        break;

      case LFAP_FER_OPCODE:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_fer_rcvd++;

        if ( LFAP_conn_state == LFAP_DATA_EXPORT_NEGOTIATION )
          handleLfapFER_(msg->buf, msg->len);
        else
          status = LFAP_FAILURE;
        break;

      case LFAP_KA_OPCODE:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_ka_rcvd++;

        /* the spec says that KA's start after the session
         * is established but I think that is something people
         * will get wrong so I'll allow it. after DATA_EXPORT_NEGOTIATION.
        */
        if ( LFAP_conn_state >= LFAP_DATA_EXPORT_NEGOTIATION)
          handleLfapKA_(msg->buf, msg->len);
        else
          status = LFAP_FAILURE;
        break;

      case LFAP_DR_OPCODE:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_dr_rcvd++;

        if ( LFAP_conn_state >= LFAP_SECURITY_CHECK )
          handleLfapDR_(msg->buf, msg->len);
        else
          status = LFAP_FAILURE;
        break;

      case LFAP_AR_OPCODE:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_ar_rcvd++;

        if ( LFAP_conn_state >= LFAP_DATA_EXPORT_NEGOTIATION )
          handleLfapAR_(msg->buf, msg->len);
        else 
          status = LFAP_FAILURE;
        break;

      case LFAP_ARA_OPCODE:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_ara_rcvd++;

        if ( LFAP_conn_state >= LFAP_DATA_EXPORT_NEGOTIATION )
          handleLfapARA_(msg->buf, msg->len);
        else 
          status = LFAP_FAILURE;
        break;
        
      default:
        if (LFAP_curr_stats) LFAP_curr_stats->msg_unk_rcvd++;
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_WARNINGS) {
          /* SSR - error log */
          ERR_errlog(0, lfap_efacid, LFAP_W_UNKNOWN_MSG, 1, lfap_msg_get_opcode(msg->buf));
        }
        #endif
        break;
    }

    if ( status == LFAP_FAILURE ) {
      LFAPDisconnect_(LFAP_server);
      LFAPConnectionLost(LFAP_server);
      LFAP_retry_server_list = LFAP_WAIT_IF_LIST_END;
    }
    
}




/******************************************************************************/
/*                                                                            */
/*  Function: LFAPReceiveHighQueueMsgs_                                       */
/*                                                                            */
/*  Processes high-q received messages.                                       */
/*                                                                            */
/******************************************************************************/
static void LFAPReceiveHighQueueMsgs_ (void)
{
    LFAPRcvdMsg* msg;
    LFAPNode* node = 0;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, " LFAPReceiveHighQueueMsgs_, msgs: ");
        fprintf(stderr, "%d\n", msgsInLfapHighRcvdQueue());
    }
    #endif

    while ((node = getMsgFromLfapHighRcvdQueue())) {
        msg = (LFAPRcvdMsg*) lfapNodeInfo(node);
        if (lfap_debug & LFAP_DBG_WARNINGS) {
          if (LFAPMsgPriority_(msg->buf, msg->len) != High)
            /* PAC - How to log */
            fprintf(stderr, "Low priority message in high queue\n");
        }
        
        LFAPDispatchMsg_(msg);

        msg->buf = 0;
        msg->len = 0;
        returnLfapRcvdMsg(node);
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPReceiveRegQueueMsgs_                                        */
/*                                                                            */
/*  Processes reg-q received messages.                                        */
/*                                                                            */
/******************************************************************************/
static void LFAPReceiveRegQueueMsgs_ (void)
{
    LFAPRcvdMsg* msg;
    LFAPNode* node = 0;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, " LFAPReceiveRegQueueMsgs_, msgs: ");
        fprintf(stderr, "%d\n", msgsInLfapRegRcvdQueue());
    }
    #endif

    while ((node = getMsgFromLfapRegRcvdQueue())) {
        msg = (LFAPRcvdMsg*) lfapNodeInfo(node);
        if (lfap_debug & LFAP_DBG_WARNINGS) {
          if (LFAPMsgPriority_(msg->buf, msg->len) != Low)
            /* PAC - How to log */
            fprintf(stderr, "LFAP-WARNING - High priority message in low queue\n");
        }
        
        LFAPDispatchMsg_(msg);

        msg->buf = 0;
        msg->len = 0;
        returnLfapRcvdMsg(node);
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSendVector_                                                 */
/*                                                                            */
/*  Sends the LFAP_send_vector to the server.  Returns LFAP_SEND_SUCCESS if   */
/*  successful, otherwise LFAP_SEND_FAILURE.                                  */
/*                                                                            */
/******************************************************************************/
static int LFAPSendVector_ (void)
{
    int i;
    int sent;
    LFAPNode* node;
    LFAPSendMsg* msg;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, "  LFAPSendVector_ ... entries: ");
        fprintf(stderr, "%d, ", LFAP_send_vector_entries);
        fprintf(stderr, "len: %d .... ", LFAP_send_vector_len);
    }
    #endif

    if ( LFAP_send_vector_entries == 0 )
      return  LFAP_SEND_SUCCESS;
    

    if (lfap_debug_fas)
        sent = LFAPSendVector(LFAP_send_vector, LFAP_send_vector_entries,
                              LFAP_send_vector_len, LFAP_lost_conn_count);
    else
        sent = LFAP_SEND_SUCCESS;

    if (sent == LFAP_SEND_SUCCESS) {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_FUNCTIONS)
            fprintf(stderr, "SUCCESS!!\n");
        #endif

        for (i = 0; i < LFAP_send_vector_entries; i++) {
            node = LFAP_send_vector_nodes[i];
            msg = (LFAPSendMsg*) lfapNodeInfo(node);

            /* API Stats */
            if (LFAP_curr_stats) {
                LFAP_curr_stats->msgs_sent_success++;
                LFAP_curr_stats->total_bytes_sent += msg->len;

                switch(lfap_msg_get_opcode(msg->buf)) {
                    case LFAP_VR_OPCODE:
                        LFAP_curr_stats->msg_vr_sent++;
                        break;
                    case LFAP_AR_OPCODE:
                        LFAP_curr_stats->msg_ar_sent++;
                        break;
                    case LFAP_ARA_OPCODE:
                        LFAP_curr_stats->msg_ara_sent++;
                        break;
                    case LFAP_FAR_OPCODE:
                        LFAP_curr_stats->msg_far_sent++;
                        break;
                    case LFAP_FUN_OPCODE:
                        LFAP_curr_stats->msg_fun_sent++;
                        break;
                    case LFAP_CR_OPCODE:
                      LFAP_curr_stats->msg_cr_sent++;
                      break;
                    
                }
            }

            lfap_free(msg->buf);
            msg->buf = 0; msg->len = 0;
            returnLfapSendMsg(node);
            LFAP_send_vector_nodes[i] = 0;
        }

        LFAP_send_vector_len = 0;
        LFAP_send_vector_entries = 0;
    }
    else {
        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_FUNCTIONS)
            fprintf(stderr, "FAILURE!!\n");
        #endif

        if (LFAP_curr_stats)
            LFAP_curr_stats->msgs_sent_failure += LFAP_send_vector_entries;
    }

    return sent;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetSrcAS_                                                   */
/*                                                                            */
/*  SSR: Gets the source AS, and places it into the FAR message buffer, buf.  */
/*  Important note: this assumes the AS is the first optional IE that is      */
/*  encoded (after the mandatory IEs are encoded).                            */
/*                                                                            */
/******************************************************************************/
static void LFAPGetSrcAS_ (void* buf)
{
    lfapui32_t src_as = 0;

    /* check to make sure the src/dest AS is where we think it should be      */
    /* HACK - this check currently keeps us from putting src AS in MPLS flows */
    if ( *((lfapui16_t*)buf+28) == LFAP_SRC_AS ) {
        src_as = LFAPGetSrcAS( *((lfapui32_t*)buf+10) );
        memcpy((char *)buf+60, &src_as, 4);
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSendRegQueueMsgs_                                           */
/*                                                                            */
/*  Sends reg-q messages to the server.                                       */
/*                                                                            */
/******************************************************************************/
static int LFAPSendRegQueueMsgs_ (lfapui8_t flush)
{
    LFAPSendMsg* msg;
    LFAPNode* node = 0;
    int work_completed = LFAP_WORK_INCOMPLETE;
    int status;
    

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, " LFAPSendRegQueueMsgs_, msgs: ");
        fprintf(stderr, "%d\n", msgsInLfapRegSendQueue());
    }
    #endif

    if (!flush && msgsInLfapRegSendQueue() < LFAP_MAX_MSGS_IN_SEND_VECTOR)
      return LFAP_WORK_COMPLETED;

    while (LFAP_send_vector_entries < LFAP_MAX_MSGS_IN_SEND_VECTOR) {
        if ((node = getMsgFromLfapRegSendQueue())) {
            msg = (LFAPSendMsg*) lfapNodeInfo(node);

            if (((lfapui16_t) ((char*)msg->buf)[1]) == LFAP_FAR_OPCODE) {
                /* SSR add flow to stats list on LFAP task */
                setLfapCheckpointTimer_( *((lfapui32_t*)msg->buf+5) );

                /* SSR src AS: 128 is LFAP_SEND_SRC_AS in lfap_main.h */
                if (lfap_send_ies & 128) LFAPGetSrcAS_(msg->buf);
            }

            LFAP_send_vector[LFAP_send_vector_entries].buf = msg->buf;
            LFAP_send_vector[LFAP_send_vector_entries].len = msg->len;
            LFAP_send_vector_nodes[LFAP_send_vector_entries] = node;

            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_WARNINGS) {
                if (msg->len <= 0) {
                    /* SSR - error log */
                    ERR_errlog(0, lfap_efacid, LFAP_W_INVALID_LEN, 1, msg->len);
                }
            }
            #endif

            LFAP_send_vector_len += msg->len;
            LFAP_send_vector_entries++;

            /* PAC_REMOVE printf("\n\n***** low Queue: Msg opcode = %d\n\n\n", lfap_msg_get_opcode(msg->buf)); */
            
        }
        else
            break;
    }

    status = LFAPSendVector_();

    if ( status == LFAP_SEND_SUCCESS && msgsInLfapRegSendQueue() < LFAP_MAX_MSGS_IN_SEND_VECTOR)
      work_completed = LFAP_WORK_COMPLETED;

    return work_completed;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSendBothQueueMsgs_                                          */
/*                                                                            */
/*  Sends high-q and reg-q messages to the server.                            */
/*                                                                            */
/******************************************************************************/
/* PAC_REMOVE static int LFAPSendBothQueueMsgs_ (void) */
/* PAC_REMOVE { */
/* PAC_REMOVE     LFAPSendMsg* msg; */
/* PAC_REMOVE     LFAPNode* node = 0; */
/* PAC_REMOVE  */
/* PAC_REMOVE     #ifdef CS_DEBUG */
/* PAC_REMOVE     if (lfap_debug & LFAP_DBG_FUNCTIONS) { */
/* PAC_REMOVE         fprintf(stderr, " LFAPSendBothQueueMsgs_\n"); */
/* PAC_REMOVE         fprintf(stderr, "      high: %d\n", msgsInLfapHighSendQueue()); */
/* PAC_REMOVE         fprintf(stderr, "       reg: %d\n", msgsInLfapRegSendQueue()); */
/* PAC_REMOVE     } */
/* PAC_REMOVE     #endif */
/* PAC_REMOVE  */
/* PAC_REMOVE     while (LFAP_send_vector_entries < LFAP_MAX_MSGS_IN_SEND_VECTOR) { */
/* PAC_REMOVE         if ((node = getMsgFromLfapHighSendQueue())) { */
/* PAC_REMOVE             msg = (LFAPSendMsg*) lfapNodeInfo(node); */
/* PAC_REMOVE             LFAP_send_vector[LFAP_send_vector_entries].buf = msg->buf; */
/* PAC_REMOVE             LFAP_send_vector[LFAP_send_vector_entries].len = msg->len; */
/* PAC_REMOVE             LFAP_send_vector_nodes[LFAP_send_vector_entries] = node; */
/* PAC_REMOVE             LFAP_send_vector_len += msg->len; */
/* PAC_REMOVE             LFAP_send_vector_entries++; */
/* PAC_REMOVE  */
/* PAC_REMOVE             printf("\n\n***** High Queue: Msg opcode = %d\n\n\n", lfap_msg_get_opcode(msg->buf)); */
/* PAC_REMOVE              */
/* PAC_REMOVE         } */
/* PAC_REMOVE         else */
/* PAC_REMOVE             break; */
/* PAC_REMOVE     } */
/* PAC_REMOVE  */
/* PAC_REMOVE     return LFAPSendRegQueueMsgs_(0); */
/* PAC_REMOVE } */

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPSendHighQueueMsgs_                                          */
/*                                                                            */
/*  Sends high-q                                                              */
/*                                                                            */
/******************************************************************************/
static int LFAPSendHighQueueMsgs_ (void)
{
    LFAPSendMsg* msg;
    LFAPNode* node = 0;
    int status = LFAP_SEND_SUCCESS;
    int count = 0;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, " LFAPSendHighQueueMsgs_\n");
        fprintf(stderr, "      high: %d\n", msgsInLfapHighSendQueue());
    }
    #endif

    /* We don't use the send vector but lets not send
     * any more in one wack than LFAPSendVector_ would send
    */
    while ( count++ < LFAP_MAX_MSGS_IN_SEND_VECTOR && 
            status == LFAP_SEND_SUCCESS ) {

      /* If we sent all of the last message, get a new one */
      if ( !LFAP_send_message ) {
        if ( !(node = getMsgFromLfapHighSendQueue()))
          break; /* no more left, break out */
        else {
            msg = (LFAPSendMsg*) lfapNodeInfo(node);
            LFAP_send_message = msg->buf;
            LFAP_send_message_len = msg->len;
            LFAP_send_message_node = node;
        }
      }

      status = LFAPSendMessage_(!CLEAR);
    }

    /* we need to send all messages in the high queue */
    if ( status == LFAP_SEND_SUCCESS && msgsInLfapHighSendQueue() == 0 )
      return LFAP_WORK_COMPLETED;
    else
      return LFAP_WORK_INCOMPLETE;
    
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGarbageCollector_                                           */
/*  GARBAGE_COLLECTOR                                                         */
/*  Garbage collector goes through the linked list of flows in chunks         */
/*  removing the flows in an effecient manner if they have already been       */
/*  deleted.                                                                  */
/*                                                                            */
/*  NOTE - if the LFAP_stats_list is changed to a more efficent structre for  */
/*         deletes (e.g. hash) then the whole garbage collector issue needs   */
/*         to be reworked.  search for GARBAGE_COLLECTOR                      */
/*                                                                            */
/******************************************************************************/
static lfapui16_t LFAPGarbageCollector_ (void)
{
    lfapui32_t flow;
    int flow_num = 0;
    LFAPNode* last_flow;
    LFAPNode* prev_node = 0;
    LFAPNode* next_node = 0;
    lfapui16_t work_completed = LFAP_WORK_INCOMPLETE;
    const LFAP_CB_ENTRYPOINT cb = LFAPGarbageCollectorIntervalExpired_;

    if (!LFAP_stats_num_flows)
        work_completed = LFAP_WORK_COMPLETED;

    if ((!LFAP_gcoll_flow) ||
         (LFAP_gcoll_flow == tailLfapNode(&LFAP_stats_list)))
        LFAP_gcoll_flow = firstLfapNode(&LFAP_stats_list);

    last_flow = lastLfapNode(&LFAP_stats_list);

    while ((flow_num++ < LFAP_GARBAGE_COLLECT_NUM_FLOWS) &&
           (work_completed != LFAP_WORK_COMPLETED)) {

        flow = (lfapui32_t) lfapNodeInfo(LFAP_gcoll_flow);
        if (LFAPDoesFlowExist(flow)) {
            /* flow exists! */

            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_FUNCTIONS)
                fprintf(stderr, "  Flow %lu exists\n", flow);
            #endif

            if (LFAP_gcoll_flow == last_flow) {
                LFAP_gcoll_flow = firstLfapNode(&LFAP_stats_list);
                work_completed = LFAP_WORK_COMPLETED;
            }
            else {
                prev_node = LFAP_gcoll_flow;
                LFAP_gcoll_flow = nextLfapNode(LFAP_gcoll_flow);
            }
        }
        else {
            /* flow has been deleted! */
            if (prev_node) {
                /* we can delete it */

                #ifdef CS_DEBUG
                if (lfap_debug & LFAP_DBG_FUNCTIONS)
                    fprintf(stderr, "  DELETING flow %lu\n", flow);
                #endif

                next_node = nextLfapNode(LFAP_gcoll_flow);

                if (LFAP_timer_flow == LFAP_gcoll_flow)
                    LFAP_timer_flow = next_node;

                removeNextLfapNode(&LFAP_stats_list, prev_node);
                LFAP_stats_num_flows = lfapListEntries(&LFAP_stats_list);

                if (LFAP_gcoll_flow == last_flow) {
                    LFAP_gcoll_flow = firstLfapNode(&LFAP_stats_list);
                    work_completed = LFAP_WORK_COMPLETED;
                }
                else
                    LFAP_gcoll_flow = next_node;
            }
            else if (firstLfapNode(&LFAP_stats_list) == last_flow) {
                /* there is only one flow in the list */

                #ifdef CS_DEBUG
                if (lfap_debug & LFAP_DBG_FUNCTIONS) {
                    fprintf(stderr, "  Deleting the ONLY flow in the list: ");
                    fprintf(stderr, "%lu\n", flow);
                }
                #endif

                next_node = nextLfapNode(LFAP_gcoll_flow);

                if (LFAP_timer_flow == LFAP_gcoll_flow)
                    LFAP_timer_flow = next_node;

                removeNextLfapNode(&LFAP_stats_list, &LFAP_stats_list._head_);
                LFAP_stats_num_flows = 0;

                LFAP_gcoll_flow = 0;
                work_completed = LFAP_WORK_COMPLETED;
            }
            else if (firstLfapNode(&LFAP_stats_list) == LFAP_gcoll_flow) {
                /* trying to delete the first flow, but there is more than 1
                   flow in the list */

                #ifdef CS_DEBUG
                if (lfap_debug & LFAP_DBG_FUNCTIONS) {
                    fprintf(stderr, "  Deleting the FIRST flow in the list ");
                    fprintf(stderr, "(but there is more!): %lu\n", flow);
                }
                #endif

                next_node = nextLfapNode(LFAP_gcoll_flow);

                if (LFAP_timer_flow == LFAP_gcoll_flow)
                    LFAP_timer_flow = next_node;

                removeNextLfapNode(&LFAP_stats_list, &LFAP_stats_list._head_);
                LFAP_stats_num_flows = lfapListEntries(&LFAP_stats_list);

                LFAP_gcoll_flow = next_node;
            }
            else {
                /* could not delete yet, get it on the next pass */

                #ifdef CS_DEBUG
                if (lfap_debug & LFAP_DBG_FUNCTIONS) {
                    fprintf(stderr, "  Could not delete flow %lu this ", flow);
                    fprintf(stderr, "pass .. wait until next time\n");
                }
                #endif

                if (LFAP_gcoll_flow == last_flow) {
                    LFAP_gcoll_flow = firstLfapNode(&LFAP_stats_list);
                    work_completed = LFAP_WORK_COMPLETED;
                }

                prev_node = LFAP_gcoll_flow;
                LFAP_gcoll_flow = nextLfapNode(LFAP_gcoll_flow);
            }
        }
    }

    if (work_completed == LFAP_WORK_COMPLETED) {
        LFAPSetSystemTimer(cb, 0, LFAP_GARBAGE_COLLECT_INTERVAL);
        LFAP_garbage_collector_timer = LFAP_TIMER_WAITING;
    }

    setLfapFunUpdateInterval_();

    return work_completed;
}


/******************************************************************************/
/*                                                                            */
/*  Function: LFAPInit                                                        */
/*                                                                            */
/*  Sends the TCP connection request, as long as there is at least one server */
/*  in the server_list.  This function sets the tunable global variables      */
/*  LFAP_bytes_stats, LFAP_packets_stats, LFAP_poll_interval,                 */
/*  LFAP_batch_size, LFAP_batch_interval, LFAP_lost_conn_interval,            */
/*  LFAP_server_retry_interval, LFAP_max_send_queue_size, and                 */
/*  LFAP_server_list, if stats_type, poll_interval, batch_size,               */
/*  lost_conn_interval, server_retry_interval, max_send_queue_size, or        */
/*  server_list are set respectively.  The server_list has server IP          */
/*  addresses in the form of lfapui32_t.  If a value is not set (LFAP_DEFAULT */
/*  is passed in for the value), then the default value is used for the       */
/*  global variable (excluding the LFAP_server_list).                         */
/*                                                                            */
/******************************************************************************/
void LFAPInit (lfapui16_t stats_type, lfapui32_t poll_interval,
               lfapui16_t batch_size, lfapui32_t batch_interval,
               lfapui32_t lost_conn_interval, lfapui32_t server_retry_interval,
               lfapui32_t max_send_queue_size, lfapui32_t* server_list,
               lfapui16_t server_list_len)
{
    lfap_init_encode();
    lfap_init_decode();
    initializeLfapFlowMgrGlobals_();
    initializeLfapMsgQueues();

    initLfapList(&LFAP_server_list);
    initLfapList(&LFAP_stats_list);
    setLfapStatsType_(stats_type);

    if (poll_interval != LFAP_DEFAULT)
        LFAP_poll_interval = poll_interval;
    else
        LFAP_poll_interval = LFAP_DEFAULT_POLL_INTERVAL;

    if (batch_size != LFAP_DEFAULT)
        LFAP_batch_size = batch_size;
    else
        LFAP_batch_size = LFAP_DEFAULT_BATCH_SIZE;

    setLfapFunUpdateInterval_();

    if (batch_interval != LFAP_DEFAULT)
        LFAP_batch_interval = batch_interval;
    else
        LFAP_batch_interval = LFAP_DEFAULT_BATCH_INTERVAL;

    if (lost_conn_interval != LFAP_DEFAULT)
        LFAP_lost_conn_interval = lost_conn_interval;
    else
        LFAP_lost_conn_interval = LFAP_DEFAULT_LOST_CONN_INTERVAL;

    if (server_retry_interval != LFAP_DEFAULT)
        LFAP_server_retry_interval = server_retry_interval;
    else
        LFAP_server_retry_interval = LFAP_DEFAULT_SRVR_RETRY_INTRVL;

    if (max_send_queue_size != LFAP_DEFAULT)
        LFAP_max_send_queue_size = max_send_queue_size;
    else
        LFAP_max_send_queue_size = LFAP_DEFAULT_MAX_SEND_QUEUE;

    if (server_list_len) {
        if (updateLfapServerList_(server_list, server_list_len))
            updateLfapApiStats_();
    }

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, "LFAPInit()\n");
        fprintf(stderr, "    stats_type                 = ");
        if (stats_type == LFAP_TOTAL_STATS)
            fprintf(stderr, "total\n");
        else
            fprintf(stderr, "delta\n");
        fprintf(stderr, "    LFAP_poll_interval         = %lu\n",
                             LFAP_poll_interval);
        fprintf(stderr, "    LFAP_batch_size            = %u\n",
                             LFAP_batch_size);
        fprintf(stderr, "    LFAP_batch_interval        = %lu\n",
                             LFAP_batch_interval);
        fprintf(stderr, "    LFAP_lost_conn_interval    = %lu\n",
                             LFAP_lost_conn_interval);
        fprintf(stderr, "    LFAP_server_retry_interval = %lu\n",
                             LFAP_server_retry_interval);
        fprintf(stderr, "    LFAP_max_send_queue_size   = %lu\n",
                             LFAP_max_send_queue_size);
        fprintf(stderr, "    server_list                = ");
        if (server_list_len) {
            if (server_list_len == 1)
                fprintf(stderr, "1 entry\n");
            else
                fprintf(stderr, "%d entries\n",server_list_len);
        }
        else
            fprintf(stderr, "empty\n");
        }
        #endif
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPUpdateParameters                                            */
/*                                                                            */
/*  Sets the tunable global variables LFAP_bytes_stats, LFAP_packets_stats,   */
/*  LFAP_poll_interval, LFAP_batch_size, LFAP_batch_interval,                 */
/*  LFAP_lost_conn_interval, LFAP_server_retry_interval,                      */
/*  LFAP_max_send_queue_size, and LFAP_server_list, if stats_type,            */
/*  poll_interval, batch_size, batch_interval, lost_conn_interval,            */
/*  server_retry_interval, max_send_queue_size, or server_list are set        */
/*  respectively.  The server_list has server IP addresses in the form of     */
/*  lfapui32_t.  If a value is not set (LFAP_DEFAULT is passed in for the     */
/*  value), then the default value is used for the global variable (excluding */
/*  the LFAP_server_list).                                                    */
/*                                                                            */
/******************************************************************************/
void LFAPUpdateParameters (lfapui16_t stats_type, lfapui32_t poll_interval,
                           lfapui16_t batch_size, lfapui32_t batch_interval,
                           lfapui32_t lost_conn_interval,
                           lfapui32_t server_retry_interval,
                           lfapui32_t max_send_queue_size,
                           lfapui32_t* server_list, lfapui16_t server_list_len)
{
    LFAP_AR ar;
    setLfapStatsType_(stats_type);

    if (poll_interval != LFAP_DEFAULT)
        LFAP_poll_interval = poll_interval;
    else
        LFAP_poll_interval = LFAP_DEFAULT_POLL_INTERVAL;

    if (batch_size != LFAP_DEFAULT)
        LFAP_batch_size = batch_size;
    else
        LFAP_batch_size = LFAP_DEFAULT_BATCH_SIZE;

    setLfapFunUpdateInterval_();

    if (batch_interval != LFAP_DEFAULT)
        LFAP_batch_interval = batch_interval;
    else
        LFAP_batch_interval = LFAP_DEFAULT_BATCH_INTERVAL;

    if (lost_conn_interval != LFAP_DEFAULT)
        LFAP_lost_conn_interval = lost_conn_interval;
    else
        LFAP_lost_conn_interval = LFAP_DEFAULT_LOST_CONN_INTERVAL;

    if (server_retry_interval != LFAP_DEFAULT)
        LFAP_server_retry_interval = server_retry_interval;
    else
        LFAP_server_retry_interval = LFAP_DEFAULT_SRVR_RETRY_INTRVL;

    if (max_send_queue_size != LFAP_DEFAULT)
        LFAP_max_send_queue_size = max_send_queue_size;
    else
        LFAP_max_send_queue_size = LFAP_DEFAULT_MAX_SEND_QUEUE;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, "LFAPUpdateParameters()\n");
        fprintf(stderr, "    stats_type                 = ");
        if (stats_type == LFAP_TOTAL_STATS)
            fprintf(stderr, "total\n");
        else
            fprintf(stderr, "delta\n");
        fprintf(stderr, "    LFAP_poll_interval         = %lu\n",
                             LFAP_poll_interval);
        fprintf(stderr, "    LFAP_batch_size            = %u\n",
                             LFAP_batch_size);
        fprintf(stderr, "    LFAP_batch_interval        = %lu\n",
                             LFAP_batch_interval);
        fprintf(stderr, "    LFAP_lost_conn_interval    = %lu\n",
                             LFAP_lost_conn_interval);
        fprintf(stderr, "    LFAP_server_retry_interval = %lu\n",
                             LFAP_server_retry_interval);
        fprintf(stderr, "    LFAP_max_send_queue_size   = %lu\n",
                             LFAP_max_send_queue_size);
        fprintf(stderr, "    server_list                = ");
        if (server_list_len) {
            if (server_list_len == 1)
                fprintf(stderr, "1 entry\n");
            else
                fprintf(stderr, "%d entries\n",server_list_len);
        }
        else
            fprintf(stderr, "empty\n");
    }
    #endif

    if (server_list_len) {
        if (updateLfapServerList_(server_list, server_list_len)) {
            updateLfapApiStats_();
            initLfapAR(&ar);
            packLfapServerListAR_(&ar);
            encodeCompletedLfapAR(&ar);
            if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&ar), getLfapMsgLength(&ar), LFAP_REQ_MSG))
                freeLfapMsg(&ar);
        }
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPTerminate                                                   */
/*                                                                            */
/*  Shuts down the LFAP API.  Cleans up the memory for the LFAP API and       */
/*  disconnects from the server.                                              */
/*                                                                            */
/******************************************************************************/
void LFAPTerminate (void)
{
    int i;
    LFAPNode* node;
    LFAPSendMsg* msg;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "LFAPTerminate()\n");
    #endif

    if (LFAP_send_vector_entries) {
        for (i = 0; i < LFAP_send_vector_entries; i++) {
            node = LFAP_send_vector_nodes[i];
            msg = (LFAPSendMsg*) lfapNodeInfo(node);
            lfap_free(msg->buf);
            returnLfapSendMsg(node);
        }
    }

    terminateLfapMsgQueues();

    /* PAC_REMOVE    if ((LFAP_connected == LFAP_CONNECTION_ESTABLISHED) && LFAP_server)*/
    if ( LFAP_conn_state >= LFAP_TCP_CONNECTION && LFAP_server )
        LFAPDisconnect(LFAP_server);
    LFAP_server = 0;

    clearLfapList(&LFAP_stats_list);
    clearLfapList(&LFAP_server_list);

    /* freeing the stats */
    if (LFAP_stats) lfap_free(LFAP_stats);
    LFAP_stats = 0; LFAP_curr_stats = 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPProcessWork                                                 */
/*                                                                            */
/*  Main work loop.  All major processing is completed here, in chunks.       */
/*                                                                            */
/******************************************************************************/
lfapui16_t LFAPProcessWork (void)
{
    lfapui8_t flush;
    lfapui16_t done = 0;
    lfapui16_t work_completed = LFAP_WORK_COMPLETED;
    lfapui16_t status;
    const LFAP_CB_ENTRYPOINT cb = LFAPBatchIntervalExpired_;

    /* PAC_REMOVE    lfapui16_t batch_size = LFAP_MAX_MSGS_IN_SEND_VECTOR;*/

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "LFAPProcessWork ... \n");
    #endif

    /* have a connection to the FAS? */
    /* PAC_REMOVE    if ((LFAP_connected == LFAP_CONNECTION_LOST) &&*/
    if ( LFAP_conn_state <= LFAP_NO_TCP_CONNECTION &&
         LFAP_connection_timer == LFAP_TIMER_EXPIRED )
        LFAPRetryServerList_(0);

    /* If we don't have a session yet, keep the work to a minumim */
    done =  LFAP_conn_state != LFAP_SESSION_ESTABLISHED ? 1 : done;

    /* try to resend if we were unable to send last time? */
    /* PAC_REMOVE    if ((LFAP_connected == LFAP_CONNECTION_ESTABLISHED) &&*/
    if (!done && LFAP_send_vector_entries) {
        if (LFAPSendVector_() == LFAP_SEND_FAILURE) {
            work_completed = LFAP_WORK_INCOMPLETE;
            /* we're not able to send to keep work to a minumum */
            done = 1;
        }
    }

    /* if our output queue is full, keep the work to a minimun */
    if ((msgsInLfapRegSendQueue() + msgsInLfapHighSendQueue()) >= LFAP_max_send_queue_size)
        done = 1;
    
    /* receive high-q msgs */
    if (msgsInLfapHighRcvdQueue())
        LFAPReceiveHighQueueMsgs_();

    /* Check the state after each piece of work */
    done =  LFAP_conn_state != LFAP_SESSION_ESTABLISHED ? 1 : done;

    /* send FUN updates for flows */
    if (!done && 
        ((msgsInLfapRegSendQueue() + msgsInLfapHighSendQueue()) < (LFAP_max_send_queue_size/4)) &&
        (LFAP_flow_update_timer == LFAP_TIMER_EXPIRED)) 
    {
      sendLfapFlowUpdates_();
    }
    
    /* Check the state after each piece of work */
    done =  LFAP_conn_state != LFAP_SESSION_ESTABLISHED ? 1 : done;


    /* report stats on certain flows */
    if (!done && LFAP_report_on_flows != LFAP_NOT_REPORTING_FLOWS) {
        if (reportLfapFlows_() != LFAP_NOT_REPORTING_FLOWS)
            work_completed = LFAP_WORK_INCOMPLETE;
    }

    /* Check the state after each piece of work */
    done =  LFAP_conn_state != LFAP_SESSION_ESTABLISHED ? 1 : done;

    /* receive reg-q msgs */
    if (!done && msgsInLfapRegRcvdQueue())
        LFAPReceiveRegQueueMsgs_();

    /* Check the state after each piece of work */
    done =  LFAP_conn_state != LFAP_SESSION_ESTABLISHED ? 1 : done;

    /* check to see if we have lost connection */
    if (LFAP_lost_conn_timer == LFAP_TIMER_EXPIRED)
      LFAPLostConnInterval_();

    /* Check the state after each piece of work */
    done =  LFAP_conn_state != LFAP_SESSION_ESTABLISHED ? 1 : done;

    /* send msgs */
    /* PAC REMOVE   if (LFAP_batch_size < batch_size) batch_size = LFAP_batch_size; */
    switch ( LFAP_conn_state ) {
      case LFAP_INITIAL_STATE:
      case LFAP_NO_TCP_CONNECTION:
        /* can't send any messages */
        break;

      case LFAP_TCP_CONNECTION:
      case LFAP_PROTOCOL_NEGOTIATION:
      case LFAP_SECURITY_CHECK:
      case LFAP_DATA_EXPORT_NEGOTIATION:
        /* Only send high priority messages during this phase. High
         * priority messages are always sent rigth away.
         */
        status = LFAPSendHighQueueMsgs_();
        if (status == LFAP_WORK_INCOMPLETE )
          work_completed = LFAP_WORK_INCOMPLETE;
        break;

     case LFAP_SESSION_ESTABLISHED:
        status = LFAPSendHighQueueMsgs_();
        if (status == LFAP_WORK_INCOMPLETE )
          work_completed = LFAP_WORK_INCOMPLETE;

        flush = LFAP_batch_interval_timer == LFAP_TIMER_EXPIRED ? 1 : 0;
        status = LFAPSendRegQueueMsgs_(flush);
        if (status == LFAP_WORK_INCOMPLETE )
          work_completed = LFAP_WORK_INCOMPLETE;
        
        if (LFAP_batch_interval_timer == LFAP_TIMER_EXPIRED) {
          LFAPSetSystemTimer(cb, 0, LFAP_batch_interval);
          LFAP_batch_interval_timer = LFAP_TIMER_WAITING;
        }

        break;
      default:
        if (lfap_debug & LFAP_DBG_WARNINGS )
          printf("LFAP - code error - Unknown state in LFAPPrcessWork\n");
        
    }
    
    /* Check the state after each piece of work */
    done =  LFAP_conn_state != LFAP_SESSION_ESTABLISHED ? 1 : done;

    
    /* PAC_REMOVE    if (LFAP_connected == LFAP_CONNECTION_ESTABLISHED) {*/
    /* PAC_REMOVE    {*/
    /* PAC_REMOVE        if (msgsInLfapHighSendQueue()) {*/
    /* PAC_REMOVE            if (LFAPSendBothQueueMsgs_() == LFAP_WORK_INCOMPLETE)*/
    /* PAC_REMOVE                work_completed = LFAP_WORK_INCOMPLETE;*/
    /* PAC_REMOVE        }*/
    /* PAC_REMOVE        else if ((msgsInLfapRegSendQueue() >= batch_size) ||*/
    /* PAC_REMOVE                 (LFAP_batch_interval_timer == LFAP_TIMER_EXPIRED)) {*/
    /* PAC_REMOVE            if (LFAPSendRegQueueMsgs_() == LFAP_WORK_INCOMPLETE)*/
    /* PAC_REMOVE                work_completed = LFAP_WORK_INCOMPLETE;*/
    /* PAC_REMOVE        }*/
    /* PAC_REMOVE    }*/

    /* GARBAGE_COLLECTOR traverse through flow list, removing deleted flows */
    if (LFAP_garbage_collector_timer == LFAP_TIMER_EXPIRED) {
        if (LFAPGarbageCollector_() == LFAP_WORK_INCOMPLETE)
            work_completed = LFAP_WORK_INCOMPLETE;
    }

    /* PAC_REMOVE    if (msgsInLfapRegSendQueue() >= batch_size )*/
    /* PAC_REMOVE     if (msgsInLfapRegSendQueue() >= LFAP_batch_size || */
    /* PAC_REMOVE         msgsInLfapRegSendQueue() >= LFAP_MAX_MSGS_IN_SEND_VECTOR ) */
    /* PAC_REMOVE         work_completed = LFAP_WORK_INCOMPLETE; */

    return work_completed;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetFlowId                                                   */
/*                                                                            */
/*  Returns the next LFAP_cce_id flow_id to the CCE, as long as the           */
/*  LFAP_fas_id has already been initialized and the LFAP_server is up.  If   */
/*  either one is not true, 0 is returned (invalid flow_id).                  */
/*                                                                            */
/******************************************************************************/
lfapui32_t LFAPGetFlowId (void)
{
    if (LFAP_fas_hi && LFAP_fas_lo) {
        LFAP_cce_id++;

        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_FUNCTIONS)
            fprintf(stderr, " Returning A Flow ID: %lu\n\n", LFAP_cce_id);
        #endif

        return LFAP_cce_id;
    }
    else
        return 0;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPAdmitFlow                                                   */
/*                                                                            */
/*  Sends the LFAP_FAR message to the LFAP server starting the flow.          */
/*                                                                            */
/******************************************************************************/
void LFAPAdmitFlow (lfapui32_t flow_id, LFAPFlowInfo* flow_info)
{
    LFAP_FAR lfap_far;
    int type, i;
    
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "LFAPAdmitFlow()\n");
    #endif

    if (flow_id) {
        initLfapFARflowId(&lfap_far, LFAP_fas_hi, LFAP_fas_lo, flow_id);
        setTimeLfapFAR(&lfap_far, LFAPGetSysUptime());

        if ( flow_info->src_len )
          setSrcAddrLfapFAR(&lfap_far, flow_info->src_addr, flow_info->src_fam,
                            flow_info->src_len);

        if ( flow_info->dest_len )
          setDestAddrLfapFAR(&lfap_far, flow_info->dest_addr, flow_info->dest_fam,
                             flow_info->dest_len);
        if (flow_info->src_port)
            setSrcPortLfapFAR(&lfap_far, flow_info->src_port_type,
                               flow_info->src_port);
        if (flow_info->protocol[0])
            setProtocolLfapFAR(&lfap_far, flow_info->protocol,
                                LFAP_MAX_PROTOCOL_LEN);
        if (flow_info->client_data_len)
            setClientDataLfapFAR(&lfap_far, (char*) flow_info->client_data,
                                  flow_info->client_data_len);
        if (flow_info->src_cce_len)
            setSrcCCEAddrLfapFAR(&lfap_far, flow_info->src_cce_addr,
                                flow_info->src_cce_fam, flow_info->src_cce_len);
        if (flow_info->src_as != LFAP_AS_NUM_INVALID)
            setSrcASLfapFAR(&lfap_far, flow_info->src_as);
        if (flow_info->dest_as != LFAP_AS_NUM_INVALID)
            setDestASLfapFAR(&lfap_far, flow_info->dest_as);

        if (flow_info->ingress_port)
            setIngressPortLfapFAR(&lfap_far, flow_info->ingress_port);
        if (flow_info->egress_port)
            setEgressPortLfapFAR(&lfap_far, flow_info->egress_port);

        /* Since we have the protocol defined value for no vlan ID we
           can pass both directly to the set routine with no futher checks.
           But if both are nto there, don't skip the IE.
        */
        if (flow_info->src_vlan_id != LFAP_NO_VLAN_ID || 
            flow_info->dst_vlan_id != LFAP_NO_VLAN_ID)
            setVlanIDLfapFAR(&lfap_far, flow_info->src_vlan_id, flow_info->dst_vlan_id);

        /* Set the Class of Service IEs, there could be
         * several in one message.
         */

        /* Cycle through each type (e.g. IPv4, IPv6, MPLS, etc... */
        for( type = 1; type <= LFAP_COS_MAX_TYPES; type++) {
            if ( flow_info->CoS_Arr[type][LFAP_COS_RECEIVED] != LFAP_COS_INVALID )
              setCoSLfapFAR(&lfap_far, type, LFAP_COS_RECEIVED, flow_info->CoS_Arr[type][LFAP_COS_RECEIVED]);

            if ( flow_info->CoS_Arr[type][LFAP_COS_TRANSMITTED] != LFAP_COS_INVALID )
              setCoSLfapFAR(&lfap_far, type, LFAP_COS_TRANSMITTED, flow_info->CoS_Arr[type][LFAP_COS_TRANSMITTED]);
        }
        

        if (flow_info->checkpoint != LFAP_QUALIFIER_INVALID)
            setCheckpointLfapFAR(&lfap_far, flow_info->checkpoint);
        if (flow_info->cxn_priority != LFAP_QUALIFIER_INVALID)
            setCxnPriorityLfapFAR(&lfap_far, flow_info->cxn_priority);

        for(i = 0; i < LFAP_MAX_FECs; i++) {
          if ( flow_info->ldpFECLen[i] != 0 )
            setLDPFECLfapFAR(&lfap_far, flow_info->ldpFEC[i], flow_info->ldpFECLen[i]);
        }

        for(i = 0; i < LFAP_MAX_LABELS; i++) {
          if (flow_info->inMplsLabel[i] != 0)
            setMplsLabelLfapFAR(&lfap_far, LFAP_IN_LABEL, flow_info->inMplsLabel[i]);
          if (flow_info->outMplsLabel[i] != 0)
            setMplsLabelLfapFAR(&lfap_far, LFAP_OUT_LABEL, flow_info->outMplsLabel[i]);
        }

        if (flow_info->next_hop_len)
            setNextHopAddrLfapFAR(&lfap_far, flow_info->next_hop_addr,
                                flow_info->next_hop_fam, flow_info->next_hop_len);
        if (LFAP_curr_stats) {
            LFAP_curr_stats->flow_setups++;
            LFAP_curr_stats->active_flows++;
            if (LFAP_curr_stats->peak_active_flows <
                LFAP_curr_stats->active_flows)
                LFAP_curr_stats->peak_active_flows =
                                                  LFAP_curr_stats->active_flows;
        }

        #ifdef CS_DEBUG
        if (lfap_debug & LFAP_DBG_MESSAGES)
            PRINT_LFAP_MESSAGE(&lfap_far, 0, LFAP_SEND);
        #endif

/*      SSR - don't add the flow to the stats list here because we are on the flow setup task 
        and we don't want to add the cost of this call on this task. Let the LFAP
        task do it.
        setLfapCheckpointTimer_(flow_id);
*/

        encodeCompletedLfapFAR(&lfap_far);
        if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&lfap_far),
                                      getLfapMsgLength(&lfap_far),
                                      LFAP_OPT_MSG)) {
            freeLfapMsg(&lfap_far);
        }
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPFlowDeleted                                                 */
/*                                                                            */
/*  The CCE calls this to remove the flow.  A final LFAP_FUN message is       */
/*  sent.                                                                     */
/*                                                                            */
/******************************************************************************/
void LFAPFlowDeleted (lfapui32_t flow_id, LFAPFlowStats* flow_stats)
{
    LFAP_FUN fun;

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "LFAPFlowDeleted(): %lu\n",flow_id);
    #endif

    /* API Stats */
    if (LFAP_curr_stats) {
        LFAP_curr_stats->flow_teardowns++;
        LFAP_curr_stats->active_flows--;
    }

    /* GARBAGE_COLLECTOR - let the garabge collector delete it. See notes on
       garbage collector routine (search for GARBAGE_COLLECTOR)
    deleteLfapCheckpointTimer_(flow_id); */

    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, "SENDING flow Statistics 1 final time, flow: ");
        fprintf(stderr, "%lu\n", flow_id);
    }
    #endif

    initLfapFUNflowId(&fun, LFAP_fas_hi, LFAP_fas_lo, flow_id);
    setTimeLfapFUN(&fun, LFAPGetSysUptime());
    setStateLfapFUN(&fun, LFAP_INACTIVE);
    setLfapFUNStats_(&fun, flow_stats);
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_MESSAGES)
        PRINT_LFAP_MESSAGE(&fun, 0, LFAP_SEND);
    #endif

    encodeCompletedLfapFUN(&fun);
    if (!addMsgToLfapSendQueue(getLfapMsgBuffer(&fun),
                               getLfapMsgLength(&fun),
                               LFAP_OPT_MSG)) {

        /* API stats */
        if (LFAP_curr_stats) {
          LFAP_curr_stats->dropped_fun_i_msgs++;
          LFAP_curr_stats->lost_bytes_sent  += flow_stats->byte_sent;
          LFAP_curr_stats->lost_bytes_rcvd   += flow_stats->byte_rcvd;
          LFAP_curr_stats->lost_packets_sent += flow_stats->packet_sent;
          LFAP_curr_stats->lost_packets_rcvd += flow_stats->packet_rcvd;
        }

        freeLfapMsg(&fun);
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPConnectionLost                                              */
/*                                                                            */
/*  This is called when a the TCP connection from the CCE to the server is    */
/*  lost.                                                                     */
/*                                                                            */
/******************************************************************************/
void LFAPConnectionLost (lfapui32_t server_ip)
{
    #ifdef CS_DEBUG
    char buffer[256];

    if (lfap_debug & LFAP_DBG_FUNCTIONS) {
        fprintf(stderr, "LFAPConnectionLost:  ");
        sprintf(buffer, "%d.%d.%d.%d",
               (int) (server_ip >> 24) & 0xff,
               (int) (server_ip >> 16) & 0xff,
               (int) (server_ip >>  8) & 0xff,
               (int) (server_ip >>  0) & 0xff);
        fprintf(stderr, "%s\n", buffer);
    }
    #endif

    /* PAC_REMOVE    LFAP_connected = LFAP_CONNECTION_LOST;*/
    /* PAC_REMOVE    LFAP_conn_status = LFAP_CONNECTION_LOST;*/
    LFAP_conn_state = LFAP_NO_TCP_CONNECTION;
    removeLfapMsgs_();
    LFAP_lost_conn_count++;

    /* if lost connection timer expires, don't want to disconnect a 2nd time */
    LFAP_msgs_rcvd++;

    if (LFAP_curr_stats && LFAP_curr_stats->conn_time) {
        LFAP_curr_stats->up_time = LFAP_curr_stats->up_time + (LFAPGetSysUptime() -
                                   LFAP_curr_stats->conn_time);
        LFAP_curr_stats->conn_time = 0;
    }

    if (LFAP_connection_timer == LFAP_TIMER_INACTIVE) {
        LFAP_connection_timer = LFAP_TIMER_EXPIRED;

        #ifdef CS_DEBUG
        if (LFAP_server && (LFAP_server != server_ip) &&
           (lfap_debug & LFAP_DBG_ERRORS)) {
            /* SSR - error log */
            ERR_errlog(0, lfap_efacid, LFAP_E_DIFF_SERVERS, 2, LFAP_server,
                       server_ip);
        }
        #endif
    }
    #ifdef CS_DEBUG
    else if (lfap_debug & LFAP_DBG_TIMERS) {
        if (LFAP_connection_timer == LFAP_TIMER_WAITING)
            fprintf(stderr, "  Waiting for notification by a timer.\n");
        else if (LFAP_connection_timer == LFAP_TIMER_EXPIRED)
            fprintf(stderr, "  Timer has already expired.\n");
        else
            fprintf(stderr, "  Currently processing the timer.\n");
    }
    #endif
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPReceiveMessage                                              */
/*                                                                            */
/*  Receives a message from the server and gives it to the appropriate        */
/*  function to handle it.  buf_len includes the LFAP_HEADER_SIZE.            */
/*                                                                            */
/******************************************************************************/
void LFAPReceiveMessage (void* buf, int buf_len)
{
    void* ibuf;
    MsgPri_t pri;
    lfapui8_t version = 0;
    lfapui16_t msg_len = 0;

    msg_len = lfap_msg_get_msglen(buf) + LFAP_HEADER_SIZE;
    #ifdef CS_DEBUG
    if (lfap_debug & LFAP_DBG_FUNCTIONS)
        fprintf(stderr, "LFAPReceiveMessage, length: %d\n\n", msg_len);
    #endif

    if (msg_len != buf_len) {
        if (msg_len > buf_len) {
            LFAP_error.error = LFAP_RCVD_PARTIAL_MESSAGE;
            LFAP_error.time = LFAPGetSysUptime();

            /* API Stats */
            if (LFAP_curr_stats)
                LFAP_curr_stats->msgs_rcvd_error++;

            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_ERRORS) {
                /* SSR - error log */
                ERR_errlog(0, lfap_efacid, LFAP_E_PART_MSG_RCVD, 2, buf_len,
                           msg_len);
            }
            #endif
        }
        else {
            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_WARNINGS) {
                /* SSR - error log */
                ERR_errlog(0, lfap_efacid, LFAP_W_MSG_SHORT, 2, buf_len,
                           msg_len);
            }
            #endif
        }
    }

    if (msg_len <= buf_len) {
        /* API Stats */
        if (LFAP_curr_stats)
            LFAP_curr_stats->total_bytes_rcvd += msg_len;

        /* still need to do this, for now .. check later */
        ibuf = (void*) lfap_malloc(msg_len);
        lfap_memcpy(ibuf, buf, msg_len);

        version = lfap_msg_get_version(buf);
        if (version < LFAP_VERSION) {
            LFAP_error.error = LFAP_INCOMPATIBLE_VERSION;
            LFAP_error.time = LFAPGetSysUptime();

            /* API Stats */
            if (LFAP_curr_stats)
                LFAP_curr_stats->msgs_rcvd_error++;

            #ifdef CS_DEBUG
            if (lfap_debug & LFAP_DBG_ERRORS) {
                /* SSR - error log */
                ERR_errlog(0, lfap_efacid, LFAP_E_BAD_VERSION, 2, version,
                           LFAP_VERSION);
            }
            #endif
            lfap_free(ibuf);
        }
        else {
            /* API Stats */
            if (LFAP_curr_stats) {
                if (lfap_msg_get_status(buf) == LFAP_MSG_SUCCESS)
                    LFAP_curr_stats->msgs_rcvd_success++;
                else if (lfap_msg_get_status(buf) == LFAP_MSG_CORRUPTED) {
                    #ifdef CS_DEBUG
                    if (lfap_debug & LFAP_DBG_WARNINGS) {
                        /* SSR - error log */
                        ERR_errlog(0, lfap_efacid, LFAP_W_MSG_CORRUPT, 2,
                                   lfap_msg_get_opcode(buf),
                                   lfap_msg_get_msgid(buf));
                    }
                    #endif
                    LFAP_curr_stats->msgs_rcvd_corrupt++;
                }
                else
                    LFAP_curr_stats->msgs_rcvd_error++;
            }

            /* lost connection? */
            LFAP_msgs_rcvd++;

            /* regular or high rcvd queue? */
            pri = LFAPMsgPriority_(buf, buf_len);
            switch (pri) {
              case High:                
                if (!addMsgToLfapHighRcvdQueue(ibuf, msg_len)) {
                  if (LFAP_curr_stats) LFAP_curr_stats->dropped_rcvd_msgs++;
                  lfap_free(ibuf);
                }
                break;
              case Low:
              default:          
                if (!addMsgToLfapRegRcvdQueue(ibuf, msg_len)) {
                  if (LFAP_curr_stats) LFAP_curr_stats->dropped_rcvd_msgs++;
                  lfap_free(ibuf);
                }
            }
        }
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPDecodeMsgLen                                                */
/*                                                                            */
/*  Decodes the length field of an LFAP message header.  This enables an      */
/*  incoming TCP/IP stream to be divided into individual LFAP messages for    */
/*  processing.  The 'buf' parameter should point to the beginning of an LFAP */
/*  message fragment that is at least LFAP_HEADER_SIZE bytes in length.  This */
/*  function returns the length of the message body in bytes.  This does not  */
/*  include the length of the LFAP message header (which is LFAP_HEADER_SIZE  */
/*  bytes in length).                                                         */
/*                                                                            */
/******************************************************************************/
int LFAPDecodeMsgLen (void* buf)
{
    return lfap_msg_get_msglen(buf);
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetLastError                                                */
/*                                                                            */
/*  Returns the last error and the time it occurred.                          */
/*                                                                            */
/******************************************************************************/
const LFAPError* LFAPGetLastError (void)
{
    return &LFAP_error;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetApiStats                                                 */
/*                                                                            */
/*  Returns the statistics of the LFAP API.                                   */
/*                                                                            */
/******************************************************************************/
const LFAPApiStats* LFAPGetApiStats (void)
{
    lfapui32_t curr_time;

    if (LFAP_curr_stats) {
        if (LFAP_curr_stats->conn_time) {
            curr_time = LFAPGetSysUptime();
            LFAP_curr_stats->up_time = LFAP_curr_stats->up_time +
                                      (curr_time - LFAP_curr_stats->conn_time);
            LFAP_curr_stats->conn_time = curr_time;
        }

        LFAP_curr_stats->msgs_send_buf = msgsInLfapHighSendQueue() +
                                         msgsInLfapRegSendQueue();
        LFAP_curr_stats->msgs_rcvd_buf = msgsInLfapHighRcvdQueue() +
                                         msgsInLfapRegRcvdQueue();
    }

    return LFAP_stats;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetActiveServer                                             */
/*                                                                            */
/*  Returns the IP address of the current server CCE is connected to.         */
/*                                                                            */
/******************************************************************************/
lfapui32_t LFAPGetActiveServer (void)
{
    return LFAP_server;
}

/******************************************************************************/
/*                                                                            */
/*  Function: LFAPGetConnectionStatus                                         */
/*                                                                            */
/*  Returns the status of the connection to a server.                         */
/*                                                                            */
/******************************************************************************/
LFAPConnState_t LFAPGetConnectionStatus (void)
{
  /* PAC_REMOVE    return LFAP_conn_status;*/
  return LFAP_conn_state;
  
}

#endif /* ifndef is all the way at the top */
