/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "mls_stdlib.h"
#endif
/* End of auto-generated includes */

#include "LfapList.h"
#include "LfapSysDepend.h"

/*
 * The LFAP Node
 */

void initLfapNode (LFAPNode* node)
{
    node->_info = 0;
    node->_next = 0;
}

void copyLfapNode (LFAPNode* node, void* new_info)
{
    node->_info = new_info;
    node->_next = 0;
}

LFAPNode* nextLfapNode (LFAPNode* node)
{
    return node->_next;
}

LFAPNode* clearLfapNode (LFAPNode* node)
{
    LFAPNode* new_node = node->_next;
    node->_next = 0;
    return new_node;
}

void* lfapNodeInfo (LFAPNode* node)
{
    if (node)
        return node->_info;
    else
        return 0;
}


/*
 * The LFAP List
 */

static void initLfapList_ (LFAPList* list)
{
    list->_head_._next = list->_tail_._next = &list->_tail_;
    list->_last_ = &list->_head_;
    list->_head_._info = list->_tail_._info = 0;
    list->_entries_ = 0;
}

static const LFAPNode* lfapListAt_ (LFAPList* list, unsigned int i)
{
    register LFAPNode* node;

    if (i >= lfapListEntries(list)) {
        return 0;
    }

    node = list->_head_._next;

    while (i--) {
        node = node->_next;
    }

    return node;
}

static void insertAfterLfapNode_ (LFAPList* list, LFAPNode* l, LFAPNode* x)
{
    x->_next = l->_next;
    l->_next = x;

    if (l == list->_last_) {
        list->_last_ = x;
    }

    list->_entries_++;
}

static LFAPNode* findLfapLeft_ (LFAPList* list, const LFAPNode* x)
{
    LFAPNode* node = &list->_head_;

    while (node->_next != &list->_tail_) {
        if (node->_next == x) {
            return node;
        }

        node = node->_next;
    }

    return 0;
}
	
static LFAPNode* removeLfapRight_ (LFAPList* list, LFAPNode* x)
{
    LFAPNode* node = x->_next;

    x->_next = node->_next;

    if (node == list->_last_) {
        list->_last_ = x;
    }

    list->_entries_--;

    return node;
}

static LFAPNode* removeLfapFirst_ (LFAPList* list)
{
    return list->_entries_ ? removeLfapRight_(list, &list->_head_) : 0;
}

static void* stripLfapNode_ (LFAPNode* node)
{
    void* p;

    if (!node) return 0;

    p = node->_info;
    lfap_free(node);

    return p;
}

void initLfapList (LFAPList* list)
{
    initLfapList_(list);
}

void setLfapList (LFAPList* list, void* x)
{
    LFAPNode* n = (LFAPNode*) lfap_malloc(SIZEOF_LFAP_NODE);
    copyLfapNode(n, x);
    list->_head_._next = list->_last_ = n;
    n->_next = list->_tail_._next = &list->_tail_;
    list->_entries_ = 1;
}

void copyLfapList (LFAPList* list, LFAPList l)
{
    void* p;
    LFAPListIterator iter;
    initLfapIterList(&iter, &l);

    while ((p = incLfapIter(&iter))) {
        lfapListAppend(list, p);
    }
}

void* lfapListAt (LFAPList* list, unsigned int n)
{
    return (void*) lfapListAt_(list, n)->_info;
}

void* lfapListFirst (LFAPList* list)
{
    return (lfapListIsEmpty(list) ? 0 : list->_head_._next->_info);
}

void* lfapListLast (LFAPList* list)
{
    return (lfapListIsEmpty(list) ? 0 : list->_last_->_info);
}

void* lfapListAppend (LFAPList* list, void* x)
{
    LFAPNode* node;
    node = (LFAPNode*) lfap_malloc(SIZEOF_LFAP_NODE);
    copyLfapNode(node, x);

    insertAfterLfapNode_(list, list->_last_, node);

    return x;
}

void* lfapListPrepend (LFAPList* list, void* x)
{
    LFAPNode* node;
    node = (LFAPNode*) lfap_malloc(SIZEOF_LFAP_NODE);
    copyLfapNode(node, x);

    insertAfterLfapNode_(list, &list->_head_, node);

    return x;
}

void* lfapListInsertAt (LFAPList* list, unsigned int i, void* x)
{
    LFAPNode* prev;
    LFAPNode* node;

    if (i > lfapListEntries(list)) {
        return 0;
    }

    prev = i ? (LFAPNode*) lfapListAt_(list, i-1) : &list->_head_;

    node = (LFAPNode*) lfap_malloc(SIZEOF_LFAP_NODE);
    copyLfapNode(node, x);

    insertAfterLfapNode_(list, prev, node);

    return x;
}

void* lfapListInsertAfter (LFAPList* list, unsigned int i, void* x)
{
    LFAPNode* node;

    if (i > lfapListEntries(list)) {
        return 0;
    }

    node = (LFAPNode*) lfap_malloc(SIZEOF_LFAP_NODE);
    copyLfapNode(node, x);

    insertAfterLfapNode_(list, (LFAPNode*)lfapListAt_(list, i), node);

    return x;
}

void* lfapListFindRef (LFAPList* list, const void* x)
{
    LFAPNode* node = list->_head_._next;

    while (node != &list->_tail_) {
        if (node->_info == x) {
            return node->_info;
        }

        node = nextLfapNode(node);
    }

    return 0;
}

void* lfapListRemoveRef (LFAPList* list, const void* x)
{
    LFAPNode* prev = &list->_head_;

    while (prev != list->_last_) {
        if (nextLfapNode(prev)->_info == x) {
            return stripLfapNode_(removeLfapRight_(list, prev));
        }

        prev = nextLfapNode(prev);
    }

    return 0;
}

void* removeNextLfapNode (LFAPList* list, LFAPNode* prev)
{
    return stripLfapNode_(removeLfapRight_(list, prev));
}

void* lfapListGet (LFAPList* list)
{
    return list->_entries_ ? stripLfapNode_(removeLfapFirst_(list)) : 0;
}

unsigned int lfapListEntries (LFAPList* list)
{
    return list->_entries_;
}

unsigned int lfapListIsEmpty (LFAPList* list)
{
    return (list->_entries_ == 0);
}

void clearLfapList (LFAPList* list)
{
    LFAPNode* n;
    LFAPNode* node = list->_head_._next;

    while (node != &list->_tail_) {
        n = nextLfapNode(node);
        lfap_free(node);
        node = n;
    }

    initLfapList_(list);
}

void clearAndDestroyLfapList (LFAPList* list)
{
    LFAPNode* n;
    LFAPNode* node = list->_head_._next;

    while (node != &list->_tail_) {
        n = nextLfapNode(node);
        lfap_free(node->_info);
        lfap_free(node);
        node = n;
    }

    initLfapList_(list);
}

LFAPNode* lastLfapNode (LFAPList* list)
{
    return list->_last_;
}

LFAPNode* tailLfapNode (LFAPList* list)
{
    return &list->_tail_;
}

LFAPNode* firstLfapNode (LFAPList* list)
{
    return list->_head_._next;
}

LFAPNode* getFirstLfapNode (LFAPList* list)
{
    return list->_entries_ ? removeLfapFirst_(list) : 0;
}

LFAPNode* findLfapNode (LFAPList* list, const void* x)
{
    LFAPNode* node = list->_head_._next;

    while (node != &list->_tail_) {
        if (node->_info == x) {
            return node;
        }

        node = nextLfapNode(node);
    }

    return 0;
}

LFAPNode* previousLfapNode (LFAPList* list, const void* x)
{
    LFAPNode* prev = &list->_head_;

    while (prev != list->_last_) {
        if (nextLfapNode(prev)->_info == x) {
            return prev;
        }

        prev = nextLfapNode(prev);
    }

    return 0;
}

void appendLfapNode (LFAPList* list, LFAPNode* node)
{
    insertAfterLfapNode_(list, list->_last_, node);
}


/*
 * The LFAP List Iterator
 */

static void lfapIterForward_ (LFAPListIterator* i)
{
    i->_c_ = i->_c_->_next;
}

static unsigned int lfapIterActive_ (LFAPListIterator* i)
{
    return i->_c_ != &i->_l_->_head_ && i->_c_ != &i->_l_->_tail_;
}

void initLfapIter (LFAPListIterator* i)
{
    i->_l_ = 0;
    i->_c_ = 0;
}

void initLfapIterList (LFAPListIterator* i, LFAPList* l)
{
    i->_l_ = l;
    i->_c_ = &l->_head_;
}

/* operator ++, or () */
void* incLfapIter (LFAPListIterator* i)
{
    lfapIterForward_(i);

     return (i->_c_ == &i->_l_->_tail_ ? 0 : i->_c_->_info);
}

/* operator += */
void* plusLfapIter (LFAPListIterator* i, unsigned int n)
{
    void* p = keyItemLfapIter(i);

    while (n--) p = incLfapIter(i);

    return p;
}

void* keyItemLfapIter (LFAPListIterator* i)
{
    return (lfapIterActive_(i) ? i->_c_->_info : 0);
}

void* removeItemLfapIter (LFAPListIterator* i)
{
    if (!lfapIterActive_(i)) return 0;

    i->_c_ = findLfapLeft_(i->_l_, i->_c_);

    return stripLfapNode_(removeLfapRight_(i->_l_, i->_c_));
}

void resetLfapIter (LFAPListIterator* i)
{
    i->_c_ = &i->_l_->_head_;
}

void resetLfapIterList (LFAPListIterator* i, LFAPList l)
{
    i->_l_ = &l;
    resetLfapIter(i);
}
