/*******************************************************************************
**
**    Cabletron Systems Incorporated
**    Post Office Box 5005
**    Rochester, NH 03866-5005
**    (c) Copyright Cabletron Systems Inc. 1999
**
**    Workfile:             %M%
**    Logfile:              %P%
**    Original Author:      Todd Crowley
**    SCCS Delta ID:        %I%
**    Last Delta Date/Time: %G% %U%
**
*******************************************************************************/
/* 
 * WARNING WARNING
 *
 * since there are currently no clients on a windows platform
 * and since we don't want this file included when the windows
 * server is built under Visual C++, put in this ifdef
*/
#ifndef _WIN32

/* Start of auto-generated includes */
#define INCLUDES_H_
#ifdef MLS_SOL_BUILD
#include "sol_includes.h"
#endif
#ifdef MLS_MIPS_BUILD
#include "nucleus.h"
#include "mls_stdlib.h"
#endif
/* End of auto-generated includes */

#include "LfapMsgQueue.h"


static LFAPList LFAP_send_free_list;
static LFAPList LFAP_send_reg_queue;
static LFAPList LFAP_send_high_queue;

static LFAPList LFAP_rcvd_free_list;
static LFAPList LFAP_rcvd_reg_queue;
static LFAPList LFAP_rcvd_high_queue;

/* current statistics */
extern LFAPApiStats* LFAP_curr_stats;

/* max send queue size */
extern lfapui32_t LFAP_max_send_queue_size;


/******************************************************************************/
/*                                                                            */
/*  Function: newLfapMsgNode_                                                 */
/*                                                                            */
/*  Returns a new LFAPNode structure, which has a zero-initialized            */
/*  LFAPMsgHolder structure for data.                                         */
/*                                                                            */
/******************************************************************************/
static LFAPNode* newLfapMsgNode_ (void)
{
    LFAPNode* node;
    LFAPMsgHolder* msg;

    node = (LFAPNode*) lfap_malloc(SIZEOF_LFAP_NODE);
    msg = (LFAPMsgHolder*)lfap_malloc(SIZEOF_LFAP_MSG_HOLDER);
    msg->buf = 0; msg->len = 0;
    copyLfapNode(node, msg);

    return node;
}

/******************************************************************************/
/*                                                                            */
/*  Function: freeLfapMsgNode_                                                */
/*                                                                            */
/*  Frees a LFAPNode structure.  Also frees the LFAPMsgHolder structure in    */
/*  the LFAPNode's data.                                                      */
/*                                                                            */
/******************************************************************************/
static void freeLfapMsgNode_ (LFAPNode* node)
{
    lfap_free(node->_info);
    lfap_free(node);
}

/******************************************************************************/
/*                                                                            */
/*  Function: getFreeLfapSendMsgNode_                                         */
/*                                                                            */
/*  Returns a LFAPNode structure for one of the send queues, which has a      */
/*  zero-initialized LFAPMsgHolder structure for data.  0 is returned if the  */
/*  send queues are at maximum capacity.                                      */
/*                                                                            */
/******************************************************************************/
static LFAPNode* getFreeLfapSendMsgNode_ (int status)
{
    LFAPNode* free_node = 0;
    unsigned int entries =
                 LFAP_send_reg_queue._entries_ + LFAP_send_high_queue._entries_;

    if ((entries < LFAP_max_send_queue_size) || (status == LFAP_REQ_MSG)) {
        if (LFAP_send_free_list._entries_)
            free_node = getFirstLfapNode(&LFAP_send_free_list);
        else
            free_node = newLfapMsgNode_();
    }

    return free_node;
}

/******************************************************************************/
/*                                                                            */
/*  Function: getFreeLfapRcvdMsgNode_                                         */
/*                                                                            */
/*  Returns a LFAPNode structure for one of the rcvd queues, which has a      */
/*  zero-initialized LFAPMsgHolder structure for data.  0 is returned if the  */
/*  rcvd queues are at maximum capacity.                                      */
/*                                                                            */
/******************************************************************************/
static LFAPNode* getFreeLfapRcvdMsgNode_ (void)
{
    LFAPNode* free_node = 0;
    unsigned int entries =
                 LFAP_rcvd_reg_queue._entries_ + LFAP_rcvd_high_queue._entries_;

    if (entries < LFAP_max_send_queue_size) {
        if (LFAP_rcvd_free_list._entries_)
            free_node = getFirstLfapNode(&LFAP_rcvd_free_list);
        else
            free_node = newLfapMsgNode_();
    }

    return free_node;
}

/******************************************************************************/
/*                                                                            */
/*  Function: terminateLfapMsgQueue_                                          */
/*                                                                            */
/*  Destroys a message queue, cleaning up all memory in the queue.            */
/*                                                                            */
/******************************************************************************/
static void terminateLfapMsgQueue_ (LFAPList* queue)
{
    LFAPNode* n;
    LFAPMsgHolder* msg;
    LFAPNode* node = queue->_head_._next;

    while (node != &queue->_tail_) {
        n = nextLfapNode(node);
        msg = (LFAPMsgHolder*) node->_info;
        lfap_free(msg->buf);
        lfap_free(msg);
        lfap_free(node);
        node = n;
    }
}

/******************************************************************************/
/*                                                                            */
/*  Function: addMsgToLfapMsgQueue_                                           */
/*                                                                            */
/*  Adds a message to the appropriate queue.  The message should already be   */
/*  allocated.  Message is appended to the end of the queue.  Returns 1 if    */
/*  successful, 0 if not.                                                     */
/*                                                                            */
/******************************************************************************/
static lfapui16_t addMsgToLfapMsgQueue_ (LFAPList* queue, LFAPNode* free_node,
                                         void* buf, unsigned int len)
{
    LFAPMsgHolder* msg;
    lfapui16_t msg_added = 0;

    if (free_node) {
        msg_added = 1;
        msg = (LFAPMsgHolder*) free_node->_info;
        msg->buf = buf;
        msg->len = len;

        appendLfapNode(queue, free_node);
    }

    return msg_added;
}

/******************************************************************************/
/*                                                                            */
/*  Function: initializeLfapMsgQueues                                         */
/*                                                                            */
/*  Initializes the send and received message queues.                         */
/*                                                                            */
/******************************************************************************/
void initializeLfapMsgQueues (void)
{
    initLfapList(&LFAP_send_free_list);
    initLfapList(&LFAP_send_reg_queue);
    initLfapList(&LFAP_send_high_queue);

    initLfapList(&LFAP_rcvd_free_list);
    initLfapList(&LFAP_rcvd_reg_queue);
    initLfapList(&LFAP_rcvd_high_queue);
}

/******************************************************************************/
/*                                                                            */
/*  Function: terminateLfapMsgQueues                                          */
/*                                                                            */
/*  Terminates the send and received message queues.                          */
/*                                                                            */
/******************************************************************************/
void terminateLfapMsgQueues (void)
{
    clearAndDestroyLfapList(&LFAP_send_free_list);
    terminateLfapMsgQueue_(&LFAP_send_reg_queue);
    terminateLfapMsgQueue_(&LFAP_send_high_queue);

    clearAndDestroyLfapList(&LFAP_rcvd_free_list);
    terminateLfapMsgQueue_(&LFAP_rcvd_reg_queue);
    terminateLfapMsgQueue_(&LFAP_rcvd_high_queue);
}



/******************************************************************************/
/*                                                                            */
/*  Function: addMsgToLfapRegSendQueue                                        */
/*                                                                            */
/*  Adds a message to the regular send queue.  The message should already     */
/*  be allocated.  Message is appended to the end of the queue.  Returns 1 if */
/*  successful, 0 if not.                                                     */
/*                                                                            */
/******************************************************************************/
lfapui16_t addMsgToLfapRegSendQueue (void* buf, unsigned int len, int status)
{
    lfapui16_t added = 0;

    int ov = LFAPLock();

    added = addMsgToLfapMsgQueue_(&LFAP_send_reg_queue,
                                     getFreeLfapSendMsgNode_(status), buf, len);
    if (LFAP_curr_stats) {
        if (LFAP_curr_stats->peak_msgs_send_buf <
           (LFAP_send_reg_queue._entries_ + LFAP_send_high_queue._entries_))
            LFAP_curr_stats->peak_msgs_send_buf =
                 LFAP_send_reg_queue._entries_ + LFAP_send_high_queue._entries_;
    }

    LFAPUnlock(ov);
    return added;
}

/******************************************************************************/
/*                                                                            */
/*  Function: addMsgToLfapHighSendQueue                                       */
/*                                                                            */
/*  Adds a message to the high send queue.  The message should already        */
/*  be allocated.  Message is appended to the end of the queue.  Returns 1 if */
/*  successful, 0 if not.                                                     */
/*                                                                            */
/******************************************************************************/
lfapui16_t addMsgToLfapHighSendQueue (void* buf, unsigned int len, int status)
{
    lfapui16_t added = 0;

    int ov = LFAPLock();

    added = addMsgToLfapMsgQueue_(&LFAP_send_high_queue,
                                     getFreeLfapSendMsgNode_(status), buf, len);
    if (LFAP_curr_stats) {
        if (LFAP_curr_stats->peak_msgs_send_buf <
           (LFAP_send_reg_queue._entries_ + LFAP_send_high_queue._entries_))
            LFAP_curr_stats->peak_msgs_send_buf =
                 LFAP_send_reg_queue._entries_ + LFAP_send_high_queue._entries_;
    }

    LFAPUnlock(ov);
    return added;
}

/******************************************************************************/
/*                                                                            */
/*  Function: msgsInLfapRegSendQueue                                          */
/*                                                                            */
/*  Returns the number of messages currently in the send received queue.      */
/*                                                                            */
/******************************************************************************/
unsigned int msgsInLfapRegSendQueue (void)
{
    return LFAP_send_reg_queue._entries_;
}

/******************************************************************************/
/*                                                                            */
/*  Function: msgsInLfapHighSendQueue                                         */
/*                                                                            */
/*  Returns the number of messages currently in the high send queue.          */
/*                                                                            */
/******************************************************************************/
unsigned int msgsInLfapHighSendQueue (void)
{
    return LFAP_send_high_queue._entries_;
}

/******************************************************************************/
/*                                                                            */
/*  Function: getMsgFromLfapRegSendQueue                                      */
/*                                                                            */
/*  Gets a message from the regular send queue.  The message should be        */
/*  deleted, but not the LFAPNode structure.  If there are no messages        */
/*  available, 0 is returned.                                                 */
/*                                                                            */
/******************************************************************************/
LFAPNode* getMsgFromLfapRegSendQueue (void)
{
    LFAPNode* node;

    int ov = LFAPLock();

    node = getFirstLfapNode(&LFAP_send_reg_queue);

    LFAPUnlock(ov);
    return node;
}

/******************************************************************************/
/*                                                                            */
/*  Function: getMsgFromLfapHighSendQueue                                     */
/*                                                                            */
/*  Gets a message from the high send queue.  The message should be           */
/*  deleted, but not the LFAPNode structure.  If there are no messages        */
/*  available, 0 is returned.                                                 */
/*                                                                            */
/******************************************************************************/
LFAPNode* getMsgFromLfapHighSendQueue (void)
{
    LFAPNode* node;

    int ov = LFAPLock();
 
    node = getFirstLfapNode(&LFAP_send_high_queue);

    LFAPUnlock(ov);
    return node;
}

/******************************************************************************/
/*                                                                            */
/*  Function: returnLfapSendMsg                                               */
/*                                                                            */
/*  Returns the structure that holds a send message back so the queues        */
/*  can reuse it.  The internals of the LFAPSendMsg structure should be set   */
/*  to 0 before it is returned.                                               */
/*                                                                            */
/******************************************************************************/
void returnLfapSendMsg (LFAPNode* msg)
{
    int ov = LFAPLock();

    if (LFAP_send_free_list._entries_ < LFAP_MAX_FREE_SEND_QUEUE_SIZE)
        appendLfapNode(&LFAP_send_free_list, msg);
    else
        freeLfapMsgNode_(msg);

    LFAPUnlock(ov);
}



/******************************************************************************/
/*                                                                            */
/*  Function: addMsgToLfapRegRcvdQueue                                        */
/*                                                                            */
/*  Adds a message to the regular received queue.  The message should already */
/*  be allocated.  Message is appended to the end of the queue.  Returns 1 if */
/*  successful, 0 if not.                                                     */
/*                                                                            */
/******************************************************************************/
lfapui16_t addMsgToLfapRegRcvdQueue (void* buf, unsigned int len)
{
    return addMsgToLfapMsgQueue_(&LFAP_rcvd_reg_queue,
                                           getFreeLfapRcvdMsgNode_(), buf, len);
}

/******************************************************************************/
/*                                                                            */
/*  Function: addMsgToLfapHighRcvdQueue                                       */
/*                                                                            */
/*  Adds a message to the high received queue.  The message should already    */
/*  be allocated.  Message is appended to the end of the queue.  Returns 1 if */
/*  successful, 0 if not.                                                     */
/*                                                                            */
/******************************************************************************/
lfapui16_t addMsgToLfapHighRcvdQueue (void* buf, unsigned int len)
{
    return addMsgToLfapMsgQueue_(&LFAP_rcvd_high_queue,
                                           getFreeLfapRcvdMsgNode_(), buf, len);
}

/******************************************************************************/
/*                                                                            */
/*  Function: msgsInLfapRegRcvdQueue                                          */
/*                                                                            */
/*  Returns the number of messages currently in the regular received queue.   */
/*                                                                            */
/******************************************************************************/
unsigned int msgsInLfapRegRcvdQueue (void)
{
    return LFAP_rcvd_reg_queue._entries_;
}

/******************************************************************************/
/*                                                                            */
/*  Function: msgsInLfapHighRcvdQueue                                         */
/*                                                                            */
/*  Returns the number of messages currently in the high received queue.      */
/*                                                                            */
/******************************************************************************/
unsigned int msgsInLfapHighRcvdQueue (void)
{
    return LFAP_rcvd_high_queue._entries_;
}

/******************************************************************************/
/*                                                                            */
/*  Function: getMsgFromLfapRegRcvdQueue                                      */
/*                                                                            */
/*  Gets a message from the regular received queue.  The message should be    */
/*  deleted, but not the LFAPNode structure.  If there are no messages        */
/*  available, 0 is returned.                                                 */
/*                                                                            */
/******************************************************************************/
LFAPNode* getMsgFromLfapRegRcvdQueue (void)
{
    return getFirstLfapNode(&LFAP_rcvd_reg_queue);
}

/******************************************************************************/
/*                                                                            */
/*  Function: getMsgFromLfapHighRcvdQueue                                     */
/*                                                                            */
/*  Gets a message from the high received queue.  The message should be       */
/*  deleted, but not the LFAPNode structure.  If there are no messages        */
/*  available, 0 is returned.                                                 */
/*                                                                            */
/******************************************************************************/
LFAPNode* getMsgFromLfapHighRcvdQueue (void)
{
    return getFirstLfapNode(&LFAP_rcvd_high_queue);
}

/******************************************************************************/
/*                                                                            */
/*  Function: returnLfapRcvdMsg                                               */
/*                                                                            */
/*  Returns the structure that holds a received message back so the queues    */
/*  can reuse it.  The internals of the LFAPRcvdMsg structure should be set   */
/*  to 0 before it is returned.                                               */
/*                                                                            */
/******************************************************************************/
void returnLfapRcvdMsg (LFAPNode* msg)
{
    if (LFAP_rcvd_free_list._entries_ < LFAP_MAX_FREE_RCVD_QUEUE_SIZE)
        appendLfapNode(&LFAP_rcvd_free_list, msg);
    else
        freeLfapMsgNode_(msg);
}

#endif /* ifdef _WIN32 */
